=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Configuration - module

=head1 SYNOPSIS

use LogicBuddy::Configuration;
LogicBuddy::Configuration::set_language( $language );

=head1 ABSTRACT

LogicBuddy::Configuration abstracts LogicBuddy configuration.

It defines a lot of system global constants and variables.

Some values are stored persistent in a configuration file.

=head1 DESCRIPTION

=head3 set_language

   LogicBuddy::Configuration::set_language( $language );

defines the current language of the interface.
The language will be persistent.

=head3 language

   $language = LogicBuddy::Configuration::language();

returns the current language.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;
use utf8;

use File::HomeDir;
use Data::Dumper;
use Safe;

use Readonly;

use LogicBuddy::Version;

package LogicBuddy::Configuration;

use English '-no_match_vars';

################################################################################
# configuration
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

################################################################################
# global constants
################################################################################

use Readonly;

BEGIN {
   use Readonly;
   Readonly $LogicBuddy::Version      => '0.2.3';
   Readonly $LogicBuddy::Release_type => 'alpha';
   Readonly my $C_status_file_name    => '.logicbuddy.res';
   Readonly $LogicBuddy::C_home_dir   => File::HomeDir->my_home();
   Readonly $LogicBuddy::status_file  => "$LogicBuddy::C_home_dir/$C_status_file_name";
}; # BEGIN

Readonly my $C_File_access_timeout      => 10;

Readonly $LogicBuddy::C_max_name_length => 4;

################################################################################
# global variable
################################################################################

$LogicBuddy::G_Debug          = FALSE;
$LogicBuddy::G_language       = 'German';

################################################################################
# internal subprograms
################################################################################

sub I_initialize {

   my $compartment = new Safe( 'LogicBuddy::Settings' );
   $compartment->rdo( $LogicBuddy::status_file );
   return if $EVAL_ERROR;

   return if not $LogicBuddy::Settings::Persistent;

   $LogicBuddy::G_language = $LogicBuddy::Settings::Persistent->{language} || 'German';

   return;

}; # I_initialize

sub I_save_data {

   my $workaround = 'my $hgkutfz = $main::INC; # avoids warning "Name "main::INC" used only once ...'; # the perl god knows why

   my %saved  = ();
   $saved{language} = $LogicBuddy::G_language;

   use Fcntl qw( :DEFAULT :flock );

   my $dumper = Data::Dumper->new(
      [ \%saved ],
      [ 'Persistent' ]
   );

  eval {
      local $SIG{ALRM} = sub { alarm(0); die "File access exception: timeout" };
      alarm $C_File_access_timeout;
      sysopen( DATA, $LogicBuddy::status_file, O_WRONLY | O_CREAT ) or die "File access exception: $OS_ERROR";
      flock( DATA, LOCK_EX );
      truncate( DATA, 0 );
      $dumper->Indent( 0 );
      DATA->autoflush(TRUE);
      print DATA "$workaround\n", $dumper->Dump;
   };
   alarm 0;
   close DATA;
   print STDERR "Error saving $LogicBuddy::status_file $EVAL_ERROR\n" if $EVAL_ERROR;
   return;

}; # I_save_data

################################################################################
# exported subprograms
################################################################################

sub set_language {
   ( $LogicBuddy::G_language ) = @_;
   return;
}; # set_language

sub language {
   return $LogicBuddy::G_language;
}; # language

################################################################################
# initialize begin end
################################################################################

I_initialize();

END {
   I_save_data();
   return;
}; # END

################################################################################
# end LogicBuddy::Configuration
################################################################################

1;

