=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische_Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Espresso - module

=head1 SYNOPSIS

   use LogicBuddy::Espresso;

   $espresso = LogicBuddy::Espresso->new( $viewer );
   $espresso->generate();

=head1 ABSTRACT

LogicBuddy::Espresso abstracts interpreting and calculation of espresso tables.

=head1 DESCRIPTION

=head3 new

    $espresso = LogicBuddy::Espresso->new( $viewer );

Instantiates an espresso processor using $viewer for output.

=head3 generate

   $espresso->generate();

Takes the truth table from the associated viewer,
generates the espresso result functions and displays it using the associated viewer.

=head1 AUTHOR2

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische_Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;
use warnings;

use File::Temp;

package LogicBuddy::Espresso;

use English '-no_match_vars';

################################################################################
# constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

use Readonly;

Readonly my $C_Error_margin     => ' ' x 3;
Readonly my $C_Error            => "Error generating Espresso table\n$C_Error_margin";
 
Readonly my $C_espresso_file    => Non_OO_temp_file();
Readonly my $C_espresso_command => Non_OO_locate_espresso();

################################################################################
# class LogicBuddy::Espresso
################################################################################

sub new {
   my ( $proto, $viewer )= @_;

   my $class = ref($proto) || $proto;
   my $self;

   $self->{Viewer}  = $viewer;
   $self->{error}   = $C_Error;

   bless $self, $class;

   return $self;

}; # new

################################################################################
# Non OO internal methods
################################################################################

sub Non_OO_locate_espresso {

   if ( $^O eq 'MSWin32' ) {
      DIR: foreach my $dir ( (split /;/, $ENV{PATH}), @INC ) {
         next DIR if $dir =~ m{ \A [.] }xms;
         my $win_dir = $dir;
         $win_dir =~ s{ / }{\\}gxms;
         return "$win_dir\\espresso.exe"      if -e "$dir/espresso.exe";
         return "$win_dir\\bin\\espresso.exe" if -e "$dir/bin/espresso.exe";
         return "$win_dir\\etc\\espresso.exe" if -e "$dir/etc/espresso.exe";
      }; # DIR
      return "espresso.exe"      if -e "espresso.exe";
      return "bin\\espresso.exe" if -e "bin/espresso.exe";
      return "etc\\espresso.exe" if -e "etc/espresso.exe";
      return "LogicBuddy\\espresso.exe" if -e "LogicBuddy/espresso.exe";
      return "LogicBuddy\\Support\\espresso.exe" if -e "LogicBuddy/Support/espresso.exe";
   } else {
      foreach my $dir ( (split /:/, $ENV{PATH}), @INC ) {
         return "$dir/espresso"     if -e "$dir/espresso";
         return "$dir/bin/espresso" if -e "$dir/bin/espresso";
         return "$dir/etc/espresso" if -e "$dir/etc/espresso";
      };
   };

   return 'espresso';

}; # Non_OO_locate_espresso

sub Non_OO_temp_file {

   my $dir = ( $^O eq 'MSWin32' )  ?  $ENV{TEMP}  :  '/tmp';
   $dir = EMPTY if not $dir;

   return File::Temp->new( TEMPLATE => 'temp_EspresSo_XXXXX',
                           DIR      => $dir,
                           SUFFIX   => '.dat'
                         );

}; # Non_OO_temp_file

################################################################################
# internal methods
################################################################################

sub I_get_values_of {
   my ( $self, @values )= @_;

   return ( map { $_  eq 'd'  ?  '-'  : $_ } @values );

}; # I_get_values_of

sub I_generated_espresso_file {
   my ( $self )= @_;

# 2-bit by 2-bit binary adder (with no carry input)
#                .i 4
#                .o 3
#             .ilb a1 a0 b1 b0
#             .ob s2 s1 s0
#                0000  000
#                0001  001
#                0010  010
#                0011  011
#                0100  001
#                0101  010
#                0110  011
#                0111  100
#                1000  010
#                1001  011
#                1010  100
#                1011  101
#                1100  011
#                1101  100
#                1110  101
#                1111  110
#             .e

   my @input_labels            = $self->{Viewer}->input_labels();
   my @output_labels           = $self->{Viewer}->output_labels();
   my $number_of_input_labels  = scalar @input_labels;
   my $number_of_output_labels = scalar @output_labels;

   my $ESPRESSO;
   if ( not open $ESPRESSO, '>', $C_espresso_file ) {
      $self->{Viewer}->message( "$self->{error} $C_espresso_file: $!.\n" );
      $self->{error} = $C_Error_margin;
      return FALSE;
   };
 
   print $ESPRESSO ".i $number_of_input_labels\n";
   print $ESPRESSO ".o $number_of_output_labels\n\n";

   print $ESPRESSO ".ilb ", (join ' ', @input_labels), "\n";
   print $ESPRESSO ".ob ",  (join ' ', @output_labels), "\n\n";

   foreach my $index ( 0 .. (2**$number_of_input_labels) - 1 ) {
      my $input_bits    = join EMPTY, LogicBuddy::Utilities::bits_of( $index, $number_of_input_labels );
      my $output_values = join EMPTY, $self->I_get_values_of( $self->{Viewer}->values_of( $index ) );
      print $ESPRESSO "$input_bits $output_values\n";
   };

   print $ESPRESSO "\n.e\n";

   if ( not close $ESPRESSO ) {
      $self->{Viewer}->message( "$self->{error} $C_espresso_file: $!.\n" );
      $self->{error} = $C_Error_margin;
      return FALSE;
   };

   return TRUE;

}; # I_generated_espresso_file

sub I_run_espresso {
   my ( $self )= @_;

   return if not -e $C_espresso_file;
   return `$C_espresso_command $C_espresso_file 2>&1`;

}; # I_run_espresso

sub I_interpret_espresso {
   my ( $self, $string )= @_;

   if ( not $string ) {
      $self->{Viewer}->message( LogicBuddy::Singular::Error_interpreting_Espresso_table() . ".\n" .
                                LogicBuddy::Singular::No_table_defined() . ".\n"
                              );
      return;
   };

   my $error        = FALSE;
   my $productions;
   my @terms        = ();
   my $line         = EMPTY;
   foreach $line ( split /\n/, $string ) {
      if ( $line =~ m/ \A espresso: /xms ) {
         $self->{Viewer}->message( LogicBuddy::Singular::Error_interpreting_Espresso_table() . ".\n" ) if not $error;
         $self->{Viewer}->message( "  $line\n" );
         $error = TRUE;
      } else {
         if ( $line =~ m/ \A [.]i \s (\d+) \s* \n /xms ) {
            my $number_of_input = $1;
            if ( $number_of_input != scalar $self->{Viewer}->input_labels() ) {
               $self->{Viewer}->message( LogicBuddy::Singular::Error_interpreting_Espresso_table() . ".\n" ) if not $error;
               $self->{Viewer}->message( "   " . LogicBuddy::Singular::Number_of_input_names() . " = $number_of_input " .
                                                 LogicBuddy::Singular::is_not_expected() .
                                         " " .  (scalar $self->{Viewer}->input_labels() ) . "\n"
                                       );
               $error = TRUE;
            };
         } elsif ( $line =~ m/ \A [.]i \s (\d+) \s* \n /xms ) {
            my $number_of_output = $1;
            if ( $number_of_output != scalar $self->{Viewer}->output_labels() ) {
               $self->{Viewer}->message( LogicBuddy::Singular::Error_interpreting_Espresso_table() . ".\n" ) if not $error;
               $self->{Viewer}->message( "   " . LogicBuddy::Singular::Number_of_output_names() . " = $number_of_output " .
                                                 LogicBuddy::Singular::is_not_expected() .
                                         " " .  (scalar $self->{Viewer}->output_labels() ) . "\n"
                                       );
               $error = TRUE;
            };
         } elsif ( $line =~ s/ \A [.]ilb \s+ //xms ) {
            my %known = map { ( $_ => TRUE ) } $self->{Viewer}->input_labels();
            foreach my $name ( split / /, $line ) {
               if ( not exists $known{$name} ) {
                  $self->{Viewer}->message( LogicBuddy::Singular::Error_interpreting_Espresso_table() . ".\n" ) if not $error;
                  $self->{Viewer}->message( "   " . LogicBuddy::Singular::Output_name() . " = $name " .
                                                    LogicBuddy::Singular::is_not_known() . ".\n"
                                          );
                  $error = TRUE;
               };
            };
         } elsif ( $line =~ s/ \A [.]ob \s+ //xms ) {
            my %known = map { ( $_ => TRUE ) } $self->{Viewer}->output_labels();
            foreach my $name ( split / /, $line ) {
               if ( not exists $known{$name} ) {
                  $self->{Viewer}->message( LogicBuddy::Singular::Error_interpreting_Espresso_table() . ".\n" ) if not $error;
                  $self->{Viewer}->message( "   " . LogicBuddy::Singular::Output_name() . " = $name " .
                                                    LogicBuddy::Singular::is_not_known() . ".\n"
                                          );
                  $error = TRUE;
               };
            };
         } elsif ( $line =~ m/ \( warning \): input \s+ line \s+ \#\d+ \s+ ignored /xms ) {
         } elsif ( $line =~ m/ \A [.]p \s+ ( \d+ ) \s* \z /xms ) {
            $productions = $1;
         } elsif ( $line =~ m/ \A ( [\-01]+ ) \s+ ( [01]+ ) /xms ) {
            push @terms, [ $1, $2 ];
         };
      };
   };

   if ( (not defined $productions) or $productions != scalar @terms ) {
      $productions = 'undefined' if not defined $productions;
      $self->{Viewer}->message( LogicBuddy::Singular::Error_interpreting_Espresso_table() . "\n" ) if not $error;
      $self->{Viewer}->message( "   " . LogicBuddy::Singular::Number_of_generated_terms() . " " . (scalar @terms) . " " .
                                LogicBuddy::Singular::is_not_expected() . " $productions\n"
                              );
      $error = TRUE;
   };

   return if $error;

   $self->I_generate_result( @terms );

}; # I_interpret_espresso

sub I_generate_result {
   my ( $self, @terms ) = @_;

   my $result        = EMPTY;
   my $output_select = '1';
   foreach my $output_name ( $self->{Viewer}->output_labels() ) {
      my $preformatted  = EMPTY;
      my $first_term    = TRUE;
      my $margin        = ' ' x length $output_name;
      my $is_zero       = TRUE;
      my $only_one_term = TRUE;
      $preformatted .= "$output_name =    ( ";
      foreach my $term ( @terms ) {
         my ( $input, $output ) = @$term;
         if ( $output =~ m/\A$output_select/ ) {
            $is_zero = FALSE;
            if ( not $first_term ) {
               $preformatted .= ")\n$margin   OR ( ";
               $only_one_term = FALSE;
            };
            $first_term = FALSE;
            my $first   = TRUE;
            $preformatted .= '1 ' if $input =~ m/ \A -+ \z /xms; # special case
            INPUT: foreach my $input_name ( $self->{Viewer}->input_labels() ) {
               next INPUT if $input =~ s/ \A - //xms;
               $preformatted .= 'AND ' if not $first;
               $first = FALSE;
               if ( $input =~ s/ \A 0 //xms ) {
                  $preformatted .= '^';
               } elsif ( $input =~ s/ \A 1 //xms ) {
               };
               $preformatted .= "$input_name ";
            }; # INPUT
         };
      };
      $preformatted .= '0 ' if $is_zero;
      $preformatted .= ")\n";
      $preformatted =~ s/[()]/ /g if $only_one_term;
      $result .= $self->I_convert_expression( $preformatted );
      $output_select = ".$output_select";
      $result =~ s{ ( [\s\n]* ) \z }{.$1}xms;
   };
   $self->{Viewer}->solution( "$result", "Minimized Boolean Expression of above Table" );

}; # I_generate_result

sub I_convert_expression {
   my ( $self, $function ) = @_;

   if ( $self->{Viewer}->logic_type() eq 'NAND' ) {
      return $self->I_convert_to_NAND( $function );
   } else {
      return $function;
   };

}; # I_convert_expression

sub I_convert_to_NAND {
   my ( $self, $function ) = @_;

   return $function if $function =~ m/ \s = \s+ [01] \s* \z /xms;

   if ( $function !~ s/ \A ([a-zA-Z0-9_]+) \s+ = \s+ //xms ) {
      $self->{Viewer}->message( LogicBuddy::Singular::Error_interpreting_Espresso_table() . "\n" );
      $self->{Viewer}->message( LogicBuddy::Singular::Output_name_missing() . "\n" );
      return;
   };

   my $result = "$1 = ";
   my $margin = ' ' x length $result;

   if ( $function =~ m/ \s OR \s /xms ) {
      $result  .= 'NAND ( ';
      my $first = TRUE;
      foreach my $and_term ( split /\s*OR\s*/, $function ) {
         my $term = $self->I_convert_term_to_NAND( $and_term );
         $result .= ",\n$margin       " if not $first;
         $result .= $term;
         $first   = FALSE;
      };
      $result .= "\n$margin      )\n";
   } elsif ( $function =~ m/AND/xms ) {
      my $term = $self->I_convert_term_to_NAND( $function, $margin );
      $result .= "NOT( $term\n$margin    )\n";
   } elsif ( $function =~ m/\^/xms ) {
      my $term = $self->I_convert_term_to_NAND( $function );
      $result .= "$term\n";
   } else {
      $result .= $function;
   };

   return $result;

}; # I_convert_to_NAND

sub I_convert_term_to_NAND {
   my ( $self, $term ) = @_;

   if ( $term =~ s/ \s+ AND /,/gxms ) {
              if ( $term =~ s/ \A \s* \( \s* //xms ) {
                   $term =~ s/ \s* \) (\s|\n)* \z //xms;
              };
              $term =~ s/ \^ ([a-zA-Z0-9_]+) (\s|\n)* /NOT($1)/gxms;
              $term =~ s/ (\s|\n)* \z //xms;
      return "NAND ( $term )";
   } elsif ( $term =~ s/ \( \s* \^ \s* ([a-zA-Z0-9_]+) (\s|\n)* \) (\s|\n)* /$1/xms ) {
      return $term;
   } elsif ( $term =~ s/ \s* \^ \s* ([a-zA-Z0-9_]+) (\s|\n)* /NOT($1)/xms ) {
      return $term;
   } elsif ( $term =~ s/ \( \s* ([a-zA-Z0-9_]+) (\s|\n)* \) (\s|\n)* /NOT($1)/xms ) {
      return $term;
   } elsif ( $term =~ s/ \s* ([a-zA-Z0-9_]+) (\s|\n)* /NOT($1)/xms ) {
      return $term;
   } else {
      return $term;
   };

}; # I_convert_term_to_NAND

################################################################################
# exported methods
################################################################################

sub generate {
   my ( $self )= @_;

   return if not $self->I_generated_espresso_file();

   my $result = $self->I_run_espresso();

   $self->I_interpret_espresso( $result );

   return;

}; # generate

################################################################################
# end LogicBuddy::Espresso
################################################################################

1;

