=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Locale - module

=head1 SYNOPSIS

use LogicBuddy::Locale;

   $singular_plural_ref = LogicBuddy::Locale::TranslationOf( $token );
   $singular_plural_ref = LogicBuddy::Locale::TranslationOf( $token, $text_1 );
   $singular_plural_ref = LogicBuddy::Locale::TranslationOf( $token, $text_1, $text_2 );

   $translated_text = LogicBuddy::Locale::SingularOf( $sentence );
   $translated_text = LogicBuddy::Locale::PluralOf( $sentence );

   LogicBuddy::Locale::SetLanguage( $language );

   $translated_text = LogicBuddy::Locale::A_sentence();

   LogicBuddy::Locale::Set_local_of( $locale, $do_not_transform, %language_values );

=head1 ABSTRACT

LogicBuddy::Locale abstracts LogicBuddy locale.

LogicBuddy::Locale defines the translation of sentences_like_this into a target language.
The sentences_like_this should be an meaningful English sentence with the spaces replaced by '_'.
If there is no translation defined the sentences_like_this is translated to 'sentences like this'.
That is all '_' are replaced by ' '.

Sentences like This_is_a_sentence_with_9_word_and_8_character are used to insert parameters into
the text. The '_9_' and '_8_' can be replaced by parameter. See LogicBuddy::Locale::TranslationOf.

=head1 DESCRIPTION

=head2 EXPORTED SUBPROGRAMS

=head3 TranslationOf

   $singular_plural_ref = LogicBuddy::Locale::TranslationOf( $token );
   $singular_plural_ref = LogicBuddy::Locale::TranslationOf( $token, $text_1 );
   $singular_plural_ref = LogicBuddy::Locale::TranslationOf( $token, $text_1, $text_2 );

Translates the $token text into the current language.
   $singular_plural_ref is [ $singular_text, $plural_text ].
   The token is a sentence like 'This_is_a_sentence_with_9_word_and_8_character'.
   The '_9_' and '_8_' are placeholder which may be replaced.
   If $text_1 is defined and $token has a _9_ element the 9 ia replaced by $text_1.
   If $text_2 is defined and $token has a _8_ element the 8 ia replaced by $text_2.
   The $singular_text is the singular version of $token, $text_1 and $text_2.
   The $plural_text is the plural version of $token, $text_1 and $text_2.

=head3 SingularOf PluralOf

   $translated_text = LogicBuddy::Locale::SingularOf( $sentence );
   $translated_text = LogicBuddy::Locale::PluralOf( $sentence );

Returns the singular/plural of the translation of $word

=head3 SetLanguage

   LogicBuddy::Locale::SetLanguage( $language );

Sets the language to be translated to.
If $language is not defined or not known English is used.

=head3 A_sentence

   $translated_text = LogicBuddy::Locale::A_sentence();

A_sentence may be any sequence of words and _.
The A_sentence should be an meaningful English sentence with the spaces replaced by '_'.

=head3 Set_local_of

   LogicBuddy::Locale::Set_local_of( $locale, $do_not_transform, %language_values );

Defines / sets the $locale sentence to the language_values.
The %language_values are like ( Language => 'Translated sentence', .. )
The Language has to be defined in LogicBuddy::Locale.
Currently English and German are defined.
If the Language is not defined an exception is raised.

If $do_not_transform is TRUE the underscores etc. in the translated sentence are not transformed.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;
use warnings;

use LogicBuddy::Configuration;

# use LogicBuddy::Persistent;

################################################################################
# internal subprograms
################################################################################

package LogicBuddy::Locale;

################################################################################
# global constants
################################################################################

use utf8;

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

use Readonly;

Readonly my $C_margin          => $Tk::Configuration::C_margin;

Readonly my $C_URL             => 'Praktische-Elektronik.dr-k.de';
Readonly my $C_Copyright_HK_de => "Diese Software unterliegt dem Copyright © 2000 - 2016 von Heiner Kuhlmann\nhttp://$C_URL";
Readonly my $C_Copyright_HK    => "This software is Copyright © 2000 - 2016 by Heiner Kuhlmann\nhttp://$C_URL";

Readonly my $C_Windows_note_de => ( $^O eq 'MSWin32' )  ?
                                 "\n\nUnter Windows wird eine 'Eingabe-Aufforderung' geöffnet. " .
                                 "Diese darf nicht geschlossen werden. Wenn sie stört, wird sie am besten minimiert.\n"
                                 :  EMPTY;
Readonly my $C_Windows_note    => ( $^O eq 'MSWin32' )  ?
                                 "\n\nOn Windows a 'Command Prompt' opens. It must not be closed. If it bothers you, it is best minimized.\n"
                                 : EMPTY;

Readonly my $C_LogicBuddy_license =>
                                 "LogicBuddy is Perl.\n" .
                                 "LogicBuddy is subject to the 'The Artistic License' of Perl " .
                                 "and the 'GNU LESSER GENERAL PUBLIC LICENSE Version 3'";
Readonly my $C_LogicBuddy_license_de =>
                                 "LogicBuddy ist in Perl geschrieben.\n" .
                                 "LogicBuddy unterliegt der der 'The Artistic License' von Perl " .
                                 "und der 'GNU LESSER GENERAL PUBLIC LICENSE Version 3'.";

Readonly my $C_Espresso       => 'Espresso is a two-level (PLA) logic minimizer developed at the University of California, Berkeley.';
Readonly my $C_Espresso_de    => 'Espresso ist ein Logik Minimierer für zwei Ebenen (PLA) der University of California, Berkeley.';

Readonly my $C_Strawberry     => 'Strawberry Perl is a distribution of the Perl programming language for Microsoft Windows.';
Readonly my $C_Strawberry_de  => 'Strawberry Perl ist eine Distribution der Programmiersprache Perl für Microsoft Windows';

Readonly my $C_Syntax_descipton_de =>"
Die logischen Funktionen können auf zwei Weisen formuliert werden:
o mathematisch:
   x = a and b.
o Umgangssprachlich:
   x ist a und b.

Im wesentlichen bestehen die Funktionen aus Aussagen.
Sie beginnen mit einer Variablen z.B.  \"x\" und enden mit einem Punkt \".\".
Es ist ohne weiteres möglich, mehrere Aussagen zu der gleichen Variablen zu machen.
Die letzte überschreibt vorherige Aussagen.
LogicBuddy gibt dann eine entsprechend Warnung aus.

Bei Fehlern in einer Aussage wird er durch LogicBuddy angezeigt und eine Hilfe gegeben.

Logische Funktionen sind häufig nicht so einfach wie die Beispiele oben.
Es sind Formulierungen mit AND, OR, NOT bzw. UND, ODER, NICHT möglich.
Meistens ist nicht jedem klar, was z.B. mit

x = a und b oder c.

gemeint ist.

Heißt es zuerst a UND b und dann ODER mit c. Also

x = (a und b) oder b.

aber es könnte auch

x = a und (b oder c).

gemeint sein.

Zunächst der Tip: Wer sich nicht sicher, ist setzt Klammern.

Die Regel ist UND vor ODER - wie beim Rechnen Punktrechnung vor Strichrechnung.
Abweichende Fälle werden durch Klammern beschrieben.

Für UND kann sowohl 'und' als auch 'and' eingesetzt werden.
Für ODER 'oder' bzw. 'or'.
Das 'ist' ist nichts anderes als '='
Diese Schreibweisen dürfen sogar beliebig gemischt werden. LogicBuddy versteht sie schon.
Anders ist es mit uns Menschen. Wir sollten uns an eine Schreibweise halten.
LogicBuddy weist darauf hin, wenn deutsch und englisch gemischt werden.

";

Readonly my $C_AND_OR         => "The logical functions are described using\n" .
                                 "$C_margin 'AND' and 'OR' and '^' for 'NOT'";
Readonly my $C_AND_OR_de      => "Die logischen Funktionen werden mit\n" .
                                 "$C_margin 'AND' und 'OR' sowie '^' für 'NOT' beschrieben";

Readonly my $C_NAND_NOT       => "The logical functions are described using\n" .
                                 "$C_margin 'NAND' and 'NOT'";
Readonly my $C_NAND_NOT_de    => "Die logischen Funktionen werden mit\n" .
                                 "$C_margin 'NAND' und 'NOT' beschrieben";

Readonly my $C_Names_de       => "Namen bestehen aus maximal $LogicBuddy::C_max_name_length Buchstaben (a-z) und Ziffern,\n" .
                                 "$C_margin Sie beginnen mit einem Buchstaben und können einzelne Unterstriche enthalten.";
Readonly my $C_Names          => "Names consist of up to $LogicBuddy::C_max_name_length letters and digits,\n" .
                                 "$C_margin starting with a letter and can contain single underscores.";

Readonly my $C_Input_names_de => "Namen von Eingangsvariablen\n" .
                                 "$C_margin $C_Names_de";
Readonly my $C_Input_names    => "Names of input variable\n" .
                                 "$C_margin $C_Names";

Readonly my $C_Function_names_de
                              => "Namen von Ausgangsvariablen bzw. Funktionen\n" .
                                 "$C_margin $C_Names_de";
Readonly my $C_Function_names => "Name of output variable or functions\n" .
                                 "$C_margin $C_Names";

Readonly my $C_Balloon_de     => 'Für alle wichtigen Felder und Buttons wird beim Berühren mit der Maus eine Kurzhilfe (Ballon-Hilfe) angezeigt. ';
Readonly my $C_Balloon        => 'For all major fields and buttons while touching the mouse a balloon help is displayed. ';

Readonly my $C_Balloon_help_de
                              => "$C_Balloon_de Mit diesem Button kann die Ballon-Hilfe ein- und ausgeschaltet werden.";
Readonly my $C_Balloon_help   => "$C_Balloon Using this button the Balloon Help can be turned on and off.";

Readonly my $C_Copy_and_Paste_de
                              => 'Der Inhalt kann mit Copy <Strg + c> und Paste <Strg + v> übertragen werden'; 
Readonly my $C_Copy_and_Paste => 'The contents can be transferred with copy <Ctrl + c> and paste <Ctrl + v>';
Readonly my $C_Description_Solution_de
                              => 'Im Fenster mit der letzten Lösung wird für alle Funktionen (Ausgangsvariablen) die minimale Lösung angezeigt. ' .
                                 'Die einzelnen Ausdrücke werden je nach der gewählten Logik (im Beispiel NAND) dargestellt. ' .
                                 'Die Lösung wird aktualisiert, wenn die Wahrheitstabelle geändert wird. ' .
                                 "$C_Copy_and_Paste_de. " .
                                 'Für umfangreiche Lösungen kann ein eigener Dialog für die Lösung geöffnet werden:';
Readonly my $C_Description_Solution
                              => 'The window with the last solution (output variables) displays the minimized solution of all functions. ' .
                                 'The individual terms are shown according to the selected logic (in the example NAND). ' .
                                 'The solution is updated when the truth table is changed. ' .
                                 "$C_Copy_and_Paste. " .
                                 'For comprehensive solutions, a separate dialog for the solution can be opened:';

Readonly my $C_Description_help_de
                              => 'Hilfe-Fenster enthalten Texte und Bilder zu einem Thema oder Dialog. ' .
                                 'Verweise auf weitere Hilfen werden blau dargestellt. Sie öffnen ein neues Hilfe-Fenster. ' .
                                 'Einige Bilder können durch anklicken vergrößert werden.';
Readonly my $C_Description_help
                              => 'Help window containing text and images on a topic or dialogue. ' .
                                 'References to other help are shown in blue. It opens a new Help window. ' .
                                 'Some images can be enlarged by clicking.';

Readonly my $C_Description_truth_table_de
                              => 'Die Wahrheitstabelle besteht aus vier Bereichen:';
Readonly my $C_Description_truth_table_1_de
                              => 'Der Kopfzeile mit den Namen der Eingangsvariablen (gelb) und logischen Funktionen bzw. Ausgangsvariablen (grau). ' .
                                 'Die Felder mit Eingangsvariablen und Ausgangsvariablen können angeklickt werden. ' .
                                 'Dann öffnet sich ein Dialog zum Ändern des Namens.';
Readonly my $C_Description_truth_table_2_de
                              => 'Die linke Spalte numeriert die Zeilen. Ein Klick aktiviert:' ;
Readonly my $C_Description_truth_table_3_de
                              => 'Die gelben Felder in der Mitte stellen alle möglichen Kombinationen der Eingangsvariablen dar.';
Readonly my $C_Description_truth_table_4_de
                              => 'Die grauen Felder auf der rechten Seite beinhalten die Werte der logischen Funktionen, ' .
                                 'jede in einer eigenen Spalte. ' .
                                 'Durch Anklicken ändern die einzelnen Felder die Werte von 0 - 1 - d . ';
Readonly my $C_Description_truth_table
                              => 'The truth table consists of four areas:';
Readonly my $C_Description_truth_table_1
                              => 'The header with the name of the input variables (yellow) and logic functions and output variables (gray). ' .
                                 'Fields with input variables and output variables can be clicked. ' .
                                 'A dialog opens for changing the name.';
Readonly my $C_Description_truth_table_2
                              => 'The left column numbered rows. A click activates:';
Readonly my $C_Description_truth_table_3
                              => 'The yellow fields in the middle represent all possible combinations of the input variables.';
Readonly my $C_Description_truth_table_4
                              => 'The gray squares on the right side contain the values of logic functions, each in its own column. ' .
                                 'Clicking on a field changes the values from 0 - 1 - d .';

Readonly my $C_LogicBuddy_does =>
                                 "LogicBuddy minimizes logic functions";
Readonly my $C_LogicBuddy_does_de
                              => "LogicBuddy minimiert logische Funktionen";
Readonly my $C_Info_1_de      => "$C_LogicBuddy_does_de.";
Readonly my $C_Info_1         => "$C_LogicBuddy_does.";

Readonly my $C_Check_Function_de
                              => 'Die eingegebene logische Funktion wird überprüft';
Readonly my $C_Check_Function => 'The entered logical function is checked';

Readonly my $C_Check_Function_and_calculate_de
                              => "$C_Check_Function_de, " .
                                 'die entsprechende Wahrheitstabelle in den Hauptdialog eingetragen ' .
                                 'und die Funktion minimiert.';
Readonly my $C_Check_Function_and_calculate
                              => "$C_Check_Function, " .
                                 'the corresponding truth table is entered into the main dialog and the function is minimized. ' .
                                 'The buttons in the truth table are without effect.';

Readonly my $C_description_multiple_change_de =>
                                 'Eine Gruppe von Tabelleneinträgen kann mit diesem Dialog geändert werden. ' .
                                 'Nur die Knöpfe im oberen Fenster sind aktiv. ' .
                                 'Die Wahrheitstabelle im unteren Fenster zeigt nur die Änderungen in ROT an. ' .
                                 'Die Buttons in der Wahrheitstabelle sind ohne Wirkung.';

Readonly my $C_description_multiple_change =>
                                 'A group of table entries can be changed with this dialog. ' .
                                 'Only the buttons in the top window are active. ' .
                                 'The truth table in the bottom window shows only the changes in RED';

Readonly my $C_multiple_change_upper_row_de =>
                                 'Die obere Reihe der Felder stellt die Eingangsvariablen (gelb) und die Funktionen (grau) dar.';

Readonly my $C_multiple_change_upper_row =>
                                 'The upper row of the fields represents the input variables (yellow) and the functions (gray).';

Readonly my $C_multiple_change_lower_row_de =>
                                 "In der unteren Reihe wird im linken grauen Feld die Zeilennummer in der Wahrheitstabelle angegeben. " .
                                 "Ein '?' zeigt an, dass mehrere Zeilen ausgewählt sind. " .
                                 "Die mittleren gelben Felder stellen den Wert der Eingangsvariablen dar: '0', '1' oder 'A' für 'beliebig'. " .
                                 "Mit dem 'A' werden mehrere Zeilen der Wahrheitstabelle ausgewählt. " .
                                 "Die rechten grauen Felder stellen die Werte der Funktionen dar: '0', '1', 'd' oder 'X' für 'nicht ändern'. " .
                                 "Mit dem 'X' werden die Werte der entsprechenden Funktion nicht geändert. " .
                                 "Die Werte der Funktionen werden in die selektierten Zeilen der Wahrheitstabelle eingetragen. ";
Readonly my $C_multiple_change_lower_row =>
                                 "In the lower row, the line number in the left gray field is indicated in the truth table. " .
                                 "A '?' indicates that multiple rows are selected. " .
                                 "The yellow fields in the middle represent the value of the input variables: '0', '1' or 'A' for 'any'. " .
                                 "The 'A' selects multiple lines of the truth table. " .
                                 "The right gray fields represent the values of the functions: '0', '1', 'd' or 'X' for 'do not change'. " .
                                 "The values of the function with a 'X' are not changed. " .
                                 "The values of the functions will be entered into the selected lines of the truth table. ";

Readonly my $C_multiple_change_short_de =>
                                 "Untere Reihe, gelbe Felder, Werte der Eingangsvariablen: '0', '1' oder 'A' für 'beliebig'.\n" .
                                 "Rechte graue Felder, Werte der Funktionen: '0', '1', 'd' oder 'X' für 'nicht ändern'.";
Readonly my $C_multiple_change_short =>
                                 "Lower row, yellow fields, values of input variables: '0', '1' or 'A' for 'any'.\n" .
                                 "Right gray fields, values of functions: '0', '1', 'd' or 'X' for 'do not change'.";

Readonly my $C_multiple_change_toggle_de =>
                                 "Die Werte der Eingangsvariablen und Funktionen werden durch Klicks zyklisch geändert.";
Readonly my $C_multiple_change_toggle =>
                                 "The values of the input variables and functions are changed cyclically by clicking.";

Readonly my $C_multiple_input_truth_table_de =>
                                 'Die selektierten Zeilen werden durch rote Werte der Eingangsvariablen hervorgehoben.';
Readonly my $C_multiple_input_truth_table =>
                                 'The selected rows are highlighted by red values of the input variables.';

Readonly my $C_multiple_function_truth_table_de =>
                                 'Die geänderten Werte von Funktionen werden rot dargestellt.';
Readonly my $C_multiple_function_truth_table =>
                                 'The changed values of functions are also red.';

Readonly my $C_Description_Editor_Function_de =>
                                 'In diesem Fenster können Texte wie in jedem Text-Fenster eingegeben, verändert und gelöscht werden. ' .
                                 "$C_Copy_and_Paste_de."; 
Readonly my $C_Description_Editor_Function =>
                                 'In this window, texts can be entered, changed and deleted as in any text window. ' .
                                 "$C_Copy_and_Paste.";

Readonly my $C_Description_Function_and_Table_match_de =>
                                 'Beschreibt, wie die aktuelle logische Funktion und die Wahrheitstabelle im Hauptdialog zusammen passen.';
Readonly my $C_Description_Function_and_Table_match =>
                                 'Describes how the current logical function and the truth table in the main dialog match.';

Readonly my $C_INPUT_NAME_de =>  "Name einer Eingangsvariablen.\n$C_margin $C_Input_names_de";
Readonly my $C_INPUT_NAME    =>  "Name of an input variable.\n$C_margin " . ucfirst $C_Input_names;

Readonly my $C_ZUWEISUNG     =>  "ZUWEISUNG wie:\n" .
                                 "$C_margin 'var = ...' oder 'var ist ...'\n" .
                                 "$C_margin 'var': $C_Function_names_de";
Readonly my $C_ASSIGNMENT    =>  "ASSIGNMENT like:\n" .
                                 "$C_margin 'var = ...' or 'var is ...'\n" .
                                 "$C_margin 'var': $C_Function_names"; 

Readonly my $C_INPUT_de      =>   "'Eingang = var ...' oder 'Eingang ist var ...'\n$C_margin $C_Input_names_de";
Readonly my $C_INPUT         =>   "'Input = var ...' or 'Input is var ...'\n$C_margin $C_Input_names";

Readonly my $C_OUTPUT_de     =>   "'Ausgang = var ...' oder 'Ausgang ist var ...\n$C_margin $C_Function_names_de";
Readonly my $C_OUTPUT        =>   "'Output = var ... c' or 'Output is var ...\n$C_margin $C_Function_names";

Readonly my $C_Output_or_ASSIGNMENT_de =>
                                 "'$C_OUTPUT_de'   oder\n" .
                                 "$C_margin $C_ZUWEISUNG";
Readonly my $C_Output_or_ASSIGNMENT =>
                                 "'$C_OUTPUT'   or\n" .
                                 "$C_margin $C_ASSIGNMENT";

Readonly my $C_EXPRESSION_de  => "AUSDRUCK wie:\n" .
                                 "$C_margin var\n" .
                                 "$C_margin wahr\n" .
                                 "$C_margin nicht var ...\n" .
                                 "$C_margin var und a1 ... \n" .
                                 "$C_margin var oder a1 ...";
Readonly my $C_EXPRESSION     => "EXPRESSION like:\n" .
                                 "$C_margin var\n" .
                                 "$C_margin true\n" .
                                 "$C_margin not var ...\n" .
                                 "$C_margin var and a1 ...\n" .
                                 "$C_margin var or a1 ...";

Readonly my $C_CONDITION      => "CONDITION like:\n" .
                                 "'when' EXPRESSION.\n" .
                                 "$C_margin $C_EXPRESSION";
Readonly my $C_CONDITION_de   => "BEDINGUNG wie:\n" .
                                 "$C_margin 'wenn' AUSDRUCK.\n" .
                                 "$C_margin $C_EXPRESSION_de";

Readonly my $C_ANY_CONDITION  => "'any' CONDITION.\n" .
                                 "$C_margin $C_CONDITION";
Readonly my $C_ANY_CONDITION_de
                              => "'beliebig' Bedingung.\n" .
                                 "$C_margin $C_CONDITION_de";

Readonly my $C_TERM_de        => "TERM wie:\n" .
                                 "$C_margin var\n" .
                                 "$C_margin wahr\n" .
                                 "$C_margin nicht var ...\n" .
                                 "$C_margin var oder a1 ...";
Readonly my $C_TERM           => "TERM like:\n" .
                                 "$C_margin var\n" .
                                 "$C_margin true\n" .
                                 "$C_margin not var ...\n" .
                                 "$C_margin var or a1 ...";

Readonly my $C_and_TERM_or_dot_de =>
                              "TERM oder 'und' oder '.'\n$C_TERM_de";
Readonly my $C_and_TERM_or_dot =>
                              "TERM or 'and' or '.'\n$C_TERM";

Readonly my $C_FUNCTION_CALL =>
                                 "FUNCTION_CALL like:\n" .
                                 "$C_margin NAND ( ... )\n" .
                                 "$C_margin NOR ( ... )";
Readonly my $C_FUNCTION_CALL_de =>
                                 "FUNKTIONSAUFRUF wie:\n" .
                                 "$C_margin NAND ( ... )\n" .
                                 "$C_margin NOR ( ... )";

Readonly my $C_Bracket_or_NAME_de =>
                                 "'(' or\n" .
                                 "$C_margin $C_Names_de\n" .
                                 "$C_margin $C_FUNCTION_CALL_de\n" .
                                 "$C_margin 'wahr' oder 'falsch' oder '0' oder '1'";
Readonly my $C_Bracket_or_NAME =>
                                 "'(' or\n" .
                                 "$C_margin $C_Names\n" .
                                 "$C_margin $C_FUNCTION_CALL\n" .
                                 "$C_margin 'true' or 'false' or '0' or '1'";

Readonly my $C_FUNCTION_CALL_or_Value_de =>
                                 "Wert oder FUNKTIONSAUFRUF oder '(' AUSDRUCK ')'\n" .
                                 "Wert wie:\n" .
                                 "$C_margin 'var' oder 'wahr' oder 'falsch' oder '0' oder '1'\n" .
                                 "$C_FUNCTION_CALL_de\n" .
                                 "$C_EXPRESSION_de";
Readonly my $C_FUNCTION_CALL_or_Value =>
                                 "Value or FUNCTION_CALL or '(' EXPRESSION ')'\n" .
                                 "Value like:\n" .
                                 "$C_margin 'var' or 'true' or 'false' or '0' or '1'\n" .
                                 "$C_FUNCTION_CALL\n" .
                                 "$C_EXPRESSION";

Readonly my $C_FACTOR =>         "FACTOR like:\n" .
                                 "$C_margin var\n" .
                                 "$C_margin not var\n" .
                                 "$C_margin ( EXPRESSION )\n" .
                                 "$C_EXPRESSION";
Readonly my $C_FACTOR_de =>      "FAKTOR wie:\n" .
                                 "$C_margin var\n" .
                                 "$C_margin nicht var\n" .
                                 "$C_margin ( AUSDRUCK )\n" .
                                 "$C_EXPRESSION_de";

Readonly my $C_AUTOMATIC_ERROR_CORRECTION_de
                         =>      "Einige Fehler werden automatisch im Editor korrigiert. " .
                                 "Sie werden im Editor-Fenster unterstrichen. ".
                                 "Wenn der Cursor im korrigierten Text ist, kann die Korrektur mit <F5> akzeptiert werden.\n" .
                                 "Jede Änderung akzeptiert die Korrektur.";

Readonly my $C_AUTOMATIC_ERROR_CORRECTION
                         =>      "Some errors are automatically corrected in the editor. " .
                                 "They are underlined in the editor window. " .
                                 "If the cursor is in the corrected text, the correction can be accepted with <F5>.\n" .
                                 "Each change accepts the correction.";

Readonly my %C_german_message  => (     #                                'neutral'
                                        #                                [ singulär, plural ]
   '9_not_found'                                                      => '9_nicht gefunden',
   'accept_change'                                                    => 'Änderung akzeptieren',
   'accept_correction'                                                => [ 'Korrektur akzeptieren', 'Korrekturen akzeptieren' ],
   'all'                                                              => 'alles',
   'and'                                                              => 'und',
   'and_TERM_or_dot'                                                  => $C_and_TERM_or_dot_de,
   'AND_OR'                                                           => $C_AND_OR_de,
   'ANY_CONDITION'                                                    => $C_ANY_CONDITION_de,
   'apply'                                                            => 'Anwenden',
   'apply_change'                                                     => [ 'Änderung übernehmen', 'Änderungen übernehmen' ],
   'are'                                                              => 'sind',
   'assign_language'                                                  => 'Sprache zuordnen',
   'assign_no_language'                                               => 'keine Sprache zuordnen',
   'ASSIGNMENT'                                                       => $C_ZUWEISUNG,
   'automatic_error_correction'                                       => 'automatische Fehlerkorrektur',
   'AUTOMATIC_ERROR_CORRECTION'                                       => $C_AUTOMATIC_ERROR_CORRECTION_de,
   'back'                                                             => 'zurück',
   'balloon'                                                          => 'Ballon',
   'balloon_help'                                                     => 'Ballon-Hilfe',
   'balloon_Help'                                                     => $C_Balloon_de,
   'balloon_Help_Full'                                                => $C_Balloon_help_de,
   'balloon_Help_on_off'                                              => 'Ballon-Hilfe ein-/ausschalten',
   'Bracket_or_NAME'                                                  => $C_Bracket_or_NAME_de,
   'but_9_is_not_accepted'                                            => 'aber_9_wird nicht akzeptiert',
   'language_of_balloon_help_will_change_on_restart'                  => 'Die Sprache der Ballon-Hilfe wird nach einem Neustart angepasst',
   'bit_status'                                                       => 'Bit-Status',
   'bold'                                                             => 'fett',
   'border'                                                           => 'Rahmen',
   'bottom'                                                           => 'Ende',
   'calculate_truth_table_of_logical_function'                        => 'Wahrheitstabelle der logischen Funktion berechnen',
   'cancel'                                                           => 'Abbruch',
   'cancel_dialog_and_ignore_changes'                                 => 'Dialog abbrechen und Änderungen verwerfen',
   'cancel_dialog'                                                    => 'Dialog abbrechen',
   'cell'                                                             => [ 'Feld', 'Felder' ],
   'change_9'                                                         => '9_ändern',
   'change_cyclic_9'                                                  => 'zyklisch_9_ändern',
   'change'                                                           => 'ändern',
   'changed_language'                                                 => 'Sprache geändert',
   'change_name'                                                      => 'Namen ändern',
   'change_input_variable_name'                                       => 'Namen der Eingangsvariablen ändern',
   'change_multiple_table_entries'                                    => 'Gruppe von Tabelleneinträgen ändern',
   'change_output_variable_name'                                      => 'Namen der Ausgangsvariablen ändern',
   'change_variable_name'                                             => 'Namen einer Variablen ändern',
   'changing_or_F5_accepts'                                           => 'akzeptieren: ändern oder F5',
   'cannot_use_output_9_in_its_own_assignment'                        => 'kann die Ausgangsvariable_9_nicht in ihrer eigenen Zuweisung verwenden',
   'check',                                                           => 'angenommen',
   'Check_Function'                                                   => $C_Check_Function_de,
   'check_and_calculate_logical_function'                             => $C_Check_Function_and_calculate_de,
   'clear'                                                            => 'löschen',
   'clear_contents_of_cell'                                           => 'Feldinhalt löschen',
   'clear_message'                                                    => [ 'Nachricht löschen', 'Nachrichten löschen' ],
   'clear_logical_function'                                           => 'logische Funktion löschen',
   'click'                                                            => 'klick',
   'click_selects_an_entry'                                           => 'Klick wählt einen Eintrag aus',
   'close'                                                            => 'beenden',
   'close_dialog'                                                     => 'Dialog beenden',
   'close_edit_dialog'                                                => 'Editor-Dialog schließen',
   'close_LogicBuddy'                                                 => 'LogicBuddy beenden',
   'CONDITION'                                                        => $C_CONDITION_de,
   'configuration'                                                    => 'Konfiguration',
   'contents'                                                         => 'Inhaltsverzeichnis',
   'copy'                                                             => 'kopieren',
   'copy_and_paste'                                                   => $C_Copy_and_Paste_de,
   'copy_logical_function_to_editor'                                  => 'logische Funktion in den Editor kopieren',
   'copy_table'                                                       => 'Tabelle kopieren',
   'copy_table_entry'                                                 => 'Tabellen-Element kopieren',
   'Copyright_HK'                                                     => $C_Copyright_HK_de,
   'correction'                                                       => 'Korrektur',
   'could_not_access_9'                                               => 'konnte_9_nicht erreichen',
   'create_9'                                                         => '9_erzeugen',
   'create_new_directory'                                             => 'neues Verzeichnis erstellen',
   'current_value_Toggle_0_1_d'                                       => 'Aktueller Wert. Klick => 0 - 1 - d ändern',
   'current_language'                                                 => 'Deutsch',
   'cut'                                                              => 'ausschneiden',
   'decrement'                                                        => 'verringern',
   'define_new_table'                                                 => 'neue Tabelle definieren',
   'delete_9'                                                         => '9_löschen',
   'delete'                                                           => 'löschen',
   'delete_column'                                                    => 'Spalte löschen',
   'delete_row'                                                       => 'Zeile löschen',
   'Describes_Function_Matches'                                       => $C_Description_Function_and_Table_match_de,
   'description_help'                                                 => $C_Description_help_de,
   'description_multiple_change'                                      => $C_description_multiple_change_de,
   'description_solution_window'                                      => $C_Description_Solution_de,
   'description_truth_table'                                          => $C_Description_truth_table_de,
   'description_truth_table_1'                                        => $C_Description_truth_table_1_de,
   'description_truth_table_2'                                        => $C_Description_truth_table_2_de,
   'description_truth_table_3'                                        => $C_Description_truth_table_3_de,
   'description_truth_table_4'                                        => $C_Description_truth_table_4_de,
   'discard_change'                                                   => [ 'Änderung verwerfen', 'Änderungen verwerfen' ],
   'discard_correction'                                               => [ 'Korrektur verwerfen', 'Korrekturen verwerfen' ],
   'displays_the_last_error_message_when_checking_the_logical_function'
                                                                      => 'zeigt die letzte Fehlermeldung beim Überprüfen der logischen Funktion',
   'dot_or_FACTOR'                                                    => "'.' oder 'oder' oder $C_FACTOR_de",
   'down',                                                            => 'abwärts',
   'edit'                                                             => 'editieren',
   'editor'                                                           => 'Editor',
   'edit_table'                                                       => '(neue) Tabelle editieren',
   'editor_window'                                                    => 'Editor-Fenster',
   'Editor_Function_Description'                                      => $C_Description_Editor_Function_de,
   'end_of_input'                                                     => 'Ende der Eingabe',
   'english'                                                          => 'englisch',
   'enter_here_new_9'                                                 => 'bitte hier_9_eingeben',
   'enter_here_the_logical_function'                                  => 'hier die logische Funktion eingeben',
   'enter_name'                                                       => 'Namen eingeben',
   'enter_and_change_logical_function'                                => 'logische Funktion eingeben und bearbeiten',
   'equal_or_is'                                                      => "'=' oder 'ist'",
   'equal_or_is_or_are'                                               => "'=' oder 'ist' oder 'sind'",
   'equivalent_reserved_word'                                         => [ 'äquivalentes reserviertes Wort', 'äquivalente reservierte Worte' ],
   'error'                                                            => 'Fehler',
   'error_interpreting_Espresso_table'                                => 'Fehler beim Auswerten der Espresso-Tabelle',
   'error_message'                                                    => [ 'Fehlermeldung', 'Fehlermeldungen' ],
   'espresso_software'                                                => $C_Espresso_de,
   'evaluate'                                                         => 'auswerten',
   'example_of_logical_function'                                      => ['Beispiel einer logischen Funktion', 'Beispiele logischer Funktionen'],
   'exit'                                                             => 'Ende',
   'expected'                                                         => 'erwartete',
   'expected_never_9_but'                                             => 'erwartete niemals_9_aber',
   'EXPRESSION'                                                       => $C_EXPRESSION_de,
   'FACTOR'                                                           => $C_FACTOR_de,
   'false'                                                            => 'falsch',
   'field'                                                            => 'Feld',
   'field_range'                                                      => 'Feld-Bereich',
   'file'                                                             => [ 'Datei', 'Dateien' ],
   'file_link'                                                        => 'Link auf Datei',
   'file_not_found'                                                   => 'Datei nicht gefunden',
   'file_9_not_found_create'                                          => 'Datei_9_nicht gefunden. Erzeugen?',
   'find'                                                             => 'suchen',
   'find_next'                                                        => 'weitersuchen',
   'find_previous'                                                    => 'zurück suchen',
   'find_in_9'                                                        => 'suche in_9',
   'forget'                                                           => 'vergessen',
   'formula'                                                          => 'Formel',
   'for_NOT'                                                          => 'für_NOT',
   'frame'                                                            => 'Rahmen',
   'full_syntax_of_logical_function'                                  => 'vollständige Syntax logischer Funktionen',
   'function'                                                         => [ 'Funktion', 'Funktionen' ],
   'FUNCTION_CALL_or_Value'                                           => $C_FUNCTION_CALL_or_Value_de,
   'german'                                                           => 'deutsch',
   'help'                                                             => 'Hilfe',
   'Help_Description_for_Balloon'                                     => "Hilfe-Dialog\nüberall mit <F1> oder\nrechter Maustaste: selektive Hilfe",
   'help_Dialog'                                                      => 'Anzeige dieses Hilfe-Dialogs',
   'help_for_help'                                                    => 'Hilfe zur Hilfe',
   'hide_message'                                                     => [ 'Nachricht verbergen', 'Nachrichten verbergen' ],
   'highlighted_word'                                                 => [ 'hervorgehobenes Wort', 'hervorgehobene Worte' ],
   'hint',                                                            => 'Hinweis',
   'icon'                                                             => [ 'Icon', 'Icons' ],
   'ignore'                                                           => 'ignorieren',
   'include'                                                          => 'inkludieren',
   'increment'                                                        => 'erhöhen',
   'input'                                                            => 'Eingang',
   'INPUT'                                                            => $C_INPUT_de,
   'INPUT_NAME'                                                       => $C_Input_names_de,
   'input_variable'                                                   => [ 'Eingangsvariable' , 'Eingangsvariablen' ],
   'illegal_name'                                                     => 'unzulässiger Name',
   'index'                                                            => 'Index',
   'info_1'                                                           => $C_Info_1_de,
   'insert'                                                           => 'einfügen',
   'insert_a_9'                                                       => '9_einfügen',
   'inserted_input_statement'                                         => 'Eingang-Anweisung eingefügt',
   'inserted_output_statement'                                        => 'Ausgang-Anweisung eingefügt',
   'is'                                                               => 'ist',
   'is_not_expected'                                                  => 'ist nicht erwartet',
   'is_a_reserved_word'                                               => 'ist ein reserviertes Wort',
   'larger'                                                           => 'grösser',
   'last_solution'                                                    => 'letzte Lösung',
   'last_solution_window'                                             => 'Fenster mit letzter Lösung',
   'license'                                                          => [ 'Lizenz', 'Lizenzen' ],
   'line'                                                             => [ 'Linie', 'Linien' ],
   'list_of_reserved_word'                                            => 'Liste der reservierten Worte',
   'license_of_LogicBuddy'                                            => [ 'Lizenz von LogicBuddy', 'Lizenzen von LogicBuddy' ],
   'logical_function_changed'                                         => 'logische Funktion wurde geändert',
   'logical_value'                                                    => [ 'logischer Wert', 'logische Werte' ],
   'logical_value_0_and_1'                                            => 'logischer Wert 0 und 1',
   'logical_value_doesnt_care'                                        => 'logischer Wert ist egal',
   'logical_value_not_changed'                                        => 'logischer Wert wird nicht geändert',
   'LogicBuddy_does'                                                  => $C_LogicBuddy_does_de,
   'LogicBuddy_license'                                               => $C_LogicBuddy_license_de,
   'main_control_buttons'                                             => 'Buttons der Steuerleiste',
   'mark'                                                             => [ 'Marke', 'Marken' ],
   'match_between_logical_function_and_table'                         => 'Übereinstimmung zwischen logischer Funktion und Tabelle',
   'mathematic'                                                       => 'Mathematik',
   'move_9_to_8'                                                      => 'verschiebe_9_nach_8',
   'message'                                                          => [ 'Nachricht', 'Nachrichten' ],
   'minimizes_logical_functions'                                      => 'minimiert logische Funktionen',
   'multiple_change_lower_row'                                        => $C_multiple_change_lower_row_de,
   'multiple_change_short'                                            => $C_multiple_change_short_de,
   'multiple_change_toggle'                                           => $C_multiple_change_toggle_de,
   'multiple_input_truth_table'                                       => $C_multiple_input_truth_table_de,
   'multiple_function_truth_table'                                    => $C_multiple_function_truth_table_de,
   'multiple_change_upper_row'                                        => $C_multiple_change_upper_row_de,
   'NAME'                                                             => $C_Names_de,
   'name_already_in_use'                                              => 'Name wird bereits verwendet',
   'name_too_long'                                                    => 'Name zu lang',
   'name_9_is_too_long_8'                                             => 'Name_9_ist zu lang_8',
   'NAND_NOT'                                                         => $C_NAND_NOT_de,
   'negative',                                                        => 'negativ',
   'new_9'                                                            => 'neu:_9',
   'new'                                                              => 'neu',
   'new_column_left'                                                  => 'neue Spalte links',
   'new_column_right'                                                 => 'neue Spalte rechts',
   'new_row_above'                                                    => 'neue Zeile oben',
   'new_row_below'                                                    => 'neue Zeile unten',
   'new_type'                                                         => 'neuer Typ',
   'next'                                                             => 'nächste',
   'no_change'                                                        => 'keine Änderung',
   'no_logical_function_defined'                                      => 'keine logische Funktion definiert',
   'non_alphanumerical_character'                                     => 'nicht alphanumerisches Zeichen',
   'no_input_defined_and_used_at_all'                                 => 'Überhaupt kein Eingang definiert und verwendet',
   'no_output_defined_and_used_at_all'                                => 'Überhaupt kein Ausgang definiert und verwendet',
   'note',                                                            => 'Hinweis',
   'nothing'                                                          => 'nichts',
   'not_found'                                                        => 'nicht gefunden',
   'not_ok',                                                          => 'Nicht-Ok',
   'number'                                                           => 'Anzahl',
   'number_of_generated_terms'                                        => 'Zahl der generierten Terme',
   'number_of_input_names'                                            => 'Anzahl der Namen für Eingangs-Variablen',
   'number_of_output_names'                                           => 'Anzahl der Namen für Ausgangs-Funktionen',
   'number_of'                                                        => 'Anzahl der',
   'ok'                                                               => 'ok',
   'one_less'                                                         => 'um eins kleiner',
   'one_more'                                                         => 'um eins höher',
   'on_Windows'                                                       => 'unter Windows',
   'or'                                                               => 'oder',
   'or_use_the_solution_dialog'                                       => 'oder den Lösungs-Dialog verwenden',
   'OUTPUT'                                                           => $C_OUTPUT_de,
   'output'                                                           => 'Ausgang',
   'output_name'                                                      => 'Name für Ausgangs-Funktion',
   'Output_or_ASSIGNMENT'                                             => $C_Output_or_ASSIGNMENT_de,
   'OUTPUT_NAME'                                                      => $C_Function_names_de,
   'output_9_used_twice'                                              => 'Name für Ausgang_9_zweifach verwendet',
   'output_variable'                                                  => [ 'Ausgangsvariable' , 'Ausgangsvariablen' ],
   'paste'                                                            => 'einfügen',
   'positive',                                                        => 'Positiv',
   'preformatted'                                                     => 'vorformatierter Paragraph',
   'print'                                                            => 'drucken',
   'question',                                                        => 'Frage',
   'quit_and_discard_changes'                                         => 'beenden und Änderungen verwerfen',
   'quit'                                                             => 'beenden',
   'quit_dialog',                                                     => 'Dialog abbrechen',
   'reject',                                                          => 'Abgelehnt',
   'release'                                                          => 'Freigabe',
   'reload'                                                           => 'neu laden',
   'remember'                                                         => 'merken',
   'rename'                                                           => 'umbenennen',
   'rename_function'                                                  => 'Funktion umbenennen',
   'replace'                                                          => 'ersetzen',
   'replace_by'                                                       => 'ersetzen durch',
   'reserved_word'                                                    => [ 'reserviertes Wort', 'reservierte Worte' ],
   'restore_backward'                                                 => 'rückwärts wiederherstellen',
   'restore_forward'                                                  => 'vorwärts wiederherstellen',
   'restore_logical_function_backwards'                               => 'logische Funktion rückwärts wiederherstellen',
   'restore_logical_function_forwards'                                => 'logische Funktion vorwärts wiederherstellen',
   'row'                                                              => 'Zeile',
   'row_number_of_table'                                              => 'Zeile der Tabelle',
   'rule_of_logical_function'                                         => ['Regel einer logischen Funktion', 'Regeln logischer Funktionen'],
   'save_and_exit'                                                    => 'sichern und beenden',
   'save_settings'                                                    => 'Einstellungen speichern',
   'save'                                                             => 'sichern',
   'save_changes_and_close_dialog'                                    => 'Änderungen sichern und Dialog beenden',
   'save_table_and_leave_table_editor'                                => 'Tabelle sichern und Tabellen-Editor verlassen',
   'search'                                                           => 'suchen',
   'select_all'                                                       => 'alles auswählen',
   'select_font'                                                      => 'Schriftart auswählen',
   'select_language'                                                  => 'Sprache auswählen, select language',
   'select_the_language'                                              => 'die Sprache auswählen',
   'select_logic'                                                     => 'Logik auswählen',
   'select_size'                                                      => 'Größe auswählen',
   'selected_language'                                                => 'ausgewählte Sprache',
   'setting'                                                          => [ 'Einstellung', 'Einstellungen' ],
   'settings_menu'                                                    => 'Konfigurations-Menü',
   'show_help_dialog'                                                 => 'Hilfe-Dialog anzeigen',
   'show_message'                                                     => [ 'Nachricht anzeigen', 'Nachrichten anzeigen' ],
   'show_solution_dialog'                                             => [ 'Dialog mit Lösung anzeigen', 'Dialog mit Lösungen anzeigen' ],
   'small'                                                            => 'klein',
   'solve_logical_function'                                           => 'logische Funktion lösen',
   'solution'                                                         => [ 'Lösung', 'Lösungen' ],
   'sort_down'                                                        => 'sortiere abwärts',
   'sort_up'                                                          => 'sortiere aufwärts',
   'special_key'                                                      => [ 'spezielle Taste', 'spezielle Tasten' ],
   'starting_with_9'                                                  => 'beginnend_mit_9',
   'strawberry_perl'                                                  => $C_Strawberry_de,
   'state'                                                            => 'Status',
   'sub_window'                                                       => 'Unterfenster im Dialog',
   'syntax_highlighting'                                              => 'Syntax-Hervorhebung',
   'syntax_of_logical_function'                                       => [ 'Syntax der logischen Funktion', 'Syntax logischer Funktionen' ],
   'table'                                                            => [ 'Tabelle', 'Tabellen' ],
   'table_changed'                                                    => 'Tabelle wurde geändert',
   'table_does_not_match_logical_function'                            => 'Tabelle entspricht nicht logischer Funktion',
   'table_does_not_match_logical_function_and_error_message'          => 'Tabelle und logische Funktion stimmen nicht überein und Fehlermeldung',
   'table_matches_logical_function'                                   => 'Tabelle entspricht logischer Funktion',
   'table_matches_logical_function_but_warning'                       => 'Übereinstimmung zwischen logischer Funktion und Tabelle aber mit Warnung',
   'the_error_messages_are_also_contained_in_the_message_window'      => 'Die Fehlermeldungen sind auch im Nachrichtenfenster enthalten',
   'take_notice_of_9'                                                 => '9_beachten',
   'text'                                                             => 'Text',
   'the_creation_of_the_table_lasts'                                  => 'die Erstellung der Tabelle dauert',
   'the_packaged_version_of_LogicBuddy_includes_9'                    => 'Die paketierte Version von LogicBuddy enthält_9',
   'the_content_of_the_dialog_is_retained'                            => [ 'der Inhalt des Dialogs bleibt erhalten',
                                                                           'die Inhalte des Dialogs bleiben erhalten'
                                                                         ],
   'the_solution_for_9_is'                                            => 'die Lösung_für_9_ist',
   'these_are'                                                        => 'Es sind',
   'there_are_9_solutions_for_8',                                     => 'es gibt_9_Lösungen_für_8',
   'there_are_various_types_representing_the_result_function'         => 'es gibt mehrere Weisen, die Ergebnisfunktionen darzustellen',
   'this_is_free_software_licensed_under'                             => 'es ist freie Software lizenziert unter',
   'TERM'                                                             => $C_TERM_de,
   'this_help'                                                        => 'diese Hilfe',
   'this_is_not_a_number'                                             => 'das ist keine Zahl',
   'top'                                                              => 'Anfang',
   'to_contents'                                                      => 'zum Inhaltsverzeichnis',
   'true'                                                             => 'wahr',
   'truth_table'                                                      => 'Wahrheitstabelle',
   'type'                                                             => 'Typ',
   'undo_last_table_changes'                                          => 'letzte Änderung an der Tabelle zurücknehmen',
   'undo'                                                             => 'rückgängig',
   'unexpected'                                                       => 'nicht erwartet',
   'unknown_input_name_9'                                             => 'unbekannter Eingangsname_9',
   'unknown_name_9'                                                   => 'unbekannter Name_9',
   'unknown_output_name_9'                                            => 'unbekannter Ausgangsname_9',
   'unselect_all'                                                     => 'nichts auswählen',
   'up',                                                              => 'aufwärts',
   'usually_a_9_is_missing'                                           => 'meistes fehlt ein_9',
   'use_NAND_and_NOT'                                                 => 'NAND und NOT verwenden',
   'use_AND_9_OR_and'                                                 => 'AND_9_OR_und',
   'using_output_9_in_its_own_assignment'                             => 'die Ausgangsvariable_9_wird in ihrer eigenen Zuweisung verwendet',
   'using_reserved_word_of_language_9'                                => 'reserviertes Wort der Sprache_9_verwendet',
   'using_values_of_current_row'                                      => 'unter Verwendung der Werte der aktuellen Reihe',
   'variable'                                                         => [ 'Variable', 'Variablen' ],
   'variable_9_previously_set_to_8'                                   => 'Variable_9_vorher_auf_8_gesetzt',
   'variable_9_never_used'                                            => 'Variable_9_wurde_nie_verwendet',
   'view'                                                             => 'anzeigen',
   'view_message'                                                     => [ 'Nachricht anzeigen', 'Nachrichten anzeigen' ],
   'view_Messages'                                                    => 'Fehler- und Systemmeldungen anzeigen',
   'view_Solution_Dialog'                                             => 'Die aktuelle Lösung in einem eigenen Fenster anzeigen',
   'when'                                                             => 'wenn',
   'when_or_dot'                                                      => "'wenn' oder '.'",
   'windows_note'                                                     => $C_Windows_note_de,
   'work',                                                            => 'Arbeit',
   'working',                                                         => 'arbeite',
   'warning'                                                          => 'Warnung',
   'wait_please'                                                      => 'bitte warten',
   'yes_change'                                                       => 'ja, ändern',
   'yes_create'                                                       => 'ja, erzeugen',
   'yes_delete'                                                       => 'ja, löschen',
);

Readonly my %C_german_plural => (
                                );

Readonly my %C_english_message  => (
   'and_TERM_or_dot'                                                  => $C_and_TERM_or_dot,
   'AND_OR'                                                           => $C_AND_OR,
   'ANY_CONDITION'                                                    => $C_ANY_CONDITION,
   'ASSIGNMENT'                                                       => $C_ASSIGNMENT,
   'AUTOMATIC_ERROR_CORRECTION'                                       => $C_AUTOMATIC_ERROR_CORRECTION,
   'balloon_Help'                                                     => $C_Balloon,
   'balloon_Help_Full'                                                => $C_Balloon_help,
   'Bracket_or_NAME'                                                  => $C_Bracket_or_NAME,
   'Check_Function'                                                   => $C_Check_Function,
   'check_and_calculate_logical_function'                             => $C_Check_Function_and_calculate,
   'copy_and_paste'                                                   => $C_Copy_and_Paste,
   'CONDITION'                                                        => $C_CONDITION,
   'Copyright_HK'                                                     => $C_Copyright_HK,
   'current_language'                                                 => 'English',
   'Describes_Function_Matches'                                       => $C_Description_Function_and_Table_match,
   'description_help'                                                 => $C_Description_help,
   'description_multiple_change'                                      => $C_description_multiple_change,
   'description_solution_window'                                      => $C_Description_Solution,
   'description_truth_table'                                          => $C_Description_truth_table,
   'description_truth_table_1'                                        => $C_Description_truth_table_1,
   'description_truth_table_2'                                        => $C_Description_truth_table_2,
   'description_truth_table_3'                                        => $C_Description_truth_table_3,
   'description_truth_table_4'                                        => $C_Description_truth_table_4,
   'dot_or_FACTOR'                                                    => "'.' or 'or' or $C_FACTOR",
   'Editor_Function_Description'                                      => $C_Description_Editor_Function,
   'equal_or_is'                                                      => "'=' or 'is'",
   'equal_or_is_or_are'                                               => "'=' or 'is' or 'are'",
   'espresso_software'                                                => $C_Espresso,
   'example_of_logical_function'                                      => ['example of logical function', 'examples of logical functions'],
   'FACTOR'                                                           => $C_FACTOR,
   'EXPRESSION'                                                       => $C_EXPRESSION,
   'file_9_not_found_create'                                          => 'file_9_not_found. Create?',
   'FUNCTION_CALL_or_Value'                                           => $C_FUNCTION_CALL_or_Value,
   'Help_Description_for_Balloon'                                     => "Help dialog\nanywhere with <F1> or\nreight mouse key: selective help",
   'info_1'                                                           => $C_Info_1,
   'INPUT'                                                            => $C_INPUT,
   'INPUT_NAME'                                                       => $C_Input_names, 
   'language_9'                                                       => 'language =_9',
   'license_of_LogicBuddy'                                            => [ 'license of LogicBuddy', 'licenses of LogicBuddy' ],
   'logical_value_doesnt_care'                                        => "logical value doesn't care",
   'LogicBuddy_does'                                                  => $C_LogicBuddy_does,
   'LogicBuddy_license'                                               => $C_LogicBuddy_license,
   'multiple_change_lower_row'                                        => $C_multiple_change_lower_row,
   'multiple_change_short'                                            => $C_multiple_change_short,
   'multiple_change_toggle'                                           => $C_multiple_change_toggle,
   'multiple_input_truth_table'                                       => $C_multiple_input_truth_table,
   'multiple_function_truth_table'                                    => $C_multiple_function_truth_table,
   'multiple_change_upper_row'                                        => $C_multiple_change_upper_row,
   'NAME',                                                            => $C_Names,
   'NAND_NOT'                                                         => $C_NAND_NOT,
   'OUTPUT'                                                           => $C_OUTPUT,
   'Output_or_ASSIGNMENT'                                             => $C_Output_or_ASSIGNMENT,
   'OUTPUT_NAME'                                                      => $C_Function_names,
   'reject_change'                                                    => [ 'Änderung verwerfen', 'Änderungen verwerfen' ],
   'rule_of_logical_function'                                         => ['rule of logical function', 'rules of logical functions'],
   'select_language'                                                  => 'select language, Sprache auswählen',
   'show_solution_dialog'                                             => [ 'show solution dialog', 'show solutions dialog' ],
   'strawberry_perl'                                                  => $C_Strawberry,
   'TERM'                                                             => $C_TERM,
   'the_content_of_the_dialog_is_retained'                            => [ 'the content of the dialog is retained',
                                                                           'the contents of the dialog are retained'
                                                                         ],
   'this_is_free_software_licensed_under'                             => 'this is free software, licensed under',
   'when_or_dot'                                                      => "'when' or '.'",
   'windows_note'                                                     => $C_Windows_note,
   'undo_last_changes'                                                => 'Undo last changes <Ctrl>+<z>',
);

################################################################################
# global variables
################################################################################

our $G_plural;
our $G_message;

my %G_german_message   = %C_german_message;
my %G_english_message  = %C_english_message;
my %G_do_not_transform = ();

################################################################################
# internal subprograms
################################################################################

sub __english_plural_of {
   my ( $word ) = @_;

   return EMPTY unless defined $word;
   return $word if $LogicBuddy::G_Debug > 1 ;
   return $word if $word !~ m{ [a-z] \z }xms;
   return $word . 's';

}; # __english_plural_of

sub __german_plural_of {
   my ( $word ) = @_;

   return EMPTY unless defined $word;

#  to be implemented?
#  plurals are defined in %G_german_message like message => [ singular_message, plural_message ]
#  See PluralOf and SingularOf
   return $word;

#  return $word if $LogicBuddy::G_Debug > 1 ;
#  return $C_german_plural{$word} if exists $C_german_plural{$word};
#  return $word unless $word =~ m{ \s \w \z }xms;
#  return $word;

}; # __german_plural_of

sub __singular_and_plural_of {
   my ( $word ) = @_;
   return ( EMPTY, EMPTY ) unless defined $word;
   return ( $word, PluralOf($word) );
}; # __singular_and_plural_of

sub __insert_parameter {
   my ( $message, $parameter_1, $parameter_2 ) = @_;

   # $parameter_1 substitutes 9
   # $parameter_2 substitutes 8

   if ( $LogicBuddy::G_Debug > 1 ) {
      if ( defined $parameter_1 and $parameter_1 ne EMPTY ) {
         if ( defined $parameter_2 and $parameter_2 ne EMPTY ) {
            return "Debug: $message( '$parameter_1', '$parameter_2' )";
         } else {
            return "Debug: $message( '$parameter_1' )";
         };
      } else {
         return $message;
      };
   #         not   a text in $parameter_1 is not (not $parameter_1) because '0' is a text
   } elsif ( not ( defined $parameter_1 and $parameter_1 ne EMPTY ) ) {
      $message =~ s{ \A [89]_ }{_}xms;
      $message =~ s{ _[89] \z }{_}xms;
      $message =~ s{ _[89]_ }{_}gxms;
      $message =~ s{_+}{ }gmxs;
      return $message;
   } elsif ( $message =~ s{ _9 \z}{_}xms ) {
      if ( defined $parameter_2 and $parameter_2 ne EMPTY ) {
         if ( $message =~ s{_8 (_)}{_$parameter_2$1}xms ) {
            $parameter_2 = EMPTY;
         } elsif ( $message =~ s{ \A 8 (_)}{_$parameter_2$1}xms ) {
            $parameter_2 = EMPTY;
         };
      } else {
         $parameter_2 = EMPTY;
      };
      $message =~ s{_+}{ }gmxs;
      return "$message$parameter_1$parameter_2 ";
   } elsif ( $message =~ s{ \A 9_}{_}xms ) {
      if ( defined $parameter_2 and $parameter_2 ne EMPTY ) {
         if ( $message =~ s{ _8 \z }{_$parameter_2}xms ) {
            $parameter_2 = EMPTY;
         } elsif ( $message =~ s{ _8 (_)}{_$parameter_2$1}xms ) {
            $parameter_2 = EMPTY;
         };
      } else {
         $parameter_2 = EMPTY;
      };
      $message =~ s{_+}{ }gmxs;
      return "$parameter_1$message$parameter_2";
   } elsif ( $message =~ m/_[89]/ ) {
      if ( defined $parameter_2 and $parameter_2 ne EMPTY ) {
         if ( $message =~ s{ \A 8 (_) }{$parameter_2$1}xms ) {
            $parameter_2 = EMPTY;
         } elsif ( $message =~ s{ _8 (_) }{_$parameter_2$1}xms ) {
            $parameter_2 = EMPTY;
         } elsif ( $message =~ s{ _8 \z }{_$parameter_2}xms ) {
            $parameter_2 = EMPTY;
         };
      };
      if ( defined $parameter_1 ) {
         if ( $message =~ s{ _9 (_) }{_$parameter_1$1}xms ){
            $parameter_1 = EMPTY;
         };
      } else {
         $parameter_1 = EMPTY;
      };
      $message =~ s{_+}{ }gmxs;
      return "$message$parameter_1$parameter_2";
   } elsif ( defined $parameter_1 ) {
      $message .= $parameter_1;
      $message .= $parameter_2 if defined $parameter_2;;
      $message =~ s{_+}{ }gmxs;
      return $message;
   };

}; # __insert_parameter

sub __ucfirst {
   my ( $message ) = @_;

   return $message if $LogicBuddy::G_Debug > 1 ;

   if ( ref $message  eq 'ARRAY' ) {
      return [ map { ucfirst $_ } @$message ];
   } else {
      return ucfirst $message;
   };

}; # __ucfirst

sub __get {
   my ( $token ) = @_;

   our $G_message;

   return $token if $LogicBuddy::G_Debug > 1 ;

   if ( exists $G_message->{$token} ) {
      return $G_message->{$token};
   } elsif ( exists $G_message->{lcfirst $token} ) {
      return __ucfirst( $G_message->{lcfirst $token} );
   } elsif ( exists $G_english_message{$token} ) {
      return $G_english_message{$token};
   } elsif ( exists $G_english_message{lcfirst $token} ) {
      return __ucfirst( $G_english_message{lcfirst $token} );
   } else {
      return $token;
   };

}; # __get

################################################################################
# exported subprograms
################################################################################

sub PluralOf {
   my ( $word ) = @_;

   our $G_plural;

   return EMPTY unless defined $word;
   return $word->[1] if ref $word eq 'ARRAY';
   return $word unless $G_plural;
   return &$G_plural( $word );

}; # PluralOf

sub SingularOf {
   my ( $word ) = @_;

   return EMPTY unless defined $word;
   return $word->[0] if ref $word eq 'ARRAY';
   return $word;

}; # SingularOf

sub TranslationOf {
   my ( $token, $text_1, $text_2 ) = @_;

   # $text_1 optional
   # $text_2 optional

   my ( $singular_1, $plural_1 ) = __singular_and_plural_of( $text_1 );
   my ( $singular_2, $plural_2 ) = __singular_and_plural_of( $text_2 );

   my $message = __get( $token );

   if ( exists $G_do_not_transform{$token} ) {
      return [ $message, $message ];
   };

   if ( not ref $message ) {
      $message = [ $message, PluralOf($message) ];
   };

   return [ __insert_parameter($message->[0], $singular_1, $singular_2), __insert_parameter($message->[1], $plural_1, $plural_2) ];

}; # TranslationOf

sub TestLocale {
   return ( \%G_english_message, \%G_german_message );
}; # TestLocale

sub SetLanguage {
   my ( $language ) = @_;

   our $G_plural;
   our $G_message;

   $language = $LogicBuddy::G_language if not $language;

   if ( $language and $language eq 'German' ) {
      $G_message = \%G_german_message;
      $G_plural  = \&__german_plural_of;
   } else {
      $G_message = undef;  # forces default English
      $G_plural  = \&__english_plural_of;
   };

   return;

}; # SetLanguage

sub Set_local_of {
   my ( $locale, $do_not_transform, %language_values ) = @_;

   foreach my $language ( keys %language_values ) {
      if ( $language eq 'English' ) {
         Carp::confess( "Error in Set_local_of($locale): no contents for language '$language'" ) if not $language_values{$language};
         $G_english_message{$locale} = $language_values{$language};
      } elsif ( $language eq 'German' ) {
         Carp::confess( "Error in Set_local_of($locale): no contents for language '$language'" ) if not $language_values{$language};
         $G_german_message{$locale} = $language_values{$language};
      } else {
         Carp::confess( "Error in Set_local_of($locale): unknown language '$language'" );
      };
   };

   $G_do_not_transform{$locale} = TRUE if $do_not_transform;
        
   return;

}; # Set_local_of

################################################################################
# AUTOLOAD
################################################################################

sub AUTOLOAD {
   my ( @arguments ) = @_;

   my $program =  $LogicBuddyn::Locale::AUTOLOAD;
   $program    =~ s{.*::}{};

   return TranslationOf( $program, @arguments );

}; # AUTOLOAD

################################################################################
# BEGIN END
################################################################################

BEGIN {
   SetLanguage();
}; # BEGIN

################################################################################
# end LogicBuddy::Locale
################################################################################

1;
