=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische_Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Logic - module

=head1 SYNOPSIS

   use LogicBuddy::Logic;

   $logic = LogicBuddy::Logic->new( $viewer );

   $logic->define_expression( $expression );

   $expression_string              = $logic->expression_string();
   $expression_ref                 = $logic->expression();
   @input_names                    = $logic->input_names();
   @output_names                   = $logic->output_names();
   ($message, $line, $column )     = $logic->last_error();
   ( $has_code, $error, $warning ) = $logic->status();

   if ( $logic->is_defined() ) ...

   $pattern                        = LogicBuddy::Logic::reserved_word_pattern();
   $pattern                        = LogicBuddy::Logic::reserved_word_pattern_no_case();
   $pattern                        = LogicBuddy::Logic::input_variable_pattern();
   $pattern                        = LogicBuddy::Logic::output_variable_pattern();
   $pattern                        = LogicBuddy::Logic::comment_pattern();

   @reserved_words                = LogicBuddy::Logic::reserved_words();

=head1 ABSTRACT

LogicBuddy::Logic abstracts interpreting and calculation of logical expressions.

=head1 DESCRIPTION

=head2 METHODS

=head3 new

    $logic = LogicBuddy::Logic->new( $viewer );

Instantiates an logic processor using $viewer for output.

=head3 define_expression

   $logic->define_expression( $expression );

Defines a set of new logic functions from $expression string.

Raises the exception $LogicBuddy::Logic::C_expression_exception when
any error is detected in the $expression.
An error message is generated using the associated viewer.
Furthermore warnings may be generates on the viewer.
If an error occurred no expression is defined.

Raised expressions:

$LogicBuddy::Logic::C_expression_exception

=head3 expression_string

   $expression_string = $logic->expression_string();

Returns the expression string of the last call of define_expression.

If define_expression was not called yet or
the last call raised an error, undef is returned.

Raised expressions: none

=head3 expression

   $expression_ref = $logic->expression();

Returns a reference to the subprogram representing the expressions defined by the
last call of define_expression.

Raises the exception $LogicBuddy::Logic::C_no_expression_defined_exception when
define_expression was never called or the last call raised an error.

Raised expressions:

$LogicBuddy::Logic::C_no_expression_defined_exception

=head3 input_names

   @input_names = $logic->input_names();

Returns the list of input names defined by the last call of define_expression.

Raises the exception $LogicBuddy::Logic::C_no_expression_defined_exception when
define_expression was never called or the last call raised an error.

Raised expressions:

$LogicBuddy::Logic::C_no_expression_defined_exception

=head3 output_names

   @output_names = $logic->output_names();

Returns the list of output names defined by the last call of define_expression.

Raises the exception $LogicBuddy::Logic::C_no_expression_defined_exception when
define_expression was never called or the last call raised an error.

Raised expressions:

$LogicBuddy::Logic::C_no_expression_defined_exception

=head3 last_error

   ($message, $line, $column ) = $logic->last_error();

Returns the last error $message detected at $line and $column.
If there was no error at the last call of define_expression $message, $line and $column are undef.

Raised expressions: none

=head3 status

   ( $has_code, $error, $warning ) = $logic->status();

Returns if $logic has_code, and if the where an $error or $warning.

Raised expressions: none

=head3 is_defined

   if ( $logic->is_defined() ) ...

Returns if there is an expression defined, i.e. the last call of define_expression
was successful.

Raised expressions: none

=head2 CLASS METHODS

=head3 reserved_word_pattern input_variable_pattern output_variable_pattern comment_pattern

   $pattern = LogicBuddy::Logic::reserved_word_pattern();
   $pattern = LogicBuddy::Logic::input_variable_pattern();
   $pattern = LogicBuddy::Logic::output_variable_pattern();
   $pattern = LogicBuddy::Logic::comment_pattern();

Returns a $pattern finding reserved words or input or output variable or comment.

Raised expressions: none

=head3 reserved_word_pattern_no_case

   $pattern = LogicBuddy::Logic::reserved_word_pattern_no_case();

Returns a $pattern finding reserved words not case sensitive.

Raised expressions: none

=head3 reserved_word

   @reserved_words = LogicBuddy::Logic::reserved_words();

Returns a list of all reserved words.

Raised expressions: none

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische_Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;
use warnings;

use Carp;

use LogicBuddy::Utilities;
use LogicBuddy::Logic::Text;
use LogicBuddy::Logic::IO;

use LogicBuddy::Singular;

package LogicBuddy::Logic;

use English '-no_match_vars';

################################################################################
# constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

use Readonly;

Readonly my $C_Margin => '   ';

Readonly $LogicBuddy::Logic::C_no_expression_defined_exception => 'LogicBuddy::Logic no expression defined';
Readonly $LogicBuddy::Logic::C_expression_exception            => 'LogicBuddy::Logic error in expression';

Readonly my @C_reserved_words_de => qw( und
                                        oder
                                        nicht
                                        Eingang
                                        Ausgang
                                        wenn
                                        ist
                                        sind
                                        falsch
                                        wahr
                                        beliebig
                                      );

Readonly my @C_reserved_words_en => qw( and
                                        or
                                        not
                                        Input
                                        Output
                                        when
                                        is
                                        are
                                        false
                                        true
                                        any
                                      );

Readonly my @C_reserved_math => qw{ and
                                    or
                                    not
                                  };

Readonly my @C_reserved_always => qw( AND
                                      OR
                                      NOT
                                      NAND
                                      NOR
                                    );

Readonly my $C_reserved_special => "[,.()=^]";

Readonly my @C_reserved_words   => ( @C_reserved_always, @C_reserved_words_en, @C_reserved_words_de );

Readonly my %C_reserved         => map { (    $_ => TRUE ) } map { lc $_ } @C_reserved_words;
Readonly my %C_reserved_en      => map { (    $_ => TRUE ) } map { lc $_ } @C_reserved_words_en;
Readonly my %C_reserved_de      => map { (    $_ => TRUE ) } map { lc $_ } @C_reserved_words_de;
Readonly my %C_reserved_math    => map { (    $_ => TRUE ) } map { $_ } @C_reserved_math;
Readonly my %C_reserved_always  => map { (    $_ => TRUE ) } @C_reserved_always;
Readonly my %C_correct_reserved => map { (    $_ => TRUE ) } @C_reserved_words;

Readonly my $C_reserved_words_description => "   " . (join "\n   ", (sort @C_reserved_words) ) . "\n";

Readonly my $C_simple_name_pattern => qr{ ( [a-z] ( _? [a-z0-9] )* ) }ixms;
Readonly my %C_lc_reserved         => map { $_ => TRUE } map { lc $_ } @C_reserved_words;
Readonly my $C_lc_alternatives     => join ' | ', (keys %C_lc_reserved);
Readonly my $C_name_pattern        => qr{ \A (?! (?: $C_lc_alternatives ) (?: \z | \W ) ) $C_simple_name_pattern }ixms;
Readonly my $C_reserved_pattern    => qr{ \A ( $C_lc_alternatives (?= \z | \W ) | $C_reserved_special ) }ixms;

################################################################################
# global Variable
################################################################################

my $G_debug = FALSE;

################################################################################
# class LogicBuddy::Logic
################################################################################

sub new {
   my ( $proto, $editor )= @_;

   my $class = ref($proto) || $proto;
   my $self;

   $self->{Editor}                = $editor;
   $self->{input_missing}         = FALSE;
   $self->{output_missing}        = TRUE;
   $self->{input_name}            = {};
   $self->{input_names}           = [];
   $self->{variable_name}         = {};
   $self->{current_variable_name} = EMPTY;
   $self->{output_names}          = [];
   $self->{code}                  = EMPTY;
   $self->{language_seen}         = FALSE;
   $self->{any_seen}              = FALSE;
   $self->{equal_seen}            = FALSE;
   $self->{command}               = undef;

   $self->{IO}                    = LogicBuddy::Logic::IO->new( $editor, $C_reserved_pattern, $C_name_pattern, $C_name_pattern );

   bless $self, $class;

   return $self;

}; # new

################################################################################
# internal methods
################################################################################

sub I_see_not_reserved_word {
   my ( $self, $name ) = @_;

   if ( exists $C_reserved{ lc $name } ) {
      my $note = EMPTY;
      if ( not exists $C_correct_reserved{ $name } ) {
         $note = "\n" . LogicBuddy::Singular::But_9_is_not_accepted( "'$name'" );
      };
      $self->{IO}->insert_warning( LogicBuddy::Singular::Warning( ":\n" ),
                                   "'" . (lc $name) . "' " . LogicBuddy::Singular::is_a_reserved_word() . "$note.",
                                   length $name
                                 );
   };

   return TRUE;

}; # I_see_not_reserved_word

sub IC_check_language {
    my ( $self, $accepted ) = @_;

    my $language;

    eval{ $language = $self->{Editor}->language() };

    return if not $language;
    return if not $accepted;
    return if exists $C_reserved_always{$accepted};
    return if $self->{language_seen};

    if ( $language eq 'German' ) {
       if ( $self->{equal_seen} and exists $C_reserved_math{$accepted} ) {
          return;
       };
       if ( exists $C_reserved_en{ lc $accepted } )  {
          $self->{IO}->insert_warning( LogicBuddy::Singular::Warning( ":\n" ),
                                       LogicBuddy::Singular::Using_reserved_word_of_language_9( LogicBuddy::Singular::English(), ".\n" ),
                                       (length $accepted)
                                     );
          $self->{language_seen} = TRUE;
       };
    } elsif ( $language eq 'English' ) {
       if ( exists $C_reserved_de{ lc $accepted } )  {
          $self->{IO}->insert_warning( LogicBuddy::Singular::Warning( ":\n" ),
                                       LogicBuddy::Singular::Using_reserved_word_of_language_9( LogicBuddy::Singular::German(), ".\n" ),
                                       (length $accepted)
                                     );
          $self->{language_seen} = TRUE;
       };
    };

    return;

}; # IC_check_language

sub I_accept_and_check_language {
   my ( $self, $pattern, $term ) = @_;

   my $accepted =  $self->{IO}->accept( $pattern, $term, sub{ $self->IC_check_language( @_ ); } );
   $self->{equal_seen} = TRUE if $accepted eq '=';
   return $accepted;

}; # I_accept_and_check_language

sub I_am_expecting_and_check_language {
   my ( $self, $pattern, $term ) = @_;

   my $accepted = $self->{IO}->expecting( $pattern, $term, sub{ $self->IC_check_language( @_ ); } );
   $self->{equal_seen} = TRUE if $accepted eq '=';
   return $accepted;

}; # I_am_expecting_and_check_language

sub I_miss_input {
   my ( $self ) = @_;

   if ( scalar @{ $self->{input_names} } ) {
      my $insert = LogicBuddy::Singular::Input( ' = ' ) . (join ', ', @{ $self->{input_names} } ) . ".\n";
      $self->{Editor}->insert_correction_at( '0.0', $insert );
      $self->{IO}->warning( LogicBuddy::Singular::Correction( ":\n" ), LogicBuddy::Singular::Inserted_input_statement( ".\n" ), 0, 0 );
   } else {
      $self->{IO}->insert_warning( LogicBuddy::Singular::Error( ":\n" ), LogicBuddy::Singular::No_input_defined_and_used_at_all( ".\n" ) );
   };
   $self->{IO}->set_error();

   return;

}; # I_miss_input

sub I_miss_output {
   my ( $self ) = @_;

   if ( scalar %{ $self->{variable_name} } ) {
      my $insert = LogicBuddy::Singular::Output( ' = ' ) . (join ', ', keys %{ $self->{variable_name} } ) . ".\n";
      $self->{Editor}->insert_at( 'end', "\n" );
      $self->{Editor}->insert_correction_at( 'end', $insert );
      $self->{IO}->warning( LogicBuddy::Singular::Correction( ":\n" ),
                            LogicBuddy::Singular::Inserted_output_statement( ".\n" ),
                            'end - 2 lines'
                          );
   } else {
      $self->{IO}->insert_warning( LogicBuddy::Singular::Error( ":\n" ), LogicBuddy::Singular::No_output_defined_and_used_at_all( ".\n" ) );
   };
   $self->{IO}->set_error();

   return;

}; # I_miss_output

################################################################################
# productions
################################################################################

sub LOGIC {
   my ( $self ) = @_;

   $self->INPUT();
   $self->ASSIGNMENTS( LogicBuddy::Singular::OUTPUT() );
   $self->OUTPUT();

   $self->{IO}->no_more_text();

   return;

}; # LOGIC

sub INPUT {
   my ( $self ) = @_;

   my $input = $self->I_accept_and_check_language( qr{ \A (Input|Eingang) (?! [a-zA-Z] ) }xms, [ 'Input' ] );

   if ( $input ) {
      $self->{code} .= "my ( ";

      $self->I_am_expecting_and_check_language( qr{ \A (=| (ist|is|are|sind) (?! [a-zA-Z] ) ) }xms, [ qw{ is are = } ] );
      $self->LIST_OF_INPUTS();
      $self->{IO}->expecting( qr{ \A ( [.] ) }xms, [ 'and', [ ',', '.' ] ] );

      $self->{code} .= " ) = \@_;\n\n$C_Margin";

      return TRUE;
   } else {
      $self->{IO}->insert_warning( LogicBuddy::Singular::Expected( ":\n" ), LogicBuddy::Singular::INPUT( "\n" ) ); 
      $self->{input_missing} = TRUE;
      return FALSE;
   };

}; # INPUT

sub LIST_OF_INPUTS {
   my ( $self ) = @_;

   my $name = $self->INPUT_NAME();
   $self->{IO}->expected( LogicBuddy::Singular::INPUT_NAME() ) if not $name;
   $self->{code} .= '$' . $name;

   while ( $self->I_accept_and_check_language( qr{ \A (,| (?: and|und ) (?! [a-zA-Z] ) ) }xms, [ 'and', [ ',', '.' ] ] ) ) {
      my $name = $self->INPUT_NAME();
      $self->{IO}->expected( LogicBuddy::Singular::INPUT_NAME() ) if not $name;
      $self->{code} .= ', $' . $name;
   };

   return TRUE;

}; # LIST_OF_INPUTS

sub INPUT_NAME {
   my ( $self ) = @_;

   my $name = $self->{IO}->accept( 'INPUT_NAME', LogicBuddy::Singular::INPUT_NAME() );
   return if not $name;

   $self->I_see_not_reserved_word( $name );

   if ( $LogicBuddy::C_max_name_length < length $name ) {
      $self->{IO}->insert_error( LogicBuddy::Singular::Error( ":\n" ),
                             LogicBuddy::Singular::Input_name_9_is_too_long( "'$name'", " (>$LogicBuddy::C_max_name_length)" ),
                             length $name
                           );
   };

   $self->{input_name}{$name} = 1;
   push @{ $self->{input_names} }, $name;
   return $name;

}; # INPUT_NAME

sub known_NAME {
   my ( $self ) = @_;

   my $name = $self->{IO}->accept( 'NAME', LogicBuddy::Singular::NAME() );

   return if not $name;

   $self->I_see_not_reserved_word( $name );

   if ( exists $self->{input_name}{$name} ) {
      $self->{input_name}{$name}++;
   } elsif ( $self->{current_variable_name} eq $name ) {
      $self->{IO}->insert_warning( LogicBuddy::Singular::Warning( ":\n" ),
                                   LogicBuddy::Singular::using_output_9_in_its_own_assignment( "'$name'", "\n" ),
                                   length $name
                                 );
      $self->{IO}->set_error();
      $self->{IO}->go_back( length $name );
      return;
   } elsif ( exists $self->{variable_name}{$name} ) {
      $self->{variable_name}{$name}++;
   } elsif ( $self->{input_missing} ) {
      if ( $LogicBuddy::C_max_name_length < length $name ) {
         $self->{IO}->insert_error( LogicBuddy::Singular::Error( ":\n" ),
                                    LogicBuddy::Singular::Input_name_9_is_too_long( "'$name'", " (>$LogicBuddy::C_max_name_length)" ),
                                    length $name
                                  );
      };
      $self->{input_name}{$name} = 2;
      push @{ $self->{input_names} }, $name;
   } else {
      $self->{IO}->insert_warning( LogicBuddy::Singular::Error( ":\n" ),
                                   LogicBuddy::Singular::Unknown_name_9( "'$name'" ),
                                   length $name
                                 );
      $self->{IO}->go_back( length $name );
      return;
   };

   return $name;

}; # known_NAME

sub OUTPUT {
   my ( $self ) = @_;

   if ( $self->{IO}->is_end_of_input() ) {
      $self->{IO}->insert_warning( LogicBuddy::Singular::Error( ":\n" ), LogicBuddy::Singular::Expected( ":\n" ) . LogicBuddy::Singular::OUTPUT() );
      $self->{IO}->set_error();
      return FALSE;
   };

   $self->I_am_expecting_and_check_language( qr{ \A (Output|Ausgang) (?! [a-zA-Z] ) }xms,    LogicBuddy::Singular::Output_or_ASSIGNMENT() );
   $self->{output_missing} = FALSE;
   $self->I_am_expecting_and_check_language( qr{ \A (=| (ist|is|are|sind) (?! [a-zA-Z] ) ) }xms, [ '=', 'is', 'are' ] );

   $self->{code} .= "\n${C_Margin}return( ";

   $self->LIST_OF_OUTPUTS();
   $self->{IO}->expecting( qr{ \A ( [.] ) }xms, [ 'and', [ ',', '.' ] ] );

   $self->{code} .= " );\n$C_Margin";

   return TRUE;

}; # OUTPUT

sub LIST_OF_OUTPUTS {
   my ( $self ) = @_;

   my $name =  $self->OUTPUT_NAME();
   $self->{IO}->expected( LogicBuddy::Singular::OUTPUT_NAME() ) if not $name;
   if ( $self->{variable_name}{$name} < 2 ) {
      $self->{IO}->insert_error( LogicBuddy::Singular::Error( ":\n" ),
                                 LogicBuddy::Singular::unknown_output_name_9( "'$name'" ),
                                 length $name
                               );
   };
   $self->{code} .= ' $' . $name;
   $self->{variable_name}{$name} = -1;

   push @{ $self->{output_names} }, $name;

   while ( $self->I_accept_and_check_language( qr{ \A (,| (?: and|und ) (?! [a-zA-Z] ) ) }xms, [ 'and', [ ',', '.' ] ] ) ) {
      $self->{code} .= ', ';
      $name = $self->OUTPUT_NAME();
      $self->{IO}->expected( LogicBuddy::Singular::OUTPUT_NAME() ) if not $name;
      if ( $self->{variable_name}{$name} == -1 ) {
         $self->{IO}->insert_error( LogicBuddy::Singular::Error( ":\n" ),
                                    LogicBuddy::Singular::output_9_used_twice( "'$name'" ),
                                    length $name
                                  );
      } elsif ( $self->{variable_name}{$name} < 2 ) {
         $self->{IO}->insert_error( LogicBuddy::Singular::Error( ":\n" ),
                                    LogicBuddy::Singular::unknown_output_name_9( "'$name'" ),
                                    length $name
                                  );
      };
      $self->{code} .= ' $' . $name;
      $self->{variable_name}{$name} = -1;

      push @{ $self->{output_names} }, $name;
   };

   return TRUE;

}; # LIST_OF_OUTPUTS

sub OUTPUT_NAME { 
   my ( $self, $symbol_or ) = @_;

   my $name = $self->{IO}->accept( 'OUTPUT_NAME', $symbol_or || LogicBuddy::Singular::OUTPUT_NAME() );
   return if not $name;

   $self->I_see_not_reserved_word( $name );

   if ( exists $self->{input_name}{$name} ) {
      $self->{IO}->insert_error( LogicBuddy::Singular::Error( ":\n" ),
                                 LogicBuddy::Singular::Is_an_input_name( 1 ),
                                 length $name
                               );
      return;
   };

   if ( $LogicBuddy::C_max_name_length < length $name ) {
      $self->{IO}->insert_error( LogicBuddy::Singular::Error( ":\n" ),
                                 LogicBuddy::Singular::Output_name_9_is_too_long( "'$name'", " (>$LogicBuddy::C_max_name_length)" ),
                                 length $name
                               );
   };

   $self->{variable_name}{$name}  = 1 if not exists $self->{variable_name}{$name};
   $self->{current_variable_name} = $name;

   return $name;

}; # OUTPUT_NAME

sub ASSIGNMENTS {
   my ( $self, $symbol_or ) = @_;

   $self->{IO}->expected( LogicBuddy::Singular::ASSIGNMENT() ) if not $self->ASSIGNMENT();

   while ( $self->{IO}->accept( qr{ \A ( [.] ) }xms, [ qw{ and or when }, [ ',' ] ] ) ) {
      $self->{code} .= ";\n$C_Margin";
      return FALSE if $self->{IO}->is_end_of_input();
      return FALSE if not $self->ASSIGNMENT( [ $symbol_or, LogicBuddy::Singular::ASSIGNMENT() ] );
   };

   $self->{IO}->expecting( qr{ \A ( [.] ) }xms, [ qw{ and or when }, [ ',', '.' ] ] );
   $self->{code} .= ";\n$C_Margin";

   return TRUE;

}; # ASSIGNMENTS

sub ASSIGNMENT {
   my ( $self, $symbol_or ) = @_;

   my $variable_name = $self->OUTPUT_NAME( $symbol_or );

   return if not $variable_name;

   $self->I_am_expecting_and_check_language( qr{ \A (= | (?: ist|is ) (?! [a-zA-Z] ) ) }xms, [ qw{ = is } ] );

   $self->{code} .= 'my ' if $self->{variable_name}{$variable_name} < 2;
   $self->{code} .= "\$$variable_name = ";

   if ( $self->ANY_CONDITION( $variable_name ) ) {
      $self->{variable_name}{$variable_name}++;
      return TRUE;
   } elsif ( $self->{any_seen} ) {
      $self->{IO}->expected( LogicBuddy::Singular::ANY_CONDITION() );
   } elsif ( $self->VALUE( $variable_name ) ) {
      $self->{variable_name}{$variable_name}++;
      return TRUE;
   };

   return FALSE;

}; # ASSIGNMENT

sub ANY_CONDITION {
   my ( $self, $variable_name ) = @_;

   if ( $self->I_accept_and_check_language( qr{ \A (any|beliebig) }xms, LogicBuddy::Singular::EXPRESSION_OR_ANY_CONDITION() ) ) {
      $self->{any_seen} = TRUE;
      $self->{code} .= "'d'"; # the doesn't care of the truth table
      $self->{IO}->expected( LogicBuddy::Singular::CONDITION() ) if not $self->CONDITION( $variable_name, [ 'when' ] );
      return TRUE;
   };

   return FALSE;

}; # ANY_CONDITION

sub CONDITION {
   my ( $self, $variable_name, $expecting ) = @_;

   if ( $self->I_accept_and_check_language( qr{ \A (when|wenn) (?! [a-zA-Z] ) }xms, $expecting ) ) {

      if ( $self->{variable_name}{$variable_name} < 2 ) {
         my $false = LogicBuddy::Singular::false();
         $self->{IO}->insert_warning( LogicBuddy::Singular::Note( ":\n" ),
                                      LogicBuddy::Singular::variable_9_previously_set_to_8( "'$variable_name'", "'$false'" ) . '.',
                                      4 # length when
                                    );
      };

      $self->{code} .= ' if ';
      $self->{IO}->expected( LogicBuddy::Singular::EXPRESSION() ) if not $self->EXPRESSION();
      return TRUE;
   };

   return FALSE;

}; # CONDITION

sub VALUE {
   my ( $self, $variable_name ) = @_;

   $self->{IO}->expected( LogicBuddy::Singular::EXPRESSION() ) if not $self->EXPRESSION();

   $self->CONDITION( $variable_name, [ 'when', [ '.' ] ] );

   return TRUE;

}; # VALUE

sub EXPRESSION {
   my ( $self ) = @_;

   if ( $self->I_accept_and_check_language( qr{ \A ( \^ | (?: NOT|not|nicht ) (?! [a-zA-Z] ) ) }xms, LogicBuddy::Singular::EXPRESSION() ) ) {
      $self->{code} .= "! ";
   };
   $self->{IO}->expected( LogicBuddy::Singular::TERM() ) if not $self->TERM();

   while ( $self->I_accept_and_check_language( qr{ \A (AND|and|und) (?! [a-zA-Z] ) }xms, LogicBuddy::Singular::and_TERM_or_dot() ) ) {
      $self->{code} .= " && ";
      if ( $self->I_accept_and_check_language( qr{ \A ( \^ | (?: NOT|not|nicht ) (?! [a-zA-Z] ) ) }xms, LogicBuddy::Singular::TERM() ) ) {
         $self->{code} .= "! ";
      };

      $self->{IO}->expected( LogicBuddy::Singular::TERM() ) if not $self->TERM();
   };

   return TRUE;

}; # EXPRESSION

sub TERM {
   my ( $self ) = @_;

   $self->{IO}->expected( LogicBuddy::Singular::FACTOR() ) if not $self->FACTOR();

   while ( $self->I_accept_and_check_language( qr{ \A (OR|or|oder) (?! [a-zA-Z] ) }xms, LogicBuddy::Singular::dot_or_FACTOR() ) ) {
      $self->{code} .= " || ";

      $self->{IO}->expected( LogicBuddy::Singular::FACTOR() ) if not $self->FACTOR();
   };

   return TRUE;

}; # TERM

sub FACTOR {   
   my ( $self ) = @_;

   if ( $self->I_accept_and_check_language( qr{ \A ( \^ | (?: NOT|not|nicht ) (?! [a-zA-Z] ) ) }xms, LogicBuddy::Singular::Bracket_or_NAME() ) ) {
      $self->{code} .= "! ";
   };

   if ( $self->{IO}->accept( qr{ \A ( [(] ) }xms, LogicBuddy::Singular::Bracket_or_NAME() ) ) {
      $self->{code} .= " ( ";
      $self->EXPRESSION();
      $self->{IO}->expecting( qr{ \A ( [)] ) }xms, [ qw{ and or when ) } ] );
      $self->{code} .= " ) ";
   } elsif ( $self->NAME() ) {
      # OK
   } else {
      $self->{IO}->expected( LogicBuddy::Singular::Bracket_or_NAME() ) if not $self->NAME();
   };

   return TRUE;

}; # FACTOR

sub NAME {
   my ( $self ) = @_;

   my $element;
   if ( $self->FUNCTION_CALL() ) {
   } elsif ( '1' eq $self->{IO}->accept( qr{ \A ( 1 ) }xms, LogicBuddy::Singular::FUNCTION_CALL_or_Value() ) ) {
      $self->{code} .= ' 1 ';
   } elsif ( '0' eq $self->{IO}->accept( qr{ \A ( 0 ) }xms, LogicBuddy::Singular::FUNCTION_CALL_or_Value() ) ) {
      $self->{code} .= ' 0 ';
   } elsif ( $self->I_accept_and_check_language( qr{ \A (TRUE|true|wahr) (?! [a-zA-Z] ) }xms, LogicBuddy::Singular::FUNCTION_CALL_or_Value() ) ) {
      $self->{code} .= ' 1 ';
   } elsif ( $self->I_accept_and_check_language( qr{ \A (FALSE|false|falsch) (?! [a-zA-Z] ) }xms, LogicBuddy::Singular::FUNCTION_CALL_or_Value() ) ) {
      $self->{code} .= ' 0 ';
   } elsif ( $element = $self->known_NAME() ) {
      $self->{code} .= " \$$element";
   } elsif ( $self->{IO}->error_seen() ) {
      $self->{IO}->expected( LogicBuddy::Singular::FUNCTION_CALL_or_Value() );
      return;
   } else {
      $self->{IO}->expected( LogicBuddy::Singular::FUNCTION_CALL_or_Value() );
      return;
   };

   return TRUE;

}; # NAME

sub FUNCTION_CALL  {
   my ( $self ) = @_;

   return FALSE if not $self->FUNCTION_NAME();
   $self->{IO}->expecting( qr{ \A ( [(] ) }xms, "'('" );
   $self->{code} .= ' ( ';
   $self->LIST_OF_EXPRESSIONS();
   $self->{IO}->expecting( qr{ \A ( [)] ) }xms, [ qw{ and or ) }, [ ',' ] ] );
   $self->{code} .= ' ) ';

   return TRUE;

}; # FUNCTION_CALL

sub FUNCTION_NAME {
   my ( $self ) = @_;

   if ( $self->{IO}->accept( qr{ \A (NAND) (?! [a-zA-Z] ) }xms, LogicBuddy::Singular::FUNCTION_CALL_or_Value() ) ) {
      $self->{code} .= " nand ";
   } elsif ( $self->{IO}->accept( qr{ \A (NOR) (?! [a-zA-Z] ) }xms, LogicBuddy::Singular::FUNCTION_CALL_or_Value() ) ) {
      $self->{code} .= " nor ";
   } else {
      return FALSE;
   };

   return TRUE;

}; # FUNCTION_NAME

sub LIST_OF_EXPRESSIONS {
   my ( $self ) = @_;

   $self->{IO}->expected( LogicBuddy::Singular::EXPRESSION() ) if not $self->EXPRESSION();

   while ( $self->{IO}->accept( qr{ \A ( [,] ) }xms , [ qw{ and or not }, [ ',' ] ] ) ) {
      $self->{code} .= ', ';
      $self->{IO}->expected( LogicBuddy::Singular::EXPRESSION() ) if not $self->EXPRESSION();
   };

   return TRUE;

}; # LIST_OF_EXPRESSIONS

################################################################################
# exported methods
################################################################################

sub define_expression {
   my ( $self, $expression ) = @_;

   $self->{expression}            = $expression;
   $self->{input_missing}         = FALSE;
   $self->{output_missing}        = TRUE;
   $self->{input_name}            = {};
   $self->{input_names}           = [];
   $self->{variable_name}         = {};
   $self->{current_variable_name} = EMPTY;
   $self->{output_names}          = [];
   $self->{code}                  = "sub {\n$C_Margin";
   $self->{language_seen}         = FALSE;
   $self->{any_seen}              = FALSE;
   $self->{equal_seen}            = FALSE;
   $self->{command}               = undef;

   $self->{IO}->define_expression( $expression );
   $self->LOGIC();
   $self->{code}                .= "\n};\n";

   foreach my $name ( keys %{ $self->{variable_name} } ) {
      if ( ($self->{variable_name}{$name} != -1 ) and ($self->{variable_name}{$name} < 3) ) {
         $self->{IO}->insert_warning( LogicBuddy::Singular::Error( ":\n" ), LogicBuddy::Singular::variable_9_never_used( "'$name'", '.' ) );
         $self->{IO}->set_error();
      };
   };

   foreach my $name ( keys %{ $self->{input_name} } ) {
      if ( $self->{input_name}{$name} < 2 ) {
         $self->{IO}->insert_warning( LogicBuddy::Singular::Warning( ":\n" ), LogicBuddy::Singular::variable_9_never_used( "'$name'", '.' ) );
         $self->{IO}->set_warning();
      };
   };

   $self->I_miss_input()  if $self->{input_missing};
   $self->I_miss_output() if $self->{output_missing};

   die "\n" if $self->{IO}->error_seen();
   print "code=\n$self->{code}\n" if $G_debug;

   # Define the logic function
   eval "\$LogicBuddy::logic_function = $self->{code}";

   $self->{IO}->insert_error( "$LogicBuddy::Logic::C_expression_exception:", $@ ) if $@;
   $self->{IO}->message_at( EMPTY );
   
   $self->{command} = $LogicBuddy::logic_function;

   return;

}; # define_expression

sub expression_string {
   my ( $self ) = @_;
   return $self->{IO}->expression_string();
}; # expression_string

sub expression {
   my ( $self ) = @_;
   die "$LogicBuddy::Logic::C_no_expression_defined_exception\n" if not $self->{command};
   return $self->{command};
}; # expression

sub input_names {
   my ( $self ) = @_;
   die "$LogicBuddy::Logic::C_no_expression_defined_exception\n" if not $self->{command};
   return @{ $self->{input_names} };
}; # input_names   

sub output_names {
   my ( $self ) = @_;
   die "$LogicBuddy::Logic::C_no_expression_defined_exception\n" if not $self->{command};
   return @{ $self->{output_names} };
}; # output_names   

sub last_error {
   my ( $self ) = @_;
   return->$self->{IO}->last_error();
}; # last_error

sub is_defined {
   my ( $self ) = @_;
   return defined $self->{command};
}; # is_defined

sub status {
   my ( $self ) = @_;
   return ( (not not $self->{code}), $self->{IO}->status() );
}; # status

################################################################################
# class methods
################################################################################

sub reserved_words {
   return @C_reserved_words;
}; # reserved_words

sub reserved_word_pattern {
   my $alternatives = join ' | ', @C_reserved_words;
   return qr{ (?: \A | [^a-zA-Z0-9] ) ( $alternatives ) (?: [^a-zA-Z0-9] | \z ) }xms;
}; # reserved_word_pattern

sub reserved_word_pattern_no_case {
   my $alternatives = join ' | ', @C_reserved_words;
   return qr{ (?: \A | [^a-z0-9] ) ( $alternatives ) (?: [^a-z0-9] | \z ) }ixms;
}; # reserved_word_pattern_no_case

sub input_variable_pattern {
   return $C_simple_name_pattern;
}; # input_variable_pattern

sub output_variable_pattern {
   return $C_simple_name_pattern;
}; # output_variable_pattern

sub comment_pattern {
   return qr{\#.*?\z}xms;
}; # comment_pattern

sub debug {
   ( $G_debug ) = @_;
   LogicBuddy::Logic::IO::debug( $G_debug );
   return;
}; # debug

################################################################################
# Non OO functions used in code
################################################################################

sub nand {
   my ( $first, @others ) = @_;

   foreach my $o ( @others ) {
      $first = ($first and $o);
   };
   return ( not $first );

}; # nand

sub nor {
   my ( $first, @others ) = @_;

   foreach my $o ( @others ) {
      $first = ($first or $o);
   };
   return ( not $first );

}; # nor

################################################################################
# initialize LogicBuddy::Logic
################################################################################

Readonly my $C_do_not_transform => TRUE;
Readonly my $C_do_transform     => FALSE;

LogicBuddy::Locale::Set_local_of( 'reserved_words_description',
                                  $C_do_not_transform,
                                  'English' => $C_reserved_words_description
                                );

################################################################################
# end LogicBuddy::Logic
################################################################################

1;

