=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische_Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Logic:IO - module

=head1 SYNOPSIS

   use LogicBuddy::Logic:IO;

   $logic = LogicBuddy::Logic:IO->new( $viewer, $reserved_word_pattern, $input_name_pattern, $output_name_pattern );

   $logic->ignore_white_spaces();
   $token = $logic->accept( $pattern, $term, $check_language );
   $token = $logic->expecting( $pattern, $term, $check_language );

   $logic->no_more_text();
   $logic->go_back( $length );
   $logic->set_error();
   $logic->set_warning();

   if ( $logic->error_seen() ) ...

   if ( $logic->is_end_of_input() ) ...

   $logic->expected( $term );
   $logic->insert_error( $type, $message, $go_back );
   $logic->insert_warning( $type, $message, $go_back );
   $logic->warning( $type, $message, $line, $column );

   $logic->message_at( $message );

   $logic->define_expression( $expression );

   $expression_string          = $logic->expression_string();
   ($message, $line, $column ) = $logic->last_error();
   ( $error, $warning )        = $logic->status();

=head1 ABSTRACT

LogicBuddy::Logic::IO abstracts parsing and information passing of logical expressions.

=head1 DESCRIPTION

=head2 METHODS

=head3 new

   $logic = LogicBuddy::Logic:IO->new( $viewer, $reserved_word_pattern, $input_name_pattern, $output_name_pattern );

Instantiates an logic processor using $viewer for output.
Defines the pattern for variable names and reserved words.
If $input_name_pattern  is not defined the default pattern ( [a-uA-U] [a-zA-Z0-9]* ) is used.
If $output_name_pattern is not defined the default pattern ( [v-zV-Z] [a-zA-Z0-9]* ) is used.

=head3 ignore_white_spaces

   $logic->ignore_white_spaces();

ignores leading whitespace characters and comments.

=head3 accept

   $token = $logic->accept( $pattern, $term, $check_language );

ignores leading whitespace and looks if the first token matches $pattern.
If $pattern does not match nothing is returned.
If the $pattern matches the $token matched is returned. The $token is consumed from input.
If the $pattern matches the $token it is checked calling the subprogram $&check_language( $token ).
If there is end of input an error is generated with expected $term.

=head3 expecting

   $token = $logic->expecting( $pattern, $term, $check_language );

is like accept but in case $pattern does not match an error is generated with expected $term.

=head3 no_more_text

   $logic->no_more_text();

checks if there is no more text in the input but whitespace ans comment.
It there is text an error is generated.

=head3 go_back

   $logic->go_back( $length );

defines how many characters back the last token was accepted.

=head3 set_error

   $logic->set_error();

sets the error state.

=head3 set_warning

   $logic->set_warning();

sets the warning state.

=head3 error_seen

   if ( $logic->error_seen() )

returns the error state.

=head3 is_end_of_input

   if ( $logic->is_end_of_input() ) ...

returns if there are only white spaces before end of input.

=head3 expected

   $logic->expected( $term );

generates an error message that the $term was expected.

=head3 insert_error

   $logic->insert_error( $type, $message, $go_back );

inserts an error message of $type.
The error occurred at the input position $go_back characters back.

=head3 insert_warning

   $logic->insert_warning( $type, $message, $go_back );

inserts a warning of $type.
The warning occurred at the input position $go_back characters back.

=head3 warning

   $logic->warning( $type, $message, $line, $column );

inserts a warning of $type. The cursor is will be at $line and $column.

=head3 define_expression

   $logic->define_expression( $expression );

Defines a set of new logic functions from $expression string.

=head3 expression_string

   $expression_string = $logic->expression_string();

Returns the expression string of the last call of define_expression.

If define_expression was not called undef is returned.

=head3 last_error

   ($message, $line, $column ) = $logic->last_error();

Returns the last error $message detected at $line and $column.
If there was no error at the last call of define_expression $message, $line and $column are undef.

=head3 status

   ( $error, $warning ) = $logic->status();

Returns if there where an $error or $warning.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische_Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;
use warnings;

use Carp;

use LogicBuddy::Utilities;
use LogicBuddy::Logic::Text;

package LogicBuddy::Logic::IO;

use English '-no_match_vars';

################################################################################
# constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

use Readonly;

################################################################################
# global Variable
################################################################################

my $G_debug = FALSE;

################################################################################
# class LogicBuddy::Logic::IO
################################################################################

sub new {
   my ( $proto, $editor, $reserved_word_pattern, $input_name_pattern, $output_name_pattern ) = @_;

   my $class = ref($proto) || $proto;
   my $self;

   $self->{Editor}                = $editor;
   $self->{reserved_word_pattern} = $reserved_word_pattern || qr{ ( [,.()=] ) }xms;
   $self->{input_name_pattern}    = $input_name_pattern    || qr{ \A ( [a-uA-U] [a-zA-Z0-9]* ) }xms;
   $self->{output_name_pattern}   = $output_name_pattern   || qr{ \A ( [v-zV-Z] [a-zA-Z0-9]* ) }xms;
   $self->{expression}            = undef;
   $self->{input}                 = undef;
   $self->{line_number}           = undef;
   $self->{column_number}         = undef;
   $self->{pre}                   = undef;
   $self->{accepted}              = undef;
   $self->{accepted_prev}         = undef;
   $self->{message}               = undef;
   $self->{last_message_at}       = EMPTY;

   $self->{error}                 = FALSE;
   $self->{warning}               = FALSE;
   $self->{go_back}               = 0;

   bless $self, $class;

   return $self;

}; # new

################################################################################
# internal methods
################################################################################

sub I_insert_message {
   my ( $self, $type, $message, $go_back ) = @_;

   $go_back = 0 if not $go_back;

   if ( $self->{last_message_at} and $self->{last_message_at} eq $self->{pre} ) {
      my $translated = LogicBuddy::Logic::Text::translate( $message );
      return "${type}$translated\n";
   } else {
      $self->{last_message_at} = $self->{pre};
   };

   my $pre;
   if ( $self->{pre} =~ m/ ( \n ( [^\n]+? ) [.\n]*? )+? \z /xms ) {
      $pre = $1;
   } else {
      $pre = $self->{pre};
   };

   $pre =~ s/ \A \n //xms;

   my $where;
   if ( $self->{input} =~ m/ \A ( .+? ) (\n|\z) /xms ) {
      $where = "$pre$1";
   } else {
      $where = "$pre$self->{input}";
   };
   $where =~ s/ ( \s | \n )+ \z //xms;

   my $here;
   if ( $self->{pre} =~ m/ .* \n ( .*? ) \z /xms ) {
      $here = ' ' x ( (length $1) - $go_back );
      $self->{column_number} = (length $1) - $go_back;
   } else {
      $here = ' ' x ( (length $self->{pre}) - $go_back );
      $self->{column_number} = (length $self->{pre}) - $go_back;
   };

   my $translated = LogicBuddy::Logic::Text::translate( $message );
   return "\n$where\n$here^\n${type}$translated\n";

}; # I_insert_message

sub I_got_unexpected {
   my ( $self, $term, $back ) = @_;
   $self->insert_error( LogicBuddy::Singular::Unexpected( ': ' ), $term, $back );
   return;
}; # I_got_unexpected

################################################################################
# exported methods
################################################################################

sub no_more_text {
   my ( $self ) = @_;

   $self->ignore_white_spaces();

   if ( $self->{input} =~ s{ \A ( [^\s\n]+ ) }{}xms ) {
      $self->{pre} .= $1;
      $self->I_got_unexpected( LogicBuddy::Singular::text(), length $1 );
   };

   return;

}; # no_more_text

sub go_back {
   my ( $self, $length ) = @_;
   $self->{go_back} = $length;
   return;
}; # go_back

sub set_error {
   my ( $self ) = @_;
   $self->{error} = TRUE;
   return;
}; # set_error

sub set_warning {
   my ( $self ) = @_;
   $self->{warning} = TRUE;
   return;
}; # set_warning

sub error_seen {
   my ( $self ) = @_;
   return $self->{error};
}; # error_seen

sub ignore_white_spaces {
   my ( $self ) = @_;

   my $pre      = EMPTY;
   my $new_line = EMPTY;
   COMMENT: while ( TRUE ) {
      while ( $self->{input} =~ s{ \A \s*? \n }{}xms ) {
         $new_line = "\n";
         $self->{line_number}++;
      };

      if ( $self->{input} =~ s{ \A ( \s* \# .*? \n ) }{}xms ) {
         $pre = $1;
         $self->{line_number}++;
      } else {
         last COMMENT;
      };
   }; # COMMENT

   $self->{pre} .= "$new_line$pre";
   $pre          = EMPTY;

   while ( $self->{input} =~ s{ \A ( \s+? \n ) }{}xms ) {
      $pre = $1;
      $self->{line_number}++;
   };

   if ( $self->{input} =~ s{ \A ( \s+ ) }{}xms ) {
      $pre .= $1;
   };

   if ( $self->{input} ) {
      $self->{pre} .= $pre;
#  else end of input: ignore trailing whitespaces
   };

   return;

}; # ignore_white_spaces

sub is_end_of_input {
   my ( $self ) = @_;

   $self->ignore_white_spaces();
   if ( $self->{input} ) {
      return FALSE;
   } else {
      $self->{pre} =~ s/ (\s|\n)+ \z //xms;
      return TRUE;
   };

}; # is_end_of_input

sub accept {
   my ( $self, $pattern, $term, $check_language ) = @_;

# The $pattern has to be like
#    qr{ \A ( THE_PATTERN ) }xms
# or
#    qr{ \A ( THE_PATTERN ) ($! [a-zA-Z] ) }xms
#
# $pattern has to start by \A
#              the first ($1) collecting brace is accepted
#
# A lookahead assertion ($! [a-zA-Z] ) can be used to state that it should not be followed by

   $self->ignore_white_spaces();

   if ( not $self->{input} ) {
      $self->{pre} =~ s/ (\s|\n)+ \z //xms;
      if ( $term ) {
         my $translated = LogicBuddy::Logic::Text::translate( $term );
         $self->I_got_unexpected( LogicBuddy::Singular::end_of_input() . "\n" . LogicBuddy::Singular::Expected( ":\n" ) . "\n$translated", 0 );
      } else {
         $self->I_got_unexpected( LogicBuddy::Singular::end_of_input() );
      };
   };

   my $type;
   if ( $pattern eq 'INPUT_NAME' ) {
      $type    = $pattern;
      $pattern = $self->{input_name_pattern};
   } elsif ( $pattern eq 'OUTPUT_NAME' ) {
      $type    = $pattern;
      $pattern = $self->{output_name_pattern};
   } elsif ( $pattern eq 'NAME' ) {
      $type    = $pattern;
      $pattern = $self->{output_name_pattern};
   };

   if ( $self->{input} =~ s{ $pattern }{}xms ) {
      my $accepted  = $1;
      $self->{pre} .= $accepted;
      &$check_language( $accepted ) if $check_language;
      if ( $accepted =~ m/ \w \z /xms ) {
         if ( $self->{input} =~ m{ \A \w }xms ) {
            if ( $type ) {
               if ( $type eq 'INPUT_NAME' ) {
                  $self->expected( LogicBuddy::Singular::INPUT_NAME() );
               } elsif ( $type eq 'OUTPUT_NAME' ) {
                  $self->expected( LogicBuddy::Singular::OUTPUT_NAME() );
               } else {
                  $self->expected( LogicBuddy::Singular::NAME() );
               };
            } else {
               $self->expected( LogicBuddy::Singular::non_alphanumerical_character() );
            };
         };
      };
      $self->{accepted_prev} = $self->{accepted};
      $self->{accepted}      = defined $type  ?  $type  :  $accepted;

      print "accepted '$self->{accepted}' that is '$accepted' with '$pattern'\n" if $G_debug;

      return $accepted;
   };

   #      EMPTY is important because the result may be compared
   return EMPTY;

}; # accept

sub insert_error {
   my ( $self, $type, $message, $go_back ) = @_;

   print "error( $type, $message )\n" if $G_debug;

   $self->{message}  = "$type $message";

   my $error_message = $self->I_insert_message( $type, $message, $go_back || $self->{go_back} );
   $self->{Editor}->go_to( $self->{line_number}, $self->{column_number} ) if $self->{Editor};

   $self->{go_back}  = 0;
   $self->{error}    = TRUE;

   die "$error_message\n";

}; # insert_error

sub insert_warning {
   my ( $self, $type, $message, $go_back ) = @_;

   my $warning;
   if ( $self->{error} or $self->{warning} ) {
      $warning = LogicBuddy::Logic::Text::translate( $message ) . "\n";
   } else {
      $warning = $self->I_insert_message( $type, $message, $go_back || $self->{go_back} );
   };

   $self->{Editor}->message_at( $warning, $self->{line_number}, $self->{column_number} ) if $self->{Editor};
   print "$warning\n" if $G_debug or not $self->{Editor};

   return;

}; # insert_warning

sub warning {
   my ( $self, $type, $message, $line, $column ) = @_;

   my $translated = LogicBuddy::Logic::Text::translate( $message );
   $self->{Editor}->message_at( "$type$translated", $line, $column );
   return;

}; # warning

sub message_at {
   my ( $self, $message ) = @_;
   $self->{Editor}->message_at( $message, $self->{line_number}, $self->{column_number} ) if $self->{Editor};
   print STDERR "at $self->{line_number}, $self->{column_number}: $message\n" if $message and ($G_debug or not $self->{Editor});
   return;
}; # message_at

sub expected {
   my ( $self, $term ) = @_;

   if ( $self->{reserved_word_pattern} and $self->{input} =~ m{ $self->{reserved_word_pattern} }xms ) {
      $self->insert_error( LogicBuddy::Singular::Expected_never_9_but( "'$1'" ) . ":\n", $term );
   } else {
      $self->insert_error( LogicBuddy::Singular::Expected( ":\n" ), $term );
   };
   return;
}; # expected

sub expecting {
   my ( $self, $pattern, $term, $check_language ) = @_;

   my $result = $self->accept( $pattern, $term, $check_language );
   $self->expected( $term ) if not $result;
   return $result;

}; # expecting

sub define_expression {
   my ( $self, $expression ) = @_;

   $self->{input}                 = $expression;
   $self->{expression}            = $expression;
   $self->{line_number}           = 1;
   $self->{column_number}         = 0;
   $self->{pre}                   = EMPTY;
   $self->{pre}                   = undef;
   $self->{accepted}              = undef;
   $self->{message}               = undef;
   $self->{last_message_at}       = EMPTY;

   $self->{error}                 = FALSE;
   $self->{warning}               = FALSE;
   $self->{go_back}               = 0;

}; # define_expression

sub expression_string {
   my ( $self ) = @_;
   return $self->{expression};
}; # expression_string

sub last_error {
   my ( $self ) = @_;
   return if not $self->{message};
   return ( $self->{message}, $self->{line_number}, $self->{column_number} );
}; # last_error

sub status {
   my ( $self ) = @_;
   return ( $self->{error}, $self->{warning} );
}; # status

################################################################################
# class methods
################################################################################

sub debug {
   ( $G_debug ) = @_;
}; # debug

################################################################################
# end LogicBuddy::Logic::IO
################################################################################

1;

