=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische_Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Logic::Text - module

=head1 SYNOPSIS

   use LogicBuddy::Logic::Text;

   $text = LogicBuddy::Logic::Text::translate( $term );
   $text = LogicBuddy::Logic::Text::syntax_description( $language );
   $text = LogicBuddy::Logic::Text::full_syntax_description();

   $text = LogicBuddy::Logic::Text::rule( $number, $language );
   @text = LogicBuddy::Logic::Text::rules( $language );
   $text = LogicBuddy::Logic::Text::example( $number, $language );
   @text = LogicBuddy::Logic::Text::examples( $language );

=head1 ABSTRACT

LogicBuddy::Logic abstracts text to logical expressions.

=head1 DESCRIPTION

=head2 SUBPROGRAMS

=head3 translate

   $text = LogicBuddy::Logic::Text::translate( $term );

Translates the $term using LogicBuddy::Logic::Locale.
If $term is an array reference all words in the array are conneced by 'or'.
If a word in that array is a array reference then the words in the new array are usually missed.

=head3 syntax_description

   $text = LogicBuddy::Logic::Text::syntax_description( $language );

Returns the syntax description of $language.
If der $language is not defined English is used.

=head3 full_syntax_description

   $text = LogicBuddy::Logic::Text::full_syntax_description();

Returns the full syntax description in English.

=head3 rule

   $text = LogicBuddy::Logic::Text::rule( $number, $language );

Returns the rule $number for $language.
If the $language is not defined English is used.
If the rule $number is not defined undef is used.

=head3 rules

   @text = LogicBuddy::Logic::Text::rules( $language );

Returns all rules for $language.
If the $language is not defined English is used.

=head3 example

   $text = LogicBuddy::Logic::Text::example( $number, $language );

Returns the example $number for $language.
If the $language is not defined English is used.
If the example $number is not defined undef is used.

=head3 examples

   @text = LogicBuddy::Logic::Text::examples( $language );

Returns all examples for $language.
If the $language is not defined English is used.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische_Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;
use warnings;

use Carp;

use utf8;

use LogicBuddy::Utilities;
use LogicBuddy::Locale;

################################################################################
# package LogicBuddy::Logic::Text
################################################################################

package LogicBuddy::Logic::Text;

use English '-no_match_vars';

################################################################################
# constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

use Readonly;

Readonly my $C_syntax_full      => "
LOGIC               ::= INPUT ASSIGNMENTS OUTPUT;

INPUT               ::= ( 'Input' | 'Eingang' )
                        ( '=' | 'is' | 'are' | 'ist' | 'sind' )
                        LIST_OF_INPUTS '.';

LIST_OF_INPUTS      ::= NAME { ',' NAME };

NAME                ::= [a-zA-Z] [a-zA-Z0-9_]*;

OUTPUT              ::= ( 'Output' | 'Ausgang' )
                        ( '=' | 'is' | 'are' | 'ist' | 'sind' )
                        LIST_OF_OUTPUTS '.';

LIST_OF_OUTPUTS     ::= { NAME | ',' NAME };

ASSIGNMENTS         ::= ASSIGNMENT { '.' ASSIGNMENT } '.';

ASSIGNMENT          ::= NAME ( '=' | 'is' | 'ist' ) ( ANY_CONDITION | VALUE );

VALUE               ::= EXPRESSION [ CONDITION ];

ANY_CONDITION       ::= ( 'any' | 'beliebig' ) CONDITION;

CONDITION           ::= ( 'when' | 'wenn' ) EXPRESSION;

EXPRESSION          ::= [ NOT ] TERM { ( 'and' | 'und' ) [ NOT ] TERM };

NOT                 ::= ( 'not' | 'nicht' | '^' ) ;

TERM                ::= FACTOR { ( 'or' | 'oder' ) FACTOR };

FACTOR              ::= [ NOT ] ( NAME | '(' EXPRESSION ')' );

NAME                ::= OUTPUT_NAME | INPUT_NAME | FUNCTION_CALL |
                        'true' | 'false' | 'wahr' | 'falsch' | '1' | '0';

FUNCTION_CALL       ::= FUNCTION_NAME '(' LIST_OF_EXPRESSIONS ')';

FUNCTION_NAME       ::= 'NAND' | 'NOR' ;

LIST_OF_EXPRESSIONS ::= EXPRESSION { ',' EXPRESSION };
";

Readonly my $C_syntax_en        => "
LOGIC               ::= INPUT ASSIGNMENTS OUTPUT;

INPUT               ::= 'Input' ( '=' | 'is' | 'are' ) LIST_OF_INPUTS '.';

LIST_OF_INPUTS      ::= NAME { ',' NAME };

NAME                ::= [a-zA-Z] [a-zA-Z0-9_]*;

OUTPUT              ::= 'Output' ( '=' | 'is' | 'are' ) LIST_OF_OUTPUTS '.';

LIST_OF_OUTPUTS     ::= { NAME | ',' NAME };

ASSIGNMENTS         ::= ASSIGNMENT { '.' ASSIGNMENT } '.';

ASSIGNMENT          ::= OUTPUT_NAME ( '=' | 'is' ) ( ANY_CONDITION | VALUE );

VALUE               ::= EXPRESSION [ CONDITION ];

ANY_CONDITION       ::= 'any' CONDITION;

CONDITION           ::= 'when' EXPRESSION;

EXPRESSION          ::= [ NOT ] TERM { 'and' [ NOT ] TERM };

NOT                 ::= ( 'not' | '^' );

TERM                ::= FACTOR { 'or'  FACTOR };

FACTOR              ::= [ NOT ] ( NAME | '(' EXPRESSION ')' );

NAME                ::= OUTPUT_NAME | INPUT_NAME | FUNCTION_CALL |
                        'true' | 'false' | '1' | '0';

FUNCTION_CALL       ::= FUNCTION_NAME '(' LIST_OF_EXPRESSIONS ')';

FUNCTION_NAME       ::= 'NAND' | 'NOR';

LIST_OF_EXPRESSIONS ::= EXPRESSION { ',' EXPRESSION };
";

Readonly my $C_syntax_de        => "
LOGIK                ::= EINGANG ZUWEISUNGEN AUSGANG;

EINGANG              ::= 'Eingang' ( '=' | 'ist' | 'sind' ) LISTE_DER_EINGAENGE '.';

LISTE_DER_EINGAENGE  ::= NAME { ',' NAME };

NAME                 ::= [a-zA-Z] [a-zA-Z0-9_]*;

AUSGANG              ::= 'Ausgang' ( '=' | 'ist' | 'sind' ) LISTE_DER_AUSGAENGE '.';

LISTE_DER_AUSGAENGE  ::= { NAME | ',' NAME };

ZUWEISUNGEN          ::= ZUWEISUNG { '.' ZUWEISUNG } '.';

ZUWEISUNG            ::= AUSGANGS_NAME ( '=' | 'ist' ) ( BELIEBIG_BEDINGUNG | WERT );

WERT                 ::= AUSDRUCK [ BEDINGUNG ];

BELIEBIG_BEDINGUNG   ::= 'beliebig' BEDINGUNG;

BEDINGUNG            ::= 'wenn' AUSDRUCK;

AUSDRUCK             ::= [ NICHT ] TERM { 'und' [ NICHT ] TERM };

NICHT                ::= ( 'nicht' | '^' );

TERM                 ::= FAKTOR { 'oder' FAKTOR };

FAKTOR               ::= [ NICHT ] ( NAME | '(' AUSDRUCK ')' );

NAME                 ::= AUSGANGS_NAME | EINGANGS_NAME | FUNKTIONS_AUFRUF |
                         'wahr' | 'falsch' | '1' | '0';

FUNKTIONS_AUFRUF     ::= FUNKTIONS_NAME '(' LISTE_DER_AUSDRUECKE ')';

FUNKTIONS_NAME       ::= 'NAND' | 'NOR';

LISTE_DER_AUSDRUECKE ::= AUSDRUCK { ',' AUSDRUCK };
";

Readonly my @C_equivalent_symbols => ( [ qw{ Input Eingang } ],
                                       [ qw{ Output Ausgang } ],
                                       [ qw{ is ist } ],
                                       [ qw{ and und } ],
                                       [ qw{ or oder } ],
                                       [ qw{ not nicht ^ } ],
                                       [ qw{ when wenn } ],
                                       [ qw{ false falsch 0 } ],
                                       [ qw{ true wahr 1 } ],
                                       [ qw{ any beliebig } ]
                                     );

Readonly my @C_equivalent_input_symbols => ( [ qw{ is are ist sind } ],
                                             [ 'and', 'und', ',' ]
                                           );

Readonly my $C_example_de_1 =>
"Eingang = a, b, c, d.
x = a und b.       # Kommentar bis zum Ende der Zeile
y = c oder d.
v = wahr wenn d und nicht a.
z = y und ( nicht a oder nicht c ).
x = beliebig wenn nicht c.
Ausgang = x, z.
";

Readonly my $C_example_en_1 =>
"Input = a, b, c, d.
x = a and b.       # Comment to end of line
y = c or d.
v = true when d and not a.
z = y and ( not a or not c ).
x = any when not c.
Output = x, z.
";

Readonly my @C_example_en => ( EMPTY,
                               $C_example_en_1,
                             );

Readonly my @C_example_de => ( EMPTY,
                               $C_example_de_1,
                             );

Readonly my $C_rule_en_1  => "'Input ...' is the first statement";
Readonly my $C_rule_de_1  => "'Eingang ...' ist die erste Aussage";
Readonly my $C_rule_en_2  => "'Output ...' is the last statement";
Readonly my $C_rule_de_2  => "'Ausgang ...' ist die letzte Aussage";
Readonly my $C_rule_en_3  => "Each statement ends with a '.' Point.";
Readonly my $C_rule_de_3  => "Jede Aussage endet mit einem '.' Punkt.";
Readonly my $C_rule_en_4  => "All variables 'a' ... 'z'.\n" .
                             "These input variables can be up to 4 characters long and contain digits or '_'.";
Readonly my $C_rule_de_4  => "Alle Variablen beginnen mit 'a' ... 'z'.\n" .
                             "Diese Variablen dürfen maximal 4 Zeichen lang sein, Ziffern und '_' enthalten.";
Readonly my $C_rule_en_5  => "All normal statements begin with a variable.";
Readonly my $C_rule_de_5  => "Alle normalen Aussagen beginnen mit einer Variablen.";
Readonly my $C_rule_en_6  => "AND before OR: AND is calculated before OR.";
Readonly my $C_rule_de_6  => "UND vor ODER: UND wird vor ODER berechnet.";
Readonly my $C_rule_en_7  => "NOT refers only to the immediately following.";
Readonly my $C_rule_de_7  => "NICHT bezieht sich nur auf das unmittelbar folgende.";
Readonly my $C_rule_en_8  => "Brackets '(' are evaluated first";
Readonly my $C_rule_de_8  => "Klammern '(' werden zuerst ausgewertet";
Readonly my $C_rule_en_9  => "In case of ambiguity, place brackets.";
Readonly my $C_rule_de_9  => "Bei Unklarheit Klammern setzen.";
Readonly my $C_rule_en_10 => "TRUE and FALSE are allowed.";
Readonly my $C_rule_de_10 => "WAHR und FALSCH sind erlaubt.";
Readonly my $C_rule_en_11 => "A later statement goes beyond a previous one.";
Readonly my $C_rule_de_11 => "Eine spätere Aussage geht über eine vorherige.";
Readonly my $C_rule_en_12 => "For WHEN, the variable only receives the value if the statement after WHEN is true.";
Readonly my $C_rule_de_12 => "Bei WENN erhält die Variable nur dann den Wert, wenn die Aussage hinter WENN zutrifft.";
Readonly my $C_rule_en_13 => "Each variable has the initial value FALSE.";
Readonly my $C_rule_de_13 => "Jede Variable hat den anfänglichen Wert FALSCH.";
Readonly my $C_rule_en_14 => "Statements with ANY may only be immediately before OUTPUT.";
Readonly my $C_rule_de_14 => "Aussagen mit BELIEBIG dürfen nur unmittelbar vor AUSGANG stehen.";
Readonly my $C_rule_en_15 => "Uppercase and lowercase is allowed.";
Readonly my $C_rule_de_15 => "Groß- und Kleinschreibung ist erlaubt.";
Readonly my $C_rule_en_16 => "Comments begin with # and go to the end of a line.";
Readonly my $C_rule_de_16 => "Kommentare beginnen mit # und gehen bis zum Ende einer Zeile.";

Readonly my @C_rules_en   => ( undef,
                               $C_rule_en_1,
                               $C_rule_en_2,
                               $C_rule_en_3,
                               $C_rule_en_4,
                               $C_rule_en_5,
                               $C_rule_en_6,
                               $C_rule_en_7,
                               $C_rule_en_8,
                               $C_rule_en_9,
                               $C_rule_en_10,
                               $C_rule_en_11,
                               $C_rule_en_12,
                               $C_rule_en_13,
                               $C_rule_en_14,
                               $C_rule_en_15,
                               $C_rule_en_16
                             );

Readonly my @C_rules_de   => ( undef,
                               $C_rule_de_1,
                               $C_rule_de_2,
                               $C_rule_de_3,
                               $C_rule_de_4,
                               $C_rule_de_5,
                               $C_rule_de_6,
                               $C_rule_de_7,
                               $C_rule_de_8,
                               $C_rule_de_9,
                               $C_rule_de_10,
                               $C_rule_de_11,
                               $C_rule_de_12,
                               $C_rule_de_13,
                               $C_rule_de_14,
                               $C_rule_de_15,
                               $C_rule_de_16
                             );

################################################################################
# global Variable
################################################################################

################################################################################
# internal subprograms
################################################################################

################################################################################
# exported subprograms
################################################################################

sub rule {
   my ( $number, $language ) = @_;

   return $C_rules_en[$number] if not $language;
   return $C_rules_de[$number] if $language eq 'German';
   return $C_rules_en[$number];

}; # rule

sub rules {
   my ( $language ) = @_;

   return @C_rules_en[ 1 .. $#C_rules_en ] if not $language;
   return @C_rules_de[ 1 .. $#C_rules_de ] if $language eq 'German';
   return @C_rules_de[ 1 .. $#C_rules_de ] if $language eq 'Deutsch';
   return @C_rules_de[ 1 .. $#C_rules_de ] if $language eq 'de';
   return @C_rules_en[ 1 .. $#C_rules_en ];

}; # rules

sub example {
   my ( $number, $language ) = @_;

   return $C_example_en[$number] if not $language;
   return $C_example_de[$number] if $language eq 'German';
   return $C_example_de[$number] if $language eq 'Deutsch';
   return $C_example_de[$number] if $language eq 'de';
   return $C_example_en[$number];

}; # example

sub examples {
   my ( $language ) = @_;

   return @C_example_en[ 1 .. $#C_example_en ] if not $language;
   return @C_example_de[ 1 .. $#C_example_de ] if $language eq 'German';
   return @C_example_de[ 1 .. $#C_example_de ] if $language eq 'Deutsch';
   return @C_example_de[ 1 .. $#C_example_de ] if $language eq 'de';
   return @C_example_en[ 1 .. $#C_example_en ];

}; # examples

sub equivalent_symbols {
  return @C_equivalent_symbols;
}; # equivalent_symbols

sub equivalent_input_symbols {
   return @C_equivalent_input_symbols;
}; # equivalent_input_symbols

sub translate;

# $seen avoids multiple listing of words.
# If $usally then the next array ref is a list of usually missed words.
sub translate {
   my ( $term, $seen, $usally ) = @_;

   $seen          = {} if not $seen;
   my $translated = EMPTY;
   if ( ref $term ) {
      foreach my $word ( @$term ) {
         $translated .= LogicBuddy::Singular::or( ' ' ) if $translated;
         if ( ref $word ) {
            if ( $usally ) {
               $translated .= "\n" . LogicBuddy::Singular::usually_a_9_is_missing( translate( $word, $seen, $usally ) );
            } else {
               $translated .= translate( $word, $seen, $usally );
            };
         } elsif ( exists $seen->{$word} ) {
            # ignore
         } elsif ( $word =~ m/ \A [A-Z_]+ \z /xms ) {
            $translated .= LogicBuddy::Locale::TranslationOf( $word )->[0] . ' ';
            $seen->{$word} = TRUE;
         } elsif ( $word =~ m/ \n /xms ) {
            $translated .= LogicBuddy::Locale::TranslationOf( $word )->[0] . "\n$Tk::Configuration::C_margin";
            $seen->{$word} = TRUE;
         } else {
            $translated .= "'" . LogicBuddy::Locale::TranslationOf( $word )->[0] . "' ";
            $seen->{$word} = TRUE;
         };
         $usally = TRUE;
      };
   } else {
      $translated = $term;
      $seen->{$term} = TRUE;
   };

   return $translated;

}; # translate

sub syntax_description {
   my ( $language ) = @_;

   return $C_syntax_en if not $language;
   return $C_syntax_de if $language eq 'de';
   return $C_syntax_de if $language eq 'Deutsch';
   return $C_syntax_de if $language eq 'German';
   return $C_syntax_en;

}; # syntax_description

sub full_syntax_description {
   return $C_syntax_full;
}; # full_syntax_description

################################################################################
# initialize LogicBuddy::Logic
################################################################################

Readonly my $C_do_not_transform => TRUE;
Readonly my $C_do_transform     => FALSE;

LogicBuddy::Locale::Set_local_of( 'syntax_description',
                                  $C_do_not_transform,
                                  'English' => $C_syntax_en,
                                  'German'  => $C_syntax_de
                                );

LogicBuddy::Locale::Set_local_of( 'full_syntax_description',
                                  $C_do_not_transform,
                                  'English' => $C_syntax_full
                                );

################################################################################
# end LogicBuddy::Logic::Text
################################################################################

1;

