=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Tk::Balloon - module

=head1 SYNOPSIS

   use LogicBuddy::Tk::Balloon;

   $balloon = LogicBuddy::Tk::Balloon->new( $parent );

   $button = $balloon->Button( $parent );

   $balloon->attach( $widget, $message, %options );

   $balloon->detach( $widget );

   $balloon->on( $yes_or_no );

   if ( $balloon->is_on() ) ...

   $balloon->menu( $parent );

=head1 ABSTRACT

LogicBuddy::Tk::Balloon balloon provides balloon help messages while displaying LogicBuddy.

=head1 DESCRIPTION

=head2 METHODS

=head3 new

   $balloon = LogicBuddy::Tk::Balloon->new( $parent );

Instantiates a new Balloon widget.

=head3 Button

   $button = $balloon->Button( $parent );

Creates a balloon toggle button. The button is not packed.

=head3 attach

   $balloon->attach( $widget, $message, %options );

Attaches a balloon $message to the $widget.
The %options are optional and correspond to the options of Tk::Balloon.

=head3 detach

   $balloon->detach( $widget );

Detaches the balloon of $widget.

=head3 on

   $balloon->on( $yes_or_no );

The balloon is active or not

=head3 is_on

   if ( $balloon->is_on() ) ...

Returns if the balloon is on

=head3 menu

  $balloon->menu( $parent );

Creates a check button with balloon on/off

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;
use warnings;

use Tk;
use Tk::Balloon;

use LogicBuddy::Singular;
use LogicBuddy::Tk::Configuration;

package LogicBuddy::Tk::Balloon;

################################################################################
# global constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;

################################################################################
# global variable
################################################################################

################################################################################
# class LogicBuddy::Tk::Balloon;
################################################################################

sub new {
   my ( $proto, $parent )= @_;

   my $class = ref($proto) || $proto;
   my $self;

   $self->{Balloon_on}      = TRUE;
   $self->{Balloon}         = $parent->Balloon( -balloonposition => 'mouse',
                                                -background      => $Tk::Configuration::C_Balloon_background,
                                                -font            => $Tk::Configuration::C_small_font
                                              );
   $self->{Balloon_buttons} = [];

   return bless $self, $class;

}; # new

################################################################################
# exported methods
################################################################################

sub Button {
   my ( $self, $parent ) = @_;

   my $button = $parent->Button( -command => sub { $self->on( not $self->{Balloon_on} ) },
                                 -text    => LogicBuddy::Singular::balloon_Help_on_off(),
                                 -image   => Tk::Images::balloon( $parent )
                               );
   $button->configure( -relief => 'sunken' );
   $self->attach( $button, LogicBuddy::Singular::Balloon_Help_on_off() );

   push @{$self->{Balloon_buttons}}, $button;

   return $button;

}; # Button

sub attach {
   my ( $self, $widget, $message, %options )= @_;
   $self->{Balloon}->attach( $widget, -msg => $message, %options );
   return;
}; # attach

sub detach {
   my ( $self, $widget )= @_;
   $self->{Balloon}->detach( $widget );
   return;
}; # detach

sub on {
   my ( $self, $balloon_on )= @_;

   $self->{Balloon_on} = $balloon_on;

   if ( $self->{Balloon_on} ) {
      foreach my $button( @{$self->{Balloon_buttons}} ) {
         $button->configure( -relief  => 'sunken' );
      };
      $self->{Balloon}->configure( -state => 'balloon' ) if $self->{Balloon};
   } else {
      foreach my $button( @{$self->{Balloon_buttons}} ) {
         $button->configure( -relief  => 'flat' );
      };
      $self->{Balloon}->configure( -state => 'none' ) if $self->{Balloon};
   };
   return;

}; # on

sub is_on {
   my ( $self )= @_;
   return $self->{Balloon_on};
}; # is_on

sub menu {
   my ( $self, $parent )= @_;

   $parent->Checkbutton( -label    => LogicBuddy::Singular::Balloon_help(),
                         -variable => \$self->{Balloon_on},
                         -command  => sub { $self->on( $self->{Balloon_on} ) }
                       );
   return;

}; # menu

################################################################################
# end LogicBuddy::Tk::Balloon
################################################################################

1;

