=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Tk::Editor - module

=head1 SYNOPSIS

   use LogicBuddy::Tk::Editor;

   $editor = new LogicBuddy::Tk::Editor( $parent );

   $button = $editor->status_button( $parent );

   $editor->edit();

   $editor->set_language( $language );

   $language = $editor->language();

   $editor->go_to( $line, $column );

   $editor->message_at( $message, $line, $column );

   $editor->changed();

   $editor->enter_function( $function );

   $editor->insert_at( $position, $text );

   $editor->insert_correction_at( $position, $text );

=head1 ABSTRACT

LogicBuddy::Tk::Editor abstracts editing LogicBuddy logic functions.

=head1 DESCRIPTION

=head2 METHODS

=head3 new

   $editor = new LogicBuddy::Tk::Editor( $parent );

Instantiates a LogicBuddy editor for logic functions.

=head3 status_button

   $button = $editor->status_button( $parent );

Returns a button in $parent displaying the status of function and truth table.
The $button is not packed.

=head3 edit

   $editor->edit();

The editor is visible.

=head3 set_language

   $editor->set_language( $language );

Defines the language used.

=head3 language

   $language = $editor->language();

Returns the language used.

=head3 go_to

   $editor->go_to( $line, $column );

Moves the cursor in the window to $line, $column.
If $column is not defined the cursor is moved the the position described by $line.

=head3 message_at

   $editor->message_at( $message, $line, $column );

Puts the $message into the message window and moves the cursor in the edit window to $line, $column.
If $column is not defined the cursor is moved the the position described by $line.

=head3 changed

   $editor->changed();

Sets the state of the editor to changed

=head3 enter_function

   $editor->enter_function( $function );

Replaces the logical function in the editor window by $function.
All messages are deleted.

=head3 insert_at

   $editor->insert_at( $position, $text );

At the text $position the $text is inserted.

=head3 insert_correction_at

   $editor->insert_correction_at( $position, $text );

At the text $position the $text is inserted and marked as correction.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;

use Tk;
use Tk::Pane;
use Tk::Images;
use Tk::ROText;
use Tk::Help;

use LogicBuddy::Configuration;
use LogicBuddy::Plural;
use LogicBuddy::Singular;
use LogicBuddy::Utilities;
use LogicBuddy::Tk::Configuration;

use LogicBuddy::Tk::Utilities;
use LogicBuddy::Tk::Balloon;
use LogicBuddy::Tk::Help;

use LogicBuddy::Logic;

package LogicBuddy::Tk::Editor;

use English '-no_match_vars';

################################################################################
# global constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

use Readonly;

Readonly my $C_reaction_period    => 100; # ms
Readonly my $C_Pad                => $Tk::Configuration::C_Pad;
Readonly my %C_Pads               => %Tk::Configuration::C_Pads;
Readonly my $C_Pad_Standard       => $Tk::Configuration::C_Pad_Standard;
Readonly my %C_Standard_Pads      => %Tk::Configuration::C_Standard_Pads;

Readonly my $C_message_height     => $Tk::Configuration::C_message_height;
Readonly my $C_editor_height      => $Tk::Configuration::C_editor_height;

Readonly my $C_length_status      => 300; # Wrap length of status in header

Readonly my %C_status             => (
               empty              => [ LogicBuddy::Singular::No_logical_function_defined(),                             'C_sleep' ],
               differ             => [ LogicBuddy::Singular::Table_does_not_match_logical_function(),                   'C_aha' ],
               differ_and_error   => [ LogicBuddy::Singular::Table_does_not_match_logical_function_and_error_message(), 'C_angry' ],
               match_and_warning  => [ LogicBuddy::Singular::Table_matches_logical_function_but_warning(),              'C_hmm' ],
               match              => [ LogicBuddy::Singular::Table_matches_logical_function(),                          'C_grin' ]
            );

Readonly my $C_reserved_word_pattern         => LogicBuddy::Logic::reserved_word_pattern();
Readonly my $C_input_variable_pattern        => LogicBuddy::Logic::input_variable_pattern();
Readonly my $C_output_variable_pattern       => LogicBuddy::Logic::output_variable_pattern();
Readonly my $C_comment_pattern               => LogicBuddy::Logic::comment_pattern();
Readonly my $C_reserved_word_pattern_no_case => LogicBuddy::Logic::reserved_word_pattern_no_case();

################################################################################
# global variable
################################################################################

my $G_debug = FALSE;

################################################################################
## class LogicBuddy::Tk::Editor
#################################################################################

sub new {
   my ( $proto, $parent, $balloon, $viewer ) = @_;

   my $class = ref($proto) || $proto;
   my $self;

   $self->{parent}             = $parent;
   $self->{Viewer}             = $viewer;
   $self->{Balloon}            = $balloon;
   $self->{NotOnInserting}     = TRUE;
   $self->{EditorDialog}       = undef;
   $self->{Editor}             = undef;
   $self->{Message}            = undef;
   $self->{buttons}            = [];
   $self->{status}             = EMPTY;
   $self->{status_buttons}     = [];
   $self->{language}           = $LogicBuddy::G_language;
   $self->{solve_button}       = undef;
   $self->{changed_label}      = undef;
   $self->{last_contents}      = EMPTY;
   $self->{logic}              = undef;
   $self->{stack}              = [];
   $self->{unstack}            = [];
   $self->{restore_button}     = undef;
   $self->{forward_button}     = undef;
   $self->{clear_button}       = undef;
   $self->{work_button}        = undef;
   $self->{busy}               = FALSE;

   bless $self, $class;

   return $self;

}; # new

################################################################################
# internal subprograms
################################################################################

################################################################################
# internal Tk subprograms
################################################################################

sub ITk_editor {
   my ( $self ) = @_;

   if ( defined $self->{EditorDialog} and Tk::Exists($self->{EditorDialog}) ) {
      $self->{EditorDialog}->update();
      $self->{EditorDialog}->deiconify();
      $self->{EditorDialog}->raise();
      return;
   };

   my ( $text, $message );
   my $dialog = LogicBuddy::Tk::Utilities::Tk_dialog( $self->{parent}, LogicBuddy::Singular::Editor() );
   $dialog->bind( '<Key-F1>' => sub { LogicBuddy::Tk::Help::editor_help( $self->{parent} ); } );
   $dialog->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::editor_help( $self->{parent} ); } );
   $dialog->minsize( @Tk::Configuration::C_editor_minsize );

   my $buttons = $dialog->Frame()->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );

   my $quit_button = $buttons->Button( -command  => sub { $dialog->iconify() },
                                       -text     => LogicBuddy::Singular::Cancel(),
                                       -image    => Tk::Images::cancel( $dialog ),
                                     )->pack( -side => 'right', %C_Standard_Pads );
   $self->{Balloon}->attach( $quit_button, LogicBuddy::Singular::Close_dialog() );

   my $help_button = $buttons->Button( -command => sub { LogicBuddy::Tk::Help::editor_help( $self->{parent}) },
                                       -text    => LogicBuddy::Singular::Help(),
                                       -image   => Tk::Images::help( $dialog )
                                     )->pack( -side => 'right', %C_Pads );
   $self->{Balloon}->attach( $help_button, LogicBuddy::Singular::Help_Description_for_Balloon() );

   my $message_button = $buttons->Button( -command => sub { $self->{Viewer}->show_message(); },
                                          -text    => LogicBuddy::Plural::Show_message(),
                                          -image   => Tk::Images::viewtext( $dialog )
                                        )->pack( -side => 'right', %C_Pads );
   $self->{Balloon}->attach( $message_button, LogicBuddy::Plural::Show_message() );

   $self->{work_button} = $buttons->Button( -command  => sub { $self->IC_work(); },
                                            -text     => LogicBuddy::Singular::Work(),
                                            -image    => Tk::Images::b_work( $dialog ),
                                          )->pack( -side => 'right', %C_Standard_Pads );
   $self->{Balloon}->attach( $self->{work_button}, LogicBuddy::Singular::Calculate_truth_table_of_logical_function() );
   push @{ $self->{buttons} }, $self->{work_button};

   $self->{clear_button} = $buttons->Button( -command  => sub { $self->IC_clear() },
                                             -text     => LogicBuddy::Singular::Clear(),
                                             -image    => Tk::Images::clear( $dialog ),
                                           )->pack( -side => 'right', %C_Standard_Pads );
   $self->{Balloon}->attach( $self->{clear_button}, LogicBuddy::Singular::Clear_logical_function() );
   push @{ $self->{buttons} }, $self->{clear_button};

   $self->{forward_button} = $buttons->Button( -command  => sub { $self->IC_unpop_text() },
                                               -text     => LogicBuddy::Singular::Restore_forward(),
                                               -image    => Tk::Images::next( $dialog ),
                                               -state    => 'disabled'
                                             )->pack( -side => 'right', %C_Standard_Pads );
   $self->{Balloon}->attach( $self->{forward_button}, LogicBuddy::Singular::Restore_logical_function_forwards() );
   push @{ $self->{buttons} }, $self->{forward_button};

   $self->{restore_button} = $buttons->Button( -command  => sub { $self->IC_pop_text() },
                                               -text     => LogicBuddy::Singular::Restore_backward(),
                                               -image    => Tk::Images::prev( $dialog ),
                                               -state    => 'disabled'
                                             )->pack( -side => 'right', %C_Standard_Pads );
   $self->{Balloon}->attach( $self->{restore_button}, LogicBuddy::Singular::Restore_logical_function_backwards() );
   push @{ $self->{buttons} }, $self->{restore_button};

   $self->I_set_status( 'empty' );
   my $status_button = $self->status_button( $buttons )->pack( -side => 'left', %C_Standard_Pads );
   push @{ $self->{buttons} }, $status_button;

   $self->{changed_label} = $buttons->Label( -textvariable => \$self->{changed},
                                             -wraplength   => $C_length_status,
                                             -justify      => 'left',
                                             -font         => $Tk::Configuration::C_standard_font
                                           )->pack( -side => 'left', %C_Standard_Pads );
   $self->{Balloon}->attach( $self->{changed_label}, LogicBuddy::Singular::Match_between_logical_function_and_table() );
   push @{ $self->{buttons} }, $self->{changed_label};

   $text = $dialog->Scrolled( 'Text',
                              -scrollbars => 'osoe',
                              %Tk::Configuration::C_Text_configuration,
                              -height     => $C_editor_height,
                            )->pack( -side => 'top', -expand => TRUE, -fill => 'both', %C_Standard_Pads );
   $text->Subwidget('xscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $text->Subwidget('yscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $text->menu( undef ); # no popup menu

   # There are class Tk::Text bindings to the subwidget 'text' of the Pane. This removes these bindings to the clipboard
   # See in /path_to/Tk/Text.pm at ClassInit ... 
   $text->Subwidget('text')->bind( 'Tk::Text', '<Key-F1>' => sub { LogicBuddy::Tk::Help::help_editor_window( $self->{parent} ); Tk::break } );
   $text->Subwidget('text')->bind( 'Tk::Text', '<Key-F2>' => sub { LogicBuddy::Tk::Help::help_syntax_rules( $self->{parent} ); } );
   $text->Subwidget('text')->bind( 'Tk::Text', '<Key-F3>' => sub { LogicBuddy::Tk::Help::help_syntax( $self->{parent} ); } );
   $text->Subwidget('text')->bind( 'Tk::Text', '<Key-F4>' => sub { LogicBuddy::Tk::Help::help_syntax_example( $self->{parent} ); } );
   $text->bind( '<Key-F1>' => sub { LogicBuddy::Tk::Help::help_editor_window( $self->{parent} ); Tk::break } );
   $text->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_editor_window( $self->{parent} ); Tk::break } );
   $self->{Balloon}->attach( $text, LogicBuddy::Singular::Enter_and_change_logical_function() );
   push @{ $self->{buttons} }, $text;

   $text->tagConfigure( 'reserved', -foreground => $Tk::Configuration::C_reserved_color );
   $text->tagConfigure( 'input',    -foreground => $Tk::Configuration::C_input_color );
   $text->tagConfigure( 'output',   -foreground => $Tk::Configuration::C_output_color );
   $text->tagConfigure( 'comment',  -foreground => $Tk::Configuration::C_comment_color );
   $text->tagConfigure( 'inserted', -underline  => TRUE );
   $text->tagBind( 'inserted', '<Key>', [ sub { $self->IC_changed_inserted( @_ ) }, Tk::Ev('K') ] );
   $text->tagBind( 'inserted', '<Enter>', sub { $self->{NotOnInserting} = FALSE;  } );
   $text->tagBind( 'inserted', '<Leave>', sub { $self->{NotOnInserting} = TRUE; } );

   $message = $dialog->Scrolled( 'ROText',
                                 -scrollbars => 'osoe',
                                 %Tk::Configuration::C_Text_configuration,
                                 -height     => $C_message_height,
                               )->pack( -side => 'bottom', -fill => 'both', -expand => TRUE, %C_Standard_Pads );
   $message->Subwidget('xscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $message->Subwidget('yscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $message->menu( undef ); # no popup menu
   $message->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_editor_error_window( $self->{parent} ); Tk::break } );
   $self->{Balloon}->attach( $message, LogicBuddy::Plural::Error_message() );

   $self->{EditorDialog} = $dialog;
   $self->{Editor}       = $text;
   $self->{Message}      = $message;

   $self->{logic}        = LogicBuddy::Logic->new( $self );

   $dialog->repeat( $C_reaction_period, sub { $self->IC_changed() } );

   $dialog->protocol( 'WM_DELETE_WINDOW' => sub { $dialog->iconify() } );

   return $dialog;

}; # ITk_editor

################################################################################
# internal callback subprograms
################################################################################

sub IC_changed_inserted {
   my ( $self, $ignore_widget, $keycode ) = @_;

   # The inserted text is tagged by 'inserted'.
   # This call back ist bound to the <Key> and tag 'inserted'.
   # The inserted text will be accepted if it is changed or F5 is pressed.
   #                                            filter keys not changing the text tagged by 'inserted'
   return if $keycode ne 'F5' and
             $keycode =~ m/ \A ( F \d+ |
                                 Left|Right|Down|Up|Insert|Pause|Escape|End|Home|Prior|Next|Num_Lock|Caps_Lock
                                 Shift_.+ | Alt_.+ | Control_.+ | .+_Shift | KP_.+ | Super_.+
                               )
                            \z
                          /xms;

   my $Editor = $self->{Editor};
   my $here   = $Editor->index( 'insert' );

   my ( $start, $end ) = ( '0.0', '0.0' );
   TAG: while ( defined $start ) {
      ( $start, $end ) = $Editor->tagNextrange( 'inserted', $start, 'end' );
      last TAG if not defined $start;
      print "inserted '$keycode' at $start .. $end at '$here'\n" if $G_debug;
      if ( $Editor->compare( $start, '<=', $here ) and $Editor->compare( $end, '>=', $here ) ) {
         $Editor->tagRemove( 'inserted', $start, $end );
      };
      $start = $end;
   }; # TAG

   return;

}; # IC_changed_inserted

sub IC_clear {
   my ( $self ) = @_;

   $self->IC_push_text();
   $self->{Editor}->delete( '0.0', 'end' );
   $self->{Message}->delete( '0.0', 'end' );

}; # IC_clear

sub IC_changed {
   my ( $self ) = @_;

   return if $self->{busy};

   return if not $self->{Editor};

   my $contents          = $self->{Editor}->get( '0.0', 'end' );
   my $contents_is_clear = $contents =~ m/ \A [\s\n] * \s \z /xms;
   if ( $contents_is_clear ) {
      $self->{clear_button}->configure( -state => 'disabled' );
   } else {
      $self->{clear_button}->configure( -state => 'normal' );
   };

   if ( $contents_is_clear ) {
      $self->I_set_status( 'empty' );
   } elsif ( $self->{status} eq 'differ' ) {
   } elsif ( $self->{status} eq 'differ_and_error' ) {
   } elsif ( $self->{last_contents} ne $contents ) {
      $self->I_set_status( 'differ' );
   };

   if ( (not $self->{status} ) or ( $self->{status} eq 'empty' ) ) {
      $self->{Balloon}->attach( $self->{Editor}, LogicBuddy::Singular::Enter_and_change_logical_function() );
   } else {
      $self->{Balloon}->attach( $self->{Editor},
                                LogicBuddy::Singular::Automatic_error_correction( ".\n" ) .
                                LogicBuddy::Singular::Changing_or_F5_accepts( '.' ),
                                -motioncommand => sub { return $self->{NotOnInserting} }
                              );
   };

   $self->I_highlight_words(    $C_input_variable_pattern,  'input' );
   $self->I_highlight_reserved( $C_reserved_word_pattern,   'reserved' );
   $self->I_highlight_comments( $C_comment_pattern,         'comment' );

   my $can_restore =  (scalar @{ $self->{stack} })  ?  'normal'  :  'disabled';
   $self->{restore_button}->configure( -state => $can_restore ); 
   my $can_forward =  (scalar @{ $self->{unstack} })  ?  'normal'  :  'disabled';
   $self->{forward_button}->configure( -state => $can_forward ); 

   return;

}; # IC_changed

sub IC_work {
   my ( $self ) = @_;

   $self->IC_push_text();

   $self->{Editor}->tagRemove( 'inserted', '0.0', 'end' );

   $self->{Message}->delete( '0.0', 'end' );
   $self->{Viewer}->new_message( "\n   ------------------------------\n\n" );

   my $expression = $self->{Editor}->get( '0.0', 'end' );

   if ( $expression =~ m/ \A (\s|\n)* \z /xms ) {
      $self->{Message}->insert( 'end', LogicBuddy::Singular::No_logical_function_defined( '.' ) );
   } else {

      eval {
         local $SIG{__DIE__}  = undef;
         $self->{logic}->define_expression( $expression );
      };

      if ( $@ ) {
         $self->{Message}->insert( 'end', $@ );
         $self->{Viewer}->insert_message( $@ );
         my ( $message, $line, $column ) = $self->{logic}->last_error();
         if ( (defined $line) and (defined $column) ) {
            $self->{Editor}->markSet( 'insert', "$line.$column" );
            $self->{Editor}->see( "$line.$column" );
         };
         $self->I_set_status( 'differ_and_error' );
      } else {
         $self->{parent}->update();
         $self->{parent}->deiconify();
         $self->{parent}->raise();
         $self->{last_contents} = $self->{Editor}->get( '0.0', 'end' );

         $self->{Viewer}->define_input_labels( $self->{logic}->input_names() );
         $self->{Viewer}->define_output_labels( $self->{logic}->output_names() );
         $self->{Viewer}->define_table( $self->{logic}->expression() );
         my ( $has_code, $error, $warning ) = $self->{logic}->status();
         if ( $warning ) {
            $self->I_set_status( 'match_and_warning' );
         } else {
            $self->I_set_status( 'match' );
         };
      };

   };
   return;

}; # IC_work

sub IC_push_text {
   my ( $self ) = @_;

   my $text = $self->{Editor}->get( '0.0', 'end' );

   return if $text !~ m/\S/xms;

   # avoid mutiple versions of same text
   if ( scalar @{ $self->{stack} } ) {
      my $current = $self->{stack}->[-1];
      return if $current eq $text;
   };

   push @{ $self->{stack} }, $text;

   return;

}; # IC_push_text

sub IC_unpop_text {
   my ( $self ) = @_;

   my $text = $self->{Editor}->get( '0.0', 'end' );
   if ( $text =~ m/\S/xms ) {
      $text =~ s/\s+\z//;
      push @{ $self->{stack} }, $text;
   };

   if ( scalar @{ $self->{unstack} } ) {
      $self->{Editor}->delete( '0.0', 'end' );
      $self->{Editor}->insert( 'end', pop @{ $self->{unstack} } );
      $self->{Message}->delete( '0.0', 'end' );
      $self->I_set_status( 'differ' );
   };

   return;

}; # IC_unpop_text

sub IC_pop_text {
   my ( $self ) = @_;

   my $text = $self->{Editor}->get( '0.0', 'end' );
   if ( $text =~ m/\S/xms ) {
      $text =~ s/\s+\z//;
      push @{ $self->{unstack} }, $text;
   };

   if ( scalar @{ $self->{stack} } ) {
      $self->{Editor}->delete( '0.0', 'end' );
      $self->{Editor}->insert( 'end', pop @{ $self->{stack} } );
      $self->{Message}->delete( '0.0', 'end' );
      $self->I_set_status( 'differ' );
   };

   return;

}; # IC_pop_text

################################################################################
# internal subprograms
################################################################################

sub I_set_status {
   my ( $self, $status ) = @_;

   $self->{status}  = $status;
   $self->{changed} = LogicBuddy::Singular::Status( ': ' ) . $C_status{$status}[0] . '.';

   my $file         = Tk::Images::file_of(  $C_status{$status}[1] );
   my $image        = Tk::Images::image_of( $self->{parent}, $file );
   foreach my $button ( @{ $self->{status_buttons} } ) {
      $button->configure( -image => $image );
   };

   return;

}; # I_set_status

sub I_highlight_words {
   my ( $self, $pattern, $tag ) = @_;

   my $Editor = $self->{Editor};

   $Editor->tagRemove( $tag, '0.0', 'end' );

   my $length = 0;
   my $word_index = '0.0';
   my $stop_at    = 'end';

   WORD: while ( $Editor->compare( $word_index, '<=', $stop_at ) ) {
      $word_index = $Editor->search( -count => \$length, -regexp => $pattern, $word_index, $stop_at );
      if ( $word_index ) {
         my $start = $word_index;
         my $end   = $Editor->index( "$word_index + $length chars" );
         my $word  = $Editor->get( $word_index, $end );
         if ( $word =~ $C_reserved_word_pattern_no_case ) {
            $word_index = $end;
            next WORD;
         };
         $start = $Editor->index( "$start + 1 char" ) if $word =~ m/ \A \W /xms;
         $end   = $Editor->index( "$end   - 1 char" ) if $word =~ m/ \W \z /xms;
         $Editor->tagAdd( $tag, $start, $end );
         $word_index = $end;
      } else {
         last WORD;
      };
   }; # WORD

   return;

}; # I_highlight_words

sub I_highlight_comments {
   my ( $self, $pattern, $tag ) = @_;

   my $Editor = $self->{Editor};

   $Editor->tagRemove( $tag, '0.0', 'end' );

   my $length = 0;
   my $word_index = '0.0';
   my $stop_at    = 'end';

   WORD: while ( $Editor->compare( $word_index, '<=', $stop_at ) ) {
      $word_index = $Editor->search( -count => \$length, -regexp => $pattern, $word_index, $stop_at );
      if ( $word_index ) {
         my $end   = $Editor->index( "$word_index + $length chars" );
         $Editor->tagAdd( $tag, $word_index, $end );
         $word_index = $end;
      } else {
         last WORD;
      };
   }; # WORD

   return;

}; # I_highlight_comments

sub I_highlight_reserved {
   my ( $self, $pattern, $tag ) = @_;

   my $Editor = $self->{Editor};

   $Editor->tagRemove( $tag, '0.0', 'end' );

   my $length = 0;
   my $word_index = '0.0';
   my $stop_at    = 'end';

   RESERVED: while ( $Editor->compare( $word_index, '<=', $stop_at ) ) {
      $word_index = $Editor->search( -count => \$length, -regexp => $pattern, $word_index, $stop_at );
      if ( $word_index ) {
         my $start = $word_index;
         my $end   = $Editor->index( "$word_index + $length chars" );
         my $word  = $Editor->get( $word_index, $end );
         $start = $Editor->index( "$start + 1 char" ) if $word =~ m/ \A \W /xms;
         $end   = $Editor->index( "$end   - 1 char" ) if $word =~ m/ \W \z /xms;
         $Editor->tagAdd( $tag, $start, $end );
         $word_index = $end;
      } else {
         last RESERVED;
      };
   }; # RESERVED

   return;

}; # I_highlight_reserved

################################################################################
# exported subprograms
################################################################################

sub status_button {
   my ( $self, $parent ) = @_;

   my $button = $parent->Button( -text     => LogicBuddy::Singular::Status(),
                                 -image    => Tk::Images::C_sleep( $parent ),
                                 -command  => sub { LogicBuddy::Tk::Help::help_status( $self->{parent} ); }
                               );
   $self->{Balloon}->attach( $button, LogicBuddy::Singular::Match_between_logical_function_and_table() );

   push @{ $self->{status_buttons} }, $button;

   return $button;

}; # status_button

sub edit {
   my ( $self ) = @_;
   $self->ITk_editor();
   return;
}; # edit

sub set_language {
   my ( $self, $language ) = @_;

   if ( $language ) {
      $self->{language} = $language;
   };

   return;

}; # set_language

sub language {
   my ( $self ) = @_;
   return $self->{language};
}; # language

sub message_at {
   my ( $self, $message, $line, $column ) = @_;

   return if not $self->{Message};

   $self->ITk_editor();
   if ( $message ) {
      $self->{Message}->insert( 'end', $message );
      $self->{Viewer}->insert_message( $message );
   };

   $self->go_to( $line, $column );

   return;

}; # message_at

sub go_to {
   my ( $self, $line, $column ) = @_;

   if ( defined $line ) {
      if ( defined $column ) {
         $self->{Editor}->markSet( 'insert', "$line.$column" );
         $self->{Editor}->see( "$line.$column" );
      } else {
         $self->{Editor}->markSet( 'insert', $line );
         $self->{Editor}->see( $line );
      };
   };

   return;

}; # go_to

sub changed {
   my ( $self, $message ) = @_;

   return if not $self->{Message};

   $self->{Message}->insert( 'end', $message );

   my $contents = $self->{Editor}->get( '0.0', 'end' );
   if ( $contents =~ m/ \A [\s\n] * \s /xms ) {
      $self->I_set_status( 'differ' );
   };

   return;

}; # changed

sub enter_function {
   my ( $self, $function ) = @_;

   $self->ITk_editor();

   $self->IC_push_text();

   $self->{Message}->delete( '0.0', 'end' );
   $self->{Editor}->delete( '0.0', 'end' );
   $self->{Editor}->insert( 'end', $function );
   $self->{Editor}->see( '0.0' );

   $self->{last_contents} = $self->{Editor}->get( '0.0', 'end' );
   $self->I_set_status( 'match' );

   $self->IC_push_text();

   return;

}; # enter_function

sub insert_at {
   my ( $self, $position, $text ) = @_;
   $self->{Editor}->insert( $position, $text );
   return;
}; # insert_at

sub insert_correction_at {
   my ( $self, $position, $text ) = @_;

   my $my_position = $position eq 'end'  ?  $self->{Editor}->index( "$position - 1 chars" )  :  $self->{Editor}->index( $position );
   $self->{Editor}->insert( $position, $text );

   my $my_end = $self->{Editor}->index( "$position + " . (length $text) . " chars" );
   $self->{Editor}->tagAdd( 'inserted', $my_position, $my_end );

   return;

}; # insert_correction_at

sub busy {
   my ( $self, $busy ) = @_;

   if ( $busy ) {
      foreach my $button ( @{ $self->{buttons} } ) {
         $button->configure( -state => 'disabled' );
      };
      $self->{parent}->configure( -cursor => $Tk::Configuration::C_busy_cursor );
   } else {
      foreach my $button ( @{ $self->{buttons} } ) {
         $button->configure( -state => 'normal' );
      };
      $self->{parent}->configure( -cursor => $Tk::Configuration::C_standard_cursor );
   };
   $self->{busy} = $busy;
   $self->IC_changed();

   return;

}; # busy

################################################################################
# class operations LogicBuddy::Tk::Editor
################################################################################

sub defined_states {
   return %C_status;
}; # defined_states

################################################################################
# end LogicBuddy::Tk::Editor
################################################################################

1;

