=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Tk::Viewer - module

=head1 SYNOPSIS

   use LogicBuddy::Tk::Message;

   $messager = new LogicBuddy::Tk::Message( $parent );

   $messager->message( $message, ... );
   $messager->insert_message( $message, ... );
   $messager->new_message( $message );
   $messager->show_message();
   $messager->solution( $text );

=head1 ABSTRACT

LogicBuddy::Tk::Message abstracts Tk for LogicBuddy.

=head1 DESCRIPTION

=head2 METHODS

=head3 new

   $messager = new LogicBuddy::Tk::Message( $parent );
   instantiates a LogicBuddy Tk messages

=head3 message

   $messager->message( $message, ... );

Puts the messages into the message dialogs.
The message dialog will be raised if iconified.

=head3 insert_message
   $messager->insert_message( $message, ... );

Puts the messages into the message dialogs.
The message dialog will not be raised if iconified.

=head3 new_message

   $messager->new_message( $message );

Remembers the $message.
This massages will be displayed when the next $viewer->message is called.

=head3 show_message

   $messager->show_message();

Raises the message dialog.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;

use Tk;
use Tk::Images;
use Tk::ROText;

use LogicBuddy::Configuration;
use LogicBuddy::Plural;
use LogicBuddy::Singular;
use LogicBuddy::Tk::Configuration;

use LogicBuddy::Tk::Utilities;

use LogicBuddy::Tk::Help;

package LogicBuddy::Tk::Messages;

use English '-no_match_vars';

################################################################################
# global constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

use Readonly;
Readonly my %C_Standard_Pads => %Tk::Configuration::C_Standard_Pads;
Readonly my %C_Pads          => %Tk::Configuration::C_Pads;

################################################################################
# global variable
################################################################################

################################################################################
## class LogicBuddy::Tk::Messages
#################################################################################

sub new {
   my ( $proto, $parent, $balloon ) = @_;

   my $class = ref($proto) || $proto;
   my $self;

   $self->{parent}              = $parent;
   $self->{Balloon}             = $balloon;
   $self->{MessageDialog}       = undef;
   $self->{Message}             = EMPTY;
   $self->{new_message}         = EMPTY;

   bless $self, $class;

   return $self;

}; # new

################################################################################
# internal subprograms
################################################################################

################################################################################
# internal Tk subprograms
################################################################################

sub ITk_message {
   my ( $self, $iconify ) = @_;

   if ( $self->I_see_message_dialog() ) {
      return if $iconify;
      $self->{MessageDialog}->update();
      $self->{MessageDialog}->deiconify();
      $self->{MessageDialog}->raise( $self->{parent} );
      return;
   };

   my $dialog = LogicBuddy::Tk::Utilities::Tk_dialog( $self->{parent}, LogicBuddy::Plural::Message() );
   $dialog->minsize( @Tk::Configuration::C_standard_minsize );

   my $buttons = $dialog->Frame()->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );
   my $text;

   my $quit_button = $buttons->Button( -command  => sub { $dialog->iconify() },
                                       -text     => LogicBuddy::Singular::Cancel(),
                                       -image    => Tk::Images::cancel( $dialog ),
                                     )->pack( -side => 'right', %C_Pads );
   $self->{Balloon}->attach( $quit_button, LogicBuddy::Singular::Close_dialog() );

   my $clear_button = $buttons->Button( -command  => sub { $text->delete( '0.0', 'end' ); $self->{new_message} = EMPTY; },
                                        -text     => LogicBuddy::Singular::Clear(),
                                        -image    => Tk::Images::clear( $dialog ),
                                      )->pack( -side => 'right', %C_Pads );
   $self->{Balloon}->attach( $clear_button, LogicBuddy::Plural::Clear_message() );

   $text = $dialog->Scrolled( 'ROText',
                              -scrollbars => 'osoe',
                              %Tk::Configuration::C_Text_configuration
                            )->pack( -side => 'top', -expand => TRUE, -fill => 'both', %C_Standard_Pads );
   $text->Subwidget('xscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $text->Subwidget('yscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $text->menu( undef ); # no popup menu

   $self->{MessageDialog} = $dialog;
   $self->{Message}       = $text;

   $dialog->bind( '<Key-F1>'   => sub { LogicBuddy::Tk::Help::main_help( $self->{parent} ); Tk::break } );
   $dialog->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::main_help( $self->{parent} ); Tk::break } );
   $dialog->protocol( 'WM_DELETE_WINDOW' => sub { $dialog->iconify() } );

   $dialog->iconify() if $iconify;

   return;

}; # ITk_message

################################################################################
# internal callback subprograms
################################################################################

################################################################################
# internal subprograms
################################################################################

sub I_see_message_dialog {
   my ( $self ) = @_;
   return ( defined $self->{MessageDialog} and Tk::Exists($self->{MessageDialog}) );
}; # I_see_message_dialog

sub I_insert_message {
   my ( $self, @messages ) = @_;

   if ( $self->{new_message} and ($self->{Message}->get( '0.0', 'end' ) =~ m/\S/xms ) ) {
      $self->{Message}->insert( 'end', $self->{new_message} );
   };

   $self->{new_message} = EMPTY;

   foreach my $message ( @messages ) {
      $self->{Message}->insert( 'end', $message );
   };
   $self->{Message}->see( 'end' );

   $self->{MessageDialog}->update();

   return;

}; # I_insert_message

################################################################################
# exported subprograms
################################################################################

sub message {
   my ( $self, @messages ) = @_;
   $self->ITk_message();
   $self->I_insert_message( @messages );
   return;
}; # message

sub insert_message {
   my ( $self, @messages ) = @_;
   $self->ITk_message( TRUE );
   $self->I_insert_message( @messages );
   return;
}; # insert_message

sub new_message {
   my ( $self, $message ) = @_;
   $self->{new_message} = $message if $self->I_see_message_dialog();
   return;
}; # new_message

sub show_message {
   my ( $self ) = @_;
   $self->ITk_message();
   return;
}; # show_message

################################################################################
# end LogicBuddy::Tk::Messages
################################################################################

1;

