=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Tk::Solution - module

=head1 SYNOPSIS

   use LogicBuddy::Tk::Solution;

   $solution  = new LogicBuddy::Tk::Solution( $parent );

               $solution->solution( $text );
               $solution->delete();
   $text     = $solution->get();
               $solution->show();
               $solution->embed( $widget, $help );

=head1 ABSTRACT

LogicBuddy::Tk::Solution abstracts Tk viewing LogicBuddy solutions.

=head1 DESCRIPTION

=head2 METHODS

=head3 new

   $solution = new LogicBuddy::Tk::Solution( $parent );
   instantiates a LogicBuddy solution viewer

=head3 solution

   $solution->solution( $text );

Puts the $text into the solution window.

=head3 delete

   $solution->delete();

Deletes the text in the solution dialog.

=head3 get

   $text = $solution->get();

Returns the text in the solution dialog.

=head3 show

   $solution->show();

Shows the solution in its own window.

=head3 embed

   $solution->embed( $widget, $help );

Embeds the $widget into the solution window and binds the $help command to the solution window.
If $widget is defined the current text in the solution window is saved.
If $widget is not defined the last text in the solution window is restored.
If $help is defined the $help command is bound to the solution window.
If $help is defined the standard help command is bound to the solution window.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;

use Tk;
use Tk::ROText;

use LogicBuddy::Plural;
use LogicBuddy::Singular;

use LogicBuddy::Tk::Help;

package LogicBuddy::Tk::Solution;

use English '-no_match_vars';

################################################################################
# global constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

use Readonly;
Readonly my %C_Standard_Pads => %Tk::Configuration::C_Standard_Pads;
Readonly my %C_solution_size => ( -width => 72,  -height => 6   ); # character

################################################################################
# global variable
################################################################################

################################################################################
## class LogicBuddy::Tk::Solution
#################################################################################

sub new {
   my ( $proto, $viewer, $balloon ) = @_;

   my $class = ref($proto) || $proto;
   my $self;

   $self->{Viewer}              = $viewer;
   $self->{Balloon}             = $balloon;
   $self->{Solution}            = undef;
   $self->{SolutionDialog}      = undef;
   $self->{Solution_in_Dialog}  = undef;
   $self->{previous}            = EMPTY;

   bless $self, $class;

   $self->ITk_solution();

   return $self;

}; # new

################################################################################
# internal subprograms
################################################################################

################################################################################
# internal Tk subprograms
################################################################################

sub ITk_solution {
   my ( $self ) = @_;

   $self->{Solution} = $self->{Viewer}->Scrolled( 'ROText',
                                                  -scrollbars => 'osoe',
                                                  %C_solution_size,
                                                  -wrap       => 'word',
                                                  -relief     => 'ridge',
                                                  %C_Standard_Pads,
                                                  -background => $Tk::Configuration::C_background,
                                                  -font       => $Tk::Configuration::C_bold_fixed_font
                                                );
   $self->{Solution}->Subwidget('xscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $self->{Solution}->Subwidget('yscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $self->{Solution}->menu( undef ); # no popup menu
   $self->{Solution}->bind( '<Key-F1>'   => sub { LogicBuddy::Tk::Help::help_last_solution_window( $self->{Viewer} ); Tk::break } );
   $self->{Solution}->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_last_solution_window( $self->{Viewer} ); Tk::break } );
   $self->{Balloon}->attach( $self->{Solution}, LogicBuddy::Singular::Last_solution() . ".\n" .
                                                LogicBuddy::Singular::Or_use_the_solution_dialog() . '.'
                           );

   return $self->{Solution};

}; # ITk_solution

sub ITk_solution_dialog {
   my ( $self ) = @_;

   if ( defined $self->{SolutionDialog} and Tk::Exists($self->{SolutionDialog}) ) {
      $self->{SolutionDialog}->update();
      $self->{SolutionDialog}->deiconify();
      $self->{SolutionDialog}->raise( $self->{Viewer} );
      return;
   };

   my $dialog = LogicBuddy::Tk::Utilities::Tk_dialog( $self->{Viewer}, LogicBuddy::Plural::Solution() );
   $dialog->minsize( @Tk::Configuration::C_standard_minsize );

   my $text = $dialog->Scrolled( 'ROText',
                                 -scrollbars => 'osoe',
                                 %Tk::Configuration::C_Text_configuration
                               )->pack( -side => 'top', -expand => TRUE, -fill => 'both', %C_Standard_Pads );
   $text->Subwidget('xscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $text->Subwidget('yscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $self->{Balloon}->attach( $text, LogicBuddy::Singular::Last_solution() );
   $text->menu( undef ); # no popup menu

   $self->{SolutionDialog}     = $dialog;
   $self->{Solution_in_Dialog} = $text;

   $dialog->protocol( 'WM_DELETE_WINDOW' => sub { $dialog->iconify() } );

   return;

}; # ITk_solution_dialog

################################################################################
# internal callback subprograms
################################################################################

################################################################################
# internal subprograms
################################################################################

################################################################################
# exported subprograms
################################################################################

sub solution {
   my ( $self, $text ) = @_;

   $self->delete();
   $self->{Solution}->insert( 'end' => $text );
   $self->{Solution_in_Dialog}->insert( 'end' => $text ) if $self->{Solution_in_Dialog};
   return;

}; # solution

sub delete {
   my ( $self ) = @_;
   $self->{Solution}->delete( '0.0' => 'end' );
   $self->{Solution_in_Dialog}->delete( '0.0' => 'end' ) if $self->{Solution_in_Dialog};
   return;
}; # delete

sub get {
   my ( $self ) = @_;
   return $self->{Solution}->get( '0.0' => 'end' );
}; # get

sub pack {
   my ( $self, @parameters ) = @_;
   return $self->{Solution}->pack( @parameters );
}; # pack

sub show {
   my ( $self ) = @_;
   $self->ITk_solution_dialog();
   return;
}; # show

sub embed {
   my ( $self, $widget, $help ) = @_;

   if ( $widget ) {
      $self->{previous} = $self->get();
      $self->delete();
      $self->{Solution}->windowCreate( 'end', -window => $widget );
      $self->{Balloon}->attach( $self->{Solution}, EMPTY );
   } else {
      $self->delete();
      $self->{Solution}->insert( 'end', $self->{previous} );
      $self->{Balloon}->attach( $self->{Solution}, LogicBuddy::Singular::Last_solution() . ".\n" .
                                                   LogicBuddy::Singular::Or_use_the_solution_dialog() . '.'
                              );

   };
   if ( $help ) {
      $self->{Solution}->bind( '<Button-3>' => $help );
      $self->{Solution}->bind( '<Key-F1>'   => $help );
   } else {
      $self->{Solution}->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_last_solution_window( $self->{Viewer} ); Tk::break } );
      $self->{Solution}->bind( '<Key-F1>'   => sub { LogicBuddy::Tk::Help::help_last_solution_window( $self->{Viewer} ); Tk::break } );
   };

   return;

}; # embed

################################################################################
# end LogicBuddy::Tk::Solution
################################################################################

1;

