=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Tk::Table - module

=head1 SYNOPSIS

   use LogicBuddy::Tk::Table;

   $table   = new LogicBuddy::Tk::Table( $parent, $balloon, $viewer );

              $table->pack( ... );
   @labels  = $table->input_labels();
   @labels  = $table->output_labels();
   @values  = $table->values_of();
              $table->define_table( \&subprogram );
              $table->define_input_labels( @input_labels );
              $table->define_output_labels( @output_labels );
              $table->new_table();

=head1 ABSTRACT

LogicBuddy::Tk::Table abstracts displaying LogicBuddy truth tables.

=head1 DESCRIPTION

=head2 METHODS

=head3 new

   $table = new LogicBuddy::Tk::Table( $parent );
   instantiates a LogicBuddy table

=head3 pack

   $table->pack( ... );

Packs the table

=head3 input_labels

   @labels = $table->input_labels();

Returns the list of defined input variable names.

=head3 output_labels

  @labels = $table->output_labels();

Returns the list of defined output variable names.

=head3 values_of
   @values  = $table->values_of();

Returns the values of the current table.
( [ a0, a1, a2 ... ], [ b0, b1, b2 ... ] );

=head3 define_table

   $table->define_table( \&subprogram );

Generates the truth table using &subprogram.

&subprogram is called like &subprogram( $i1, $i2 ... );
$i1 ... are the logical values ( 0 or 1 ) of the input variable in n=1..number_of_inputs

=head3 define_input_labels

   $table->define_input_labels( @input_labels );

Defines new input variable names

=head3 define_output_labels

   $table->define_output_labels( @output_labels );

Defines new output variable names

head3 new_table

              $table->new_table();

Redraws the table using new parameters.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;

use Tk;
use Tk::Pane;
use Tk::Images;

use LogicBuddy::Configuration;
use LogicBuddy::Plural;
use LogicBuddy::Singular;
use LogicBuddy::Utilities;
use LogicBuddy::Tk::Configuration;
use LogicBuddy::Tk::Table::Configuration;

use LogicBuddy::Tk::Utilities;
use LogicBuddy::Tk::Help;

package LogicBuddy::Tk::Table;

use English '-no_match_vars';

################################################################################
# global constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

use Readonly;

Readonly my $C_Border_width       => $Tk::Configuration::C_Border_width;

Readonly my $C_Pad                => $Tk::Configuration::C_Pad;
Readonly my %C_Pads               => %Tk::Configuration::C_Pads;
Readonly my $C_Pad_Standard       => $Tk::Configuration::C_Pad_Standard;
Readonly my %C_Standard_Pads      => %Tk::Configuration::C_Standard_Pads;

Readonly my %C_Table_size         => %LogicBuddy::Tk::Table::C_Table_size;
Readonly my $C_Table_entry_width  => $LogicBuddy::Tk::Table::C_Table_entry_width;

Readonly my %C_standard_number_label_configuration => %LogicBuddy::Tk::Table::C_standard_number_label_configuration;
Readonly my %C_header_number_label_configuration   => %LogicBuddy::Tk::Table::C_header_number_label_configuration;
Readonly my %C_label_configuration                 => %LogicBuddy::Tk::Table::C_label_configuration;
Readonly my %C_I_Grid_Pads                         => %LogicBuddy::Tk::Table::Configuration::C_I_Grid_Pads;

################################################################################
# global variable
################################################################################

################################################################################
## class LogicBuddy::Tk::Viewer
#################################################################################

sub new {
   my ( $proto, $parent, $balloon, $viewer ) = @_;

   my $class = ref($proto) || $proto;
   my $self;

   $self->{parent}              = $parent;
   $self->{Balloon}             = $balloon;
   $self->{Viewer}              = $viewer;
   $self->{Inputs}              = [ 'a', 'b' ];
   $self->{Functions}           = [ 'x1', 'x2' ];
   $self->{Values}              = [ [0, 0], [0, 0], [0, 0], [0, 0] ];
   $self->{ValueButtons}        = [];
   $self->{Pane}                = undef;
   $self->{busy}                = FALSE;
   $self->{waiting}             = undef;

   bless $self, $class;

   $self->ITk_table( $parent );

   return $self;

}; # new

################################################################################
# internal methods
################################################################################

################################################################################
# internal Tk methods
################################################################################

sub ITk_table {
   my ( $self, $parent ) = @_;

   $self->{Header} = $parent->Frame( -background   => $Tk::Configuration::C_background,
                                     -padx         => 0,
                                     -pady         => 0,
                                     -relief       => 'flat',
                                     -borderwidth  => $C_Border_width
                                   )->pack( -side => 'top', -anchor => 'nw', -pady => $C_Pad_Standard );
   $self->{Pane} = $parent->Scrolled( 'Pane',
                                      -scrollbars   => 'oe',
                                      -sticky       => 'nw',
                                      %C_Table_size,
                                      -background   => $Tk::Configuration::C_background,
                                    );
   $self->{Pane}->Subwidget('xscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $self->{Pane}->Subwidget('yscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   LogicBuddy::Tk::Utilities::BindMouseWheel( $self->{Pane} );
   $self->{Pane}->bind( '<Key-F1>'   => sub { LogicBuddy::Tk::Help::help_truth_table_window( $self->{parent} ); Tk::break } );
   $self->{Pane}->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_truth_table_window( $self->{parent} ); Tk::break } );

   $self->IC_insert_table();

   return $self->{Pane};

}; # ITk_table

sub ITk_Row_Label {
   my ( $self, $parent, $number, @digits ) = @_;

   my $row_number_label;
   if ( defined $number ) {
      $row_number_label = $parent->Label( -text => $number,
                                          -relief  => 'groove',
                                          %C_standard_number_label_configuration,
                                          %C_label_configuration
                                        );
   } else {
      $row_number_label = $parent->Label( -image  => Tk::Images::LogicBuddy( $parent ),
                                          -relief => 'groove',
                                          %C_header_number_label_configuration,
                                          %C_label_configuration
                                        );
      $self->{Balloon}->attach( $row_number_label, LogicBuddy::Singular::Row_number_of_table() );
   };

   return $row_number_label;

}; # ITk_Row_Label

sub ITk_Table_Row {
   my ( $self, $parent, $number, @digits ) = @_;

   my $row_number_label = $self->ITk_Row_Label( $parent, $number, @digits );
   $row_number_label->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_truth_table_window( $self->{parent} ); Tk::break } );

   my @row_entries       = ();
   my $index             = 0;
   my %configuration     = %C_label_configuration;
   $configuration{-padx} += 1 if defined $number; # Compensation, the Tk gods know why
   foreach my $digit ( @digits ) {
      if ( $digit =~ m/ \A \d /xms ) {
         my $digit_label  = $parent->Label( -text       => $digit,
                                            -background => $Tk::Configuration::C_background,
                                            -relief     => 'sunken',
                                            %configuration
                                          );
         $digit_label->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_truth_table_window( $self->{parent} ); Tk::break } );
         $self->{Balloon}->attach( $digit_label, LogicBuddy::Singular::Bit_status() );
         LogicBuddy::Tk::Utilities::BindMouseWheel( $digit_label, $self->{Pane} );
         push @row_entries, $digit_label;
      } else {
         my $digit_button = $parent->Button( -text       => $digit,
                                             -background => $Tk::Configuration::C_background,
                                             -relief     => 'sunken',
                                             %configuration,
                                           );
         my $current_index = $index; # closure
         $digit_button->configure( -command => sub { $self->IC_rename_entry( $current_index, $digit_button ); } );
         $digit_button->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_truth_table_window( $self->{parent} ); Tk::break } );
         $self->{Balloon}->attach( $digit_button, LogicBuddy::Singular::Change_input_variable_name() );
         LogicBuddy::Tk::Utilities::BindMouseWheel( $digit_button, $self->{Pane} );
         push @row_entries, $digit_button;
      };
      $index++;
   };

   foreach my $function_index ( 0 .. scalar @{ $self->{Functions} } - 1 ) {

      my $command;
      my $note;
      my $result       = defined $number  ?  $self->{Values}[$number][$function_index]  :  $self->{Functions}[$function_index];

      my $value_button = $parent->Button( -text   => $result,
                                          -relief => 'raised',
                                          %C_label_configuration,
                                        );
      $value_button->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_truth_table_window( $self->{parent} ); Tk::break } );
      LogicBuddy::Tk::Utilities::BindMouseWheel( $value_button, $self->{Pane} );
      if ( defined $number ) {
         $value_button->configure( -command => sub { $self->IC_toggle_value( $number, $function_index ); } );
         $note    = LogicBuddy::Singular::Current_value_Toggle_0_1_d();
         $self->{ValueButtons}[$number][$function_index] = $value_button;
      } else {
         $value_button->configure( -command => sub { $self->IC_rename_function( $value_button, $function_index ); } );
         $note    = LogicBuddy::Singular::Rename_function();
      };
      $self->{Balloon}->attach( $value_button, $note );

      push @row_entries, $value_button;

   };

   $row_number_label->grid( @row_entries, -sticky => 'nwse', %C_I_Grid_Pads );

   return;

}; # ITk_Table_Row

sub ITk_get_name {
   my ( $self, $preset, $pattern, $message, $description, $help ) = @_;

   my $dialog = LogicBuddy::Tk::Utilities::Tk_dialog( $self->{parent}, LogicBuddy::Singular::Enter_name() );

   my $info = $dialog->Label( -text    => $description,
                              -justify => 'left'
                            )->pack( -side => 'top', -fill => 'x', -expand => TRUE, %C_Standard_Pads );

   my $name    = $preset;

   my $entry   = $dialog->Entry( -textvariable => \$name
                               )->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );

   my $buttons = $dialog->Frame( -relief      => 'groove',
                                 -borderwidth => $C_Border_width
                               )->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );

   my $ok_button   = $buttons->Button( -command  => sub { $self->IC_check_name( $dialog, $name, $preset, $pattern, $message, $info ); },
                                       -text     => 'Ok',
                                       -image    => Tk::Images::ok( $dialog ),
                                       -compound => 'left',
                                     )->pack( -side => 'left', %C_Pads );

   my $quit_button = $buttons->Button( -command  => sub { $name = undef; $dialog->destroy() if Tk::Exists($dialog); },
                                       -text     => LogicBuddy::Singular::Cancel(),
                                       -image    => Tk::Images::cancel( $dialog ),
                                       -compound => 'left',
                                     )->pack( -side => 'left', %C_Pads );

   $dialog->bind( '<Key-Escape>' => sub { $name = undef; $dialog->destroy() if Tk::Exists($dialog); } );
   $dialog->bind( '<Key-F1>'     => sub { &$help( $dialog ); Tk::break } );
   $dialog->bind( '<Button-3>'   => sub { &$help( $dialog ); Tk::break } );
   $dialog->bind( '<Key-Return>' => sub { $self->IC_check_name( $dialog, $name, $preset, $pattern, $message, $info ); } );

   $dialog->protocol( 'WM_DELETE_WINDOW' => sub { $name = undef; $dialog->destroy() if Tk::Exists($dialog); } );

   $dialog->waitWindow();

   return $name;

}; # ITk_get_name

sub ITk_new_table {
   my ( $self ) = @_;

   Readonly my $C_with_of_item_description => 20; # char
   Readonly my $C_with_of_item_entry       =>  5; # char

   my $dialog = LogicBuddy::Tk::Utilities::Tk_dialog( $self->{parent}, LogicBuddy::Singular::Table() );

   $dialog->Label( -text  => LogicBuddy::Singular::Define_new_table(),
                 )->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );

   my $entries = scalar @{ $self->{Inputs} };

   my $entries_frame = $dialog->Frame( -relief      => 'groove',
                                       -borderwidth => $C_Border_width
                                     )->pack( -side => 'top', -fill => 'x', -expand => TRUE, %C_Standard_Pads );
   $entries_frame->Label( -text        => LogicBuddy::Plural::Input_variable(),
                          -width       => $C_with_of_item_description,
                          -anchor     => 'w',
                        )->pack( -side => 'left', %C_Standard_Pads );
   my $inputs_entry   = $entries_frame->Label( -textvariable => \$entries,
                                               -width       => $C_with_of_item_entry,
                                               -relief      => 'sunken',
                                               -borderwidth => $C_Border_width
                                             )->pack( -side => 'left', -fill => 'x', -expand => TRUE, %C_Standard_Pads );
   $self->{Balloon}->attach( $inputs_entry, LogicBuddy::Plural::Entry() );
   my $entries_down_button = $entries_frame->Button( -text => '<',
                                                     -image => Tk::Images::prev( $dialog ),
                                                     -relief => 'flat',
                                                     -command => sub { $entries-- if $entries > 2 }
                                                   )->pack( -side => 'left', %C_Pads );
   $self->{Balloon}->attach( $entries_down_button, LogicBuddy::Singular::One_less() );
   my $entries_up_button = $entries_frame->Button( -text    => '>',
                                                   -image   => Tk::Images::next( $dialog ),
                                                   -relief  => 'flat',
                                                   -command => sub { $entries++ if $entries < 7 }
                                                 )->pack( -side => 'left', %C_Pads );
   $self->{Balloon}->attach( $entries_up_button, LogicBuddy::Singular::One_more() );

   my $outputs = scalar @{ $self->{Functions} };

   my $outputs_frame = $dialog->Frame( -relief      => 'groove',
                                       -borderwidth => $C_Border_width
                               )->pack( -side => 'top', -fill => 'x', -expand => TRUE, %C_Standard_Pads );
                         $outputs_frame->Label( -text        => LogicBuddy::Plural::Function(),
                                                -width       => $C_with_of_item_description,
                                                -anchor     => 'w'
                                              )->pack( -side => 'left', %C_Standard_Pads );
   my $outputs_entry   = $outputs_frame->Label( -textvariable => \$outputs,
                                                -width       => $C_with_of_item_entry,
                                                -relief      => 'sunken',
                                                -borderwidth => $C_Border_width
                                              )->pack( -side => 'left', -fill => 'x', -expand => TRUE, %C_Standard_Pads );
   $self->{Balloon}->attach( $outputs_entry, LogicBuddy::Plural::Function() );
   my $outputs_down_button = $outputs_frame->Button( -text => '<',
                                                     -image => Tk::Images::prev( $dialog ),
                                                     -relief => 'flat',
                                                     -command => sub { $outputs-- if $outputs > 1 }
                                                   )->pack( -side => 'left', %C_Pads );
   $self->{Balloon}->attach( $outputs_down_button, LogicBuddy::Singular::One_less() );
   my $outputs_up_button = $outputs_frame->Button( -text    => '>',
                                                   -image   => Tk::Images::next( $dialog ),
                                                   -relief  => 'flat',
                                                   -command => sub { $outputs++ if $outputs < 5 }
                                                 )->pack( -side => 'left', %C_Pads );
   $self->{Balloon}->attach( $outputs_up_button, LogicBuddy::Singular::One_more() );

   my $buttons = $dialog->Frame( -relief      => 'groove',
                                 -borderwidth => $C_Border_width
                               )->pack( -side => 'top', -fill => 'x', %C_Pads );

   my $quit_button = $buttons->Button( -command  => sub { $entries = undef; $dialog->destroy() if Tk::Exists($dialog); },
                                       -text     => LogicBuddy::Singular::Cancel(),
                                       -image    => Tk::Images::cancel( $dialog ),
                                       -compound => 'left',
                                     )->pack( -side => 'left', %C_Pads );
   $self->{Balloon}->attach( $quit_button, LogicBuddy::Singular::Cancel() );

   my $ok_button = $buttons->Button( -command  => sub { $dialog->destroy() if Tk::Exists($dialog); },
                                     -text     => 'Ok',
                                     -image    => Tk::Images::ok( $dialog ),
                                     -compound => 'left',
                                   )->pack( -side => 'left', %C_Pads );
   $self->{Balloon}->attach( $ok_button, LogicBuddy::Singular::Ok() );

   $dialog->bind( '<Key-Escape>' => sub { $entries = undef; $outputs = undef; $dialog->destroy() if Tk::Exists($dialog); } );
   $dialog->bind( '<Key-F1>'     => sub { LogicBuddy::Tk::Help::help_define_new_table( $self->{parent} ); Tk::break } );
   $dialog->bind( '<Button-3>'   => sub { LogicBuddy::Tk::Help::help_define_new_table( $self->{parent} ); Tk::break } );
   $dialog->protocol( 'WM_DELETE_WINDOW' => sub { $entries = undef; $outputs = undef; $dialog->destroy() if Tk::Exists($dialog); } );

   $dialog->waitWindow();

   return ( $entries, $outputs );

}; # ITk_new_table

sub ITk_wait {
   my ( $self ) = @_;
   $self->{waiting} = LogicBuddy::Tk::Utilities::working_dialog( LogicBuddy::Singular::Wait_please(),
                                                                 LogicBuddy::Singular::The_creation_of_the_table_lasts( ' ...' )
                                                               );
   return;
}; # ITk_wait

################################################################################
# internal callback methods
################################################################################

sub IC_rename_function {
   my ( $self, $button, $function_index ) = @_;

   if ( $self->{busy} ) {
      $self->{parent}->bell();
      return;
   };

   my $name = $self->ITk_get_name( $self->{Functions}[$function_index],
                                   LogicBuddy::Logic::output_variable_pattern(),
                                   LogicBuddy::Singular::NAME(),
                                   LogicBuddy::Singular::Enter_name() . "\n" . LogicBuddy::Singular::starting_with_9( 'a-z' ),
                                   \&LogicBuddy::Tk::Help::help_change_variable_name
                                 );

   if ( $name ) {
      $self->{Functions}[$function_index] = $name;
      $button->configure( -text => $name );
      $self->{Viewer}->changed();
      $self->IC_solve(); 
   };

   return;

}; # IC_rename_function

sub IC_rename_entry {
   my ( $self, $index, $button ) = @_;

   if ( $self->{busy} ) {
      $self->{parent}->bell();
      return;
   };

   my $name = $self->ITk_get_name( $self->{Inputs}[$index],
                                   LogicBuddy::Logic::input_variable_pattern(),
                                   LogicBuddy::Singular::NAME(),
                                   LogicBuddy::Singular::Enter_name() . "\n" . LogicBuddy::Singular::starting_with_9( 'a-z' ),
                                   \&LogicBuddy::Tk::Help::help_change_variable_name
                                 );

   if ( $name ) {
      $self->{Inputs}[$index] = $name;
      $button->configure( -text => $name );
      $self->{Viewer}->changed();
      $self->IC_solve();
   };

   return;

}; # IC_rename_entry

sub IC_toggle_value {
   my ( $self, $index, $function_index ) = @_;

   if ( $self->{busy} ) {
      $self->{parent}->bell();
      return;
   };

   my $current = $self->{Values}[$index][$function_index];

   if ( not defined $current ) {
      $current = 0;
   } elsif ( $current eq '1' ) {
      $current = 'd';
   } elsif ( $current eq 'd' ) {
      $current = '0';
   } else {
      $current = 1;
   };

   $self->{Values}[$index][$function_index] = $current;
   $self->{ValueButtons}[$index][$function_index]->configure( -text => $current );
   $self->{Viewer}->changed();
   $self->IC_solve();

   return;

}; # IC_toggle_value

sub IC_insert_table {
   my ( $self, $is_defined ) = @_;

   Readonly my $C_show_wating_info_if_more => 4;

   $self->ITk_Table_Row( $self->{Header}, undef, @{ $self->{Inputs} } );

   my $size = scalar @{ $self->{Inputs} };
   $self->{Table} = $self->{Pane}->Frame( -relief       => 'flat',
                                          -borderwidth  => $C_Border_width,
                                          -background   => $Tk::Configuration::C_background,
                                          -padx         => 8, # Trial and error
                                          -pady         => 0,
                                        )->pack( -side => 'top', -fill => 'both', -expand => TRUE, -padx => 2, -pady => 2 );
   $self->{Table}->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_truth_table_window( $self->{parent} ); Tk::break } );

   $self->{ValueButtons} = [];

   $self->ITk_wait() if scalar @{ $self->{Inputs} } > $C_show_wating_info_if_more;

   foreach my $number ( 0 .. (2**$size)-1 ) {
      my @digits = LogicBuddy::Utilities::bits_of( $number, $size );
      if ( not $is_defined ) {
         foreach my $function_index ( 0 .. scalar @{ $self->{Functions} } - 1 ) {
            $self->{Values}[$number][$function_index] = 0;
         };
      };
      $self->ITk_Table_Row( $self->{Table}, $number, @digits );
   };

   $self->{Table}->update();

   $self->{waiting}->destroy() if $self->{waiting} and Tk::Exists( $self->{waiting} );
   $self->{waiting} = undef;

   $self->{Viewer}->changed();

   return;

}; # IC_insert_table

sub IC_check_name {
   my ( $self, $dialog, $name, $preset, $pattern, $message, $info ) = @_;

   if ( (defined $name) and ( $name =~ m/\A$pattern\z/ ) ){
      if ( $name eq $preset ) {
         $dialog->destroy() if Tk::Exists($dialog);
      } elsif ( $LogicBuddy::C_max_name_length < length $name ) {
          $info->configure( -text => LogicBuddy::Singular::name_9_is_too_long_8( "'$name'", " (>$LogicBuddy::C_max_name_length)" ) ); 
      } else {
         foreach my $known ( @{ $self->{Inputs} }, @{ $self->{Functions} } ) {
            if ( $name eq $known ) {
               $info->configure( -text => LogicBuddy::Singular::Illegal_name() . "\n" .
                                          LogicBuddy::Singular::Name_already_in_use()
                               );
               return;
            };
         };
         $dialog->destroy() if Tk::Exists($dialog);
      };
   } else {
      $info->configure( -text => LogicBuddy::Singular::Illegal_name() . "\n$message" );
   };

   return;

}; # IC_check_name

sub IC_solve {
   my ( $self ) = @_;
   $self->{Viewer}->solve();
   return;
}; # IC_solve

################################################################################
# internal methods
################################################################################

sub I_take_logic_value_of {
   my ( $self, $value ) = @_;

   return FALSE if not $value;
   return $value;

}; # I_take_logic_value_of

################################################################################
# exported methods
################################################################################

sub pack {
   my ( $self, @parameters ) = @_;
   return $self->{Pane}->pack( @parameters );
}; # pack

sub input_labels {
   my ( $self ) = @_;
   return( @{ $self->{Inputs} } );
}; # input_labels

sub output_labels {
   my ( $self ) = @_;
   return( @{ $self->{Functions} } );
}; # output_labels

sub values_of {
   my ( $self, $input_index ) = @_;
   return @{ $self->{Values}[$input_index] };
}; # values_of

sub define_table {
   my ( $self, $command ) = @_;

   LogicBuddy::Tk::Utilities::ungrid_all_children( $self->{Table} );
   LogicBuddy::Tk::Utilities::unpack_all_children( $self->{Table} );
   foreach my $child ( $self->{Header}->gridSlaves() ) {
      LogicBuddy::Tk::Utilities::ungrid_all_children( $child );
   };

   my $size = (scalar @{ $self->{Inputs} });
   foreach my $number ( 0 .. (2**$size) - 1 ) {
      my @bits = LogicBuddy::Utilities::bits_of( $number, $size );
      my @values = map { $self->I_take_logic_value_of( $_ ) } &$command( @bits );
      $self->{Values}[$number] = [ @values ];
   };

   $self->IC_insert_table( TRUE );

   $self->IC_solve();

   return;

}; # define_table

sub define_input_labels {
   my ( $self, @input_labels ) = @_;
   $self->{Inputs} = [ @input_labels ];
   return;
}; # define_input_labels

sub define_output_labels {
   my ( $self, @output_labels ) = @_;
   $self->{Functions} = [ @output_labels ];
   return;
}; # define_output_labels

sub new_table {
   my ( $self ) = @_;

   my ( $size, $function_size ) = $self->ITk_new_table();

   return if not $size;

   my @inputs = map { chr( ord( 'a' ) + $_ ) } ( 0 .. $size - 1 );
   my @functions = map { "x$_" } ( 1 .. $function_size );

   LogicBuddy::Tk::Utilities::ungrid_all_children( $self->{Table} );
   LogicBuddy::Tk::Utilities::unpack_all_children( $self->{Table} );
   foreach my $child ( $self->{Header}->gridSlaves() ) {
      LogicBuddy::Tk::Utilities::ungrid_all_children( $child );
   };

   $self->{Inputs}            = [ @inputs ];
   $self->{Functions}         = [ @functions ];
   $self->{Values}            = [];
   $self->{ValueButtons}      = [];

   $self->IC_insert_table();

   return;

}; # new_table

################################################################################
# end LogicBuddy::Tk::Table
################################################################################

1;

