=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Tk::Table::Multiple - module

=head1 SYNOPSIS

   use LogicBuddy::Tk::Table::Multiple;

   $table  = new LogicBuddy::Tk::Table::Multiple( $parent, $balloon, $viewer );

=head1 ABSTRACT

LogicBuddy::Tk::Table::Multiple abstracts displaying LogicBuddy truth tables and changing multiple entries.

The numbering elements of a truth table invokes a dialog changing multiple entries.

=head1 DESCRIPTION

See LogicBuddy::Tk::Table

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;

use Tk;
use Tk::Images;

use LogicBuddy::Configuration;
use LogicBuddy::Plural;
use LogicBuddy::Singular;
use LogicBuddy::Tk::Configuration;
use LogicBuddy::Tk::Table::Configuration;

use LogicBuddy::Tk::Utilities;
use LogicBuddy::Tk::Help;

package LogicBuddy::Tk::Table::Multiple;

use English '-no_match_vars';

################################################################################
# global constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

use Readonly;
Readonly my $C_Pad                                 => $Tk::Configuration::C_Pad;
Readonly my %C_Pads                                => %Tk::Configuration::C_Pads;
Readonly my $C_Pad_Standard                        => $Tk::Configuration::C_Pad_Standard;
Readonly my %C_Standard_Pads                       => %Tk::Configuration::C_Standard_Pads;

Readonly my $C_Table_entry_width                   => $LogicBuddy::Tk::Table::C_Table_entry_width;

Readonly my %C_standard_number_label_configuration => %LogicBuddy::Tk::Table::C_standard_number_label_configuration;
Readonly my %C_header_number_label_configuration   => %LogicBuddy::Tk::Table::C_header_number_label_configuration;
Readonly my %C_label_configuration                 => %LogicBuddy::Tk::Table::C_label_configuration;
Readonly my %C_I_Grid_Pads                         => %LogicBuddy::Tk::Table::Configuration::C_I_Grid_Pads;

################################################################################
# global variable
################################################################################

################################################################################
## class LogicBuddy::Tk::Table::Multiple
#################################################################################

use base "LogicBuddy::Tk::Table";

sub new {
   my ( $class, $parent, $balloon, $viewer ) = @_;

   my $self;
   $self = $class->SUPER::new( $parent, $balloon, $viewer );

   $self->{complex_digits}      = [];
   $self->{complex_values}      = [];

   return $self;

}; # new

################################################################################
# internal methods
################################################################################

################################################################################
# polymorphic methods
################################################################################

################################################################################
# internal Tk methods
################################################################################

sub ITk_Row_Label {
   my ( $self, $parent, $number, @digits ) = @_;

   Readonly my $C_fit_headline_width => ( $^O eq 'MSWin32' )  ?  54  : 38; # Trial and error matching the truth table

   my $row_number_label;

   if ( defined $number ) {
      $row_number_label = $parent->Button( -text    => $number,
                                           -command => sub { $self->IC_complex_row( $number, @digits ) },
                                           %C_standard_number_label_configuration,
                                           %C_label_configuration
                                         );
      $self->{Balloon}->attach( $row_number_label,
                                LogicBuddy::Singular::Row_number_of_table() . "\n" .
                                LogicBuddy::Singular::Change_multiple_table_entries()
                              );
   } else {
      $row_number_label = $parent->Button( -image   => Tk::Images::LogicBuddy( $parent ),
                                           -command => sub { $self->IC_complex_row( $number, map { 'A' } @digits ) },
                                           %C_header_number_label_configuration,
                                           %C_label_configuration,
                                           -width   => $C_fit_headline_width
                                         );
      $self->{Balloon}->attach( $row_number_label,
                                LogicBuddy::Singular::Change_multiple_table_entries()
                              );
   };

   return $row_number_label;

}; # ITk_Row_Label

sub ITk_complex_row_dialog {
   my ( $self, $number, @digits ) = @_;

   Readonly my $C_info_wrap_length => $Tk::Configuration::C_standard_width - 50;

   $self->{complex_digits} = [ @digits ];
   if ( defined $number ) {
      $self->{complex_values} = [ @{ $self->{Values}[$number] } ];
   } else {
      $self->{complex_values} = [ map { FALSE } @{ $self->{Values}[0] } ];
   };

   my $dialog = $self->{parent}->Frame();
   $dialog->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );

   my $buttons = $dialog->Frame( -background => $Tk::Configuration::C_background
                               )->pack( -side => 'top', -expand => TRUE, -fill => 'x' );
   $buttons->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );
   my $header = $buttons->Label( -text       => LogicBuddy::Singular::Change_multiple_table_entries() . '     ',
                                 -background => $Tk::Configuration::C_background,
                                 -font       => $Tk::Configuration::C_heading_font
                               )->pack( -side => 'left', -expand => TRUE, -fill => 'x' );
   $header->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );

   my $quit_button = $buttons->Button( -command    => sub { $self->IC_complex_exit( $dialog, FALSE ) },
                                       -text       => LogicBuddy::Singular::Cancel(),
                                       -image      => Tk::Images::cancel( $dialog ),
                                       -background => $Tk::Configuration::C_background
                                     )->pack( -side => 'right', %C_Pads );
   $self->{Balloon}->attach( $quit_button, LogicBuddy::Plural::Reject_change() );
   $quit_button->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );

   my $ok_button = $buttons->Button( -command    => sub { $self->IC_complex_exit( $dialog, TRUE ) },
                                     -text       => LogicBuddy::Singular::OK(),
                                     -image      => Tk::Images::ok( $dialog ),
                                     -background => $Tk::Configuration::C_background
                                   )->pack( -side => 'right', %C_Pads );
   $self->{Balloon}->attach( $ok_button, LogicBuddy::Plural::Apply_change() );
   $ok_button->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );

   my $info = $dialog->Label( -text       => LogicBuddy::Singular::multiple_change_short(),
                              -background => $Tk::Configuration::C_background,
                              -height     => 2,
                              -anchor     => 'nw',
                              -justify    => 'left',
                              -wraplength => $C_info_wrap_length
                            )->pack( -side => 'top', -fill => 'x' );
   $info->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );

   my $frame = $dialog->Frame( -background => $Tk::Configuration::C_background
                             )->pack( -side => 'top', -anchor => 'nw', -expand => TRUE, -fill => 'x' );
   $frame->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );
   my $entries = $frame->Frame( -background => $Tk::Configuration::C_background
                               )->pack( -side => 'top', -anchor => 'nw', -expand => TRUE, -padx => 5, -pady => 5 );
   $entries->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );

   my $image = $entries->Label( -image  => Tk::Images::s_change( $entries ),
                                %C_header_number_label_configuration,
                                %C_label_configuration
                              );
   $image->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );

   my @row = ( $image );
   foreach my $input ( @{ $self->{Inputs} } ) {
      my $label = $entries->Label( -text         => $input,
                                   -background   => $Tk::Configuration::C_background,
                                   -relief       => 'sunken',
                                   %C_label_configuration
                                 );
      $self->{Balloon}->attach( $label, LogicBuddy::Singular::Input() );
      $label->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );
      push @row, $label;
   };
   foreach my $function ( @{ $self->{Functions} } ) {
      my $label = $entries->Label( -text   => $function,
                                   -relief => 'groove',
                                   %C_label_configuration
                                 );
      $self->{Balloon}->attach( $label, LogicBuddy::Singular::Function() );
      $label->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );
      push @row, $label;
   };

   my $first = shift @row;
   $first->grid( @row, -sticky => 'nwse', %C_I_Grid_Pads );

   my $row_number = $entries->Label( -text   => $number || '?',
                                     %C_standard_number_label_configuration,
                                     %C_label_configuration,
                                   );
   $self->{Balloon}->attach( $row_number, LogicBuddy::Singular::Row_number_of_table() );
   $row_number->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );

   @row           = (  $row_number );
   my $index      = 0;
   foreach my $digit ( @digits ) {
      my $my_index     = $index; # closure
      my $digit_button = $entries->Button( -text       => $digit,
                                           -background => $Tk::Configuration::C_background,
                                           -relief     => 'sunken',
                                           %C_label_configuration
                                         );
      $digit_button->configure( -command => sub { $self->IC_toggle_complex_digit( $digit_button, $row_number, $my_index ); } );
      $digit_button->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );
      $self->{Balloon}->attach( $digit_button, LogicBuddy::Singular::Change_cyclic_9( '0 - 1 - X' ) );
      $index++;
      push @row, $digit_button;
   };
   foreach my $function_index ( 0 .. scalar @{ $self->{Functions} } - 1 ) {

      my $result       = $self->{complex_values}[$function_index];
      my $value_button = $entries->Button( -text   => $result,
                                           -relief => 'raised',
                                           %C_label_configuration
                                         );
      $value_button->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break } );
      LogicBuddy::Tk::Utilities::BindMouseWheel( $value_button, $self->{Pane} );
      $value_button->configure( -command => sub { $self->IC_toggle_complex_value( $value_button, $function_index ); } );
      $self->{Balloon}->attach( $value_button, LogicBuddy::Singular::Change_cyclic_9( '0 - 1 - d - X' ) );
      push @row, $value_button;
   };

   $first = shift @row;
   $first->grid( @row, -sticky => 'nwse', %C_I_Grid_Pads );

   $self->{Viewer}->busy( TRUE );
   $self->{busy} = TRUE;

   $self->{Viewer}->embed_into_solution( $dialog,
                                         sub { LogicBuddy::Tk::Help::help_change_multiple( $self->{parent} ); Tk::break }
                                       );

   return;

}; # ITk_complex_row_dialog

################################################################################
# internal callback methods
################################################################################

sub IC_complex_row {
   my ( $self, $number, @digits ) = @_;

   if ( $self->{busy} ) {
      $self->{parent}->bell();
      return;
   };

   $self->ITk_complex_row_dialog( $number, @digits );
   return;

}; # IC_complex_row

sub IC_complex_exit {
   my ( $self, $dialog, $ok ) = @_;

   $dialog->destroy();
   $self->{Viewer}->embed_into_solution();
   if ( $ok ) {
      $self->I_insert_new_values();
      $self->IC_solve();
   };
   $self->I_unmark_table();
   $self->{busy} = FALSE;
   $self->{Viewer}->busy( FALSE );
   return;

}; # IC_complex_exit

sub IC_toggle_complex_digit {
   my ( $self, $digit_button, $row_number, $index ) = @_;

   my $current = $self->{complex_digits}[$index];
   if ( not defined $current ) {
      $current = 0;
   } elsif ( $current eq '1' ) {
      $current = 'A';
   } elsif ( $current eq 'A' ) {
      $current = '0';
   } else {
      $current = 1;
   };
   $self->{complex_digits}[$index] = $current;
   $digit_button->configure( -text => $current );

   my $number = 0;
   BIT: foreach my $bit ( @{ $self->{complex_digits} } ) {
      if ( $bit eq 'A' ) {
         $number = '?';
         last BIT;
      } else {
         $number += $number + $bit;
      };
   }; # BIT
   $row_number->configure( -text => $number );

   $self->I_mark_rows();

   return;

}; # IC_toggle_complex_digit

sub IC_toggle_complex_value {
   my ( $self, $value_button, $function_index ) = @_;

   my $current = $self->{complex_values}[$function_index];
   if ( not defined $current ) {
      $current = 0;
   } elsif ( $current eq '1' ) {
      $current = 'd';
   } elsif ( $current eq 'd' ) {
      $current = 'X';
   } elsif ( $current eq 'X' ) {
      $current = '0';
   } else {
      $current = 1;
   };
   $self->{complex_values}[$function_index] = $current;
   $value_button->configure( -text => $current );

   $self->I_mark_rows();

   return;

}; # IC_toggle_complex_value

################################################################################
# internal methods
################################################################################

sub I_mark_row {
   my ( $self, $row, $selected ) = @_;

   my $color = $selected  ?  'red'  :  'black';
   foreach my $widget ( $self->{Table}->gridSlaves( -row => $row ) ) {
      $widget->configure( -foreground => $color );
   };

   my $number_of_functions = scalar @{ $self->{complex_values} };
   if ( $selected ) {
      foreach my $function_index ( 0 .. $number_of_functions - 1 ) {
         $self->I_mark_function( $row, $function_index, $self->{complex_values}[$function_index] );
      };
   } else {
      foreach my $function_index ( 0 .. $number_of_functions - 1 ) {
         $self->I_mark_function( $row, $function_index );
      };
   };

   return;

}; # I_mark_row

sub I_mark_function {
   my ( $self, $row, $function_index, $value ) = @_;

   my @widgets = $self->{Table}->gridSlaves( -row    => $row,
                                             -column => $function_index + scalar @{ $self->{Inputs} } + 1
                                           );
   my $entry   = $widgets[0];

   if ( not defined $value ) {
      $entry->configure( -foreground => 'black', -text => $self->{Values}[$row][$function_index] );
   } elsif ( $value eq 'X' ) {
      $entry->configure( -foreground => 'black', -text => $self->{Values}[$row][$function_index] );
   } elsif ( $value eq $self->{Values}[$row][$function_index] ) {
      $entry->configure( -foreground => 'black', -text => $value );
   } else {
      $entry->configure( -foreground => 'red', -text => $value );
   };
   return;

}; # I_mark_function

sub I_mark_rows {
   my ( $self ) = @_;

   $self->I_unmark_table();
   foreach my $index ( $self->I_select_rows( @{ $self->{complex_digits} } ) ) {
      $self->I_mark_row( $index, TRUE );
   };
   return;

}; # I_mark_rows

sub I_insert_new_values {
   my ( $self ) = @_;

   my $number_of_functions = scalar @{ $self->{complex_values} };
   foreach my $row ( $self->I_select_rows( @{ $self->{complex_digits} } ) ) {
      foreach my $function_index ( 0 .. $number_of_functions - 1 ) {
         $self->I_mark_function( $row, $function_index, $self->{complex_values}[$function_index] );
         my $value = $self->{complex_values}[$function_index];
         if ( $value ne 'X' ) {
            $self->{Values}[$row][$function_index] = $value;
         };
      };
   };
   return;

}; # I_insert_new_values

sub I_unmark_table {
   my ( $self ) = @_;

   my $number_of_inputs = scalar @{ $self->{Inputs} };
   foreach my $index ( 0 .. 2**$number_of_inputs - 1 ) {
      $self->I_mark_row( $index, FALSE );
   };
   return;

}; # I_unmark_table

sub I_select_rows;
sub I_select_rows {
   my ( $self, @digits ) = @_;

   if ( scalar @digits ) {
      my $bit = pop @digits;
      if ( ($bit eq 0) or ($bit eq 1) ) {
         my @new = $self->I_select_rows( @digits );
         foreach my $value ( @new ) {
            $value = $bit + 2 * $value;
         };
         return @new;
      } else {
         return( $self->I_select_rows( @digits, 1 ), $self->I_select_rows( @digits, 0 ) );
      };
   } else {
      return( 0 );
   };

}; # I_select_rows

################################################################################
# exported methods
################################################################################

################################################################################
# end LogicBuddy::Tk::Table::Multiple
################################################################################

1;

