=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Tk::Utilities - module

=head1 SYNOPSIS

use LogicBuddy::Tk::Utilities;

LogicBuddy::Tk::Utilities::BindMouseWheel( $widget );
LogicBuddy::Tk::Utilities::BindMouseWheel( $widget, $target );
LogicBuddy::Tk::Utilities::UnBindMouseWheel( $widget );
LogicBuddy::Tk::Utilities::busy_waitWindow( $dialog );
LogicBuddy::Tk::Utilities::unpack_all_children( $widget );

LogicBuddy::Tk::Utilities::set_buttons_state( $button, $state );

LogicBuddy::Tk::Utilities::define_busy( \&command );

$selection = LogicBuddy::Tk::Utilities::menu( [ -error ] [ -title => $title ], [ -text => $text ], [ -button => $comment, $value, $image ], ... );

@menu_items = LogicBuddy::Tk::Utilities::standard_menu_items( $widget );

if ( LogicBuddy::Tk::Utilities::pointing_to_sel($button) ) { ...

LogicBuddy::Tk::Utilities::info_dialog( $heading, $info );
LogicBuddy::Tk::Utilities::error_dialog( $heading, $error_message );

$dialog = LogicBuddy::Tk::Utilities::working_dialog( $heading, $info );

$dialog = LogicBuddy::Tk::Utilities::Tk_dialog( $parent, $title );

( $dialog, $pane ) = LogicBuddy::Tk::Utilities::Tk_scrolled_dialog( $parent, $title, $width );

( $x, $y ) = LogicBuddy::Tk::Utilities::sub get_geometry( $widget );

LogicBuddy::Tk::Utilities::unpack_all_children( $widget );

LogicBuddy::Tk::Utilities::sub ungrid_all_children( $widget );

=head1 ABSTRACT

LogicBuddy::Tk::Utilities provides Tk utilities for LogicBuddy.

=head1 DESCRIPTION

=head2 EXPORTED SUBPROGRAMS

=head3 BindMouseWheel

   LogicBuddy::Tk::Utilities::BindMouseWheel( $widget );

Binds the mouse wheel to $widget;

   LogicBuddy::Tk::Utilities::BindMouseWheel( $widget, $target );

Binds the mouse wheel events of $widget to $target.

   LogicBuddy::Tk::Utilities::UnBindMouseWheel( $widget );

Binds no mouse wheel to $widget;

=head3 define_busy

   LogicBuddy::Tk::Utilities::define_busy( \&command );

The &command will be called while busy in LogicBuddy::Tk::Utilities::busy_waitWindow

=head3 busy_waitWindow

   LogicBuddy::Tk::Utilities::busy_waitWindow( $dialog );

The $dialog always remains in the foreground until it is finished.

=head3 unpack_all_children

   LogicBuddy::Tk::Utilities::unpack_all_children( $widget );

Unpacks $dialog and all its children.

=head3 ungrid_all_children

   LogicBuddy::Tk::Utilities::ungrid_all_children( $widget );

Ungrids $dialog and all its children.

=head3 set_buttons_state

   LogicBuddy::Tk::Utilities::set_buttons_state( $button, $state );

Sets the $button to normal if $state otherwise to disabled.

=head3 pointing_to_sel

   if ( LogicBuddy::Tk::Utilities::pointing_to_sel($button) ) { ...

Checks whether the cursor is pointing to the selection part of the button image.
The size of a selectable button is larger than a standard button.
The selection part is the right lower corner of that button.

=head3 menu

   $selection = LogicBuddy::Tk::Utilities::menu( [ -error ] [ -title => $title ], [ -text => $text ], [ -button => $comment, $value, $image ], ... );

Opens a dialog menu with a list of entries.
If a button is clicked the $value is returned.

$type = '-error' uses error decorations;

$type = '-title' inserts a tile with text $title.

$type = '-text' inserts the $text.

$type = '-button' inserts a button with $image and $comment.
If the button is pressed the dialog is closed and $value returned.

=head3 standard_menu_items

   @menu_items = LogicBuddy::Tk::Utilities::standard_menu_items( $widget );

Generates the list of @menu_items. Use like

   my $menu = $widget->Menu( -tearoff   => FALSE,
                             -menuitems => [
                                              @_some_menu_items,
                                              '-separator',
                                              LogicBuddy::Tk::Utilities::standard_menu_items( $widget ),
                                           ]
                           );

=head3 Tk_dialog

$dialog = LogicBuddy::Tk::Utilities::Tk_dialog( $parent, $title );

Creates a top level dialog with $title.

=head3 Tk_scrolled_dialog

( $dialog, $pane ) = LogicBuddy::Tk::Utilities::Tk_scrolled_dialog( $parent, $title, $width );

Creates a scrolled top level dialog with $title.
$dialog is the dialog itself and $pane the scrolled pane containing the dialog.
The dialog will be created with $width:

=head3 get_geometry

( $x, $y ) = LogicBuddy::Tk::Utilities::sub get_geometry( $widget );

Returns the geometry of the $widget.

=head3 info_dialog

   LogicBuddy::Tk::Utilities::info_dialog( $heading, $info );

Opens a dialog displaying $heading and $info.

=head3 error_dialog

   LogicBuddy::Tk::Utilities::error_dialog( $heading, $error_message );

Opens a dialog displaying $heading and $error_message.

=head3 working_dialog

   $dialog = LogicBuddy::Tk::Utilities::working_dialog( $heading, $info );

Opens a dialog displaying $heading and $info.
This dialog has no buttons and shall be destroyed when the info is no longer valid.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;

use Tk;
use Tk::Images;

use LogicBuddy::Tk::Configuration;

package LogicBuddy::Tk::Utilities;

use English '-no_match_vars';

################################################################################
# global constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

use Readonly;
Readonly my $C_Pad                => $Tk::Configuration::C_Pad;
Readonly my $C_Pad_Standard       => $Tk::Configuration::C_Pad_Standard;
Readonly my %C_Standard_Pads      => %Tk::Configuration::C_Standard_Pads;

################################################################################
# global variable
################################################################################

my $G_busy_command = undef;

################################################################################
# internal callback subprograms
################################################################################

################################################################################
# internal Tk subprograms
################################################################################

#   ITk_Edit_Menu is derived from Tk::Text::EditMenuItems
sub ITk_Edit_Menu {
   my ( $widget ) = @_;

   my @items = ();
   foreach my $op ($widget->clipEvents) {
      push( @items, [ 'command' => ucfirst $op,
                      -command  => [ $widget => "clipboard$op"]
                    ]
          );
   };
   push( @items,
         '-',
         [ 'command' => 'Select all',   -command => [$widget => 'selectAll'] ],
         [ 'command' => 'Unselect all', -command => [$widget => 'unselectAll'] ],
       );

  return @items;

}; # ITk_Edit_Menu

sub ITk_Search_Menu {
   my ( $widget ) = @_;

   return (
            [ 'command' => 'Find',          -command => sub { I_find_and_replace_popup( $widget, TRUE ) } ],
            [ 'command' => 'Find next',     -command => [$widget => 'FindSelectionNext'] ],
            [ 'command' => 'Find previous', -command => [$widget => 'FindSelectionPrevious'] ],
          );

}; # ITk_Search_Menu

sub ITk_View_Menu {
   my ( $widget ) = @_;

   my $configure_wrap;
   tie $configure_wrap, 'Tk::Configure', $widget, '-wrap';
   return (
            [ radiobutton => 'Wrap word',      -variable => \$configure_wrap, -value => 'word' ],
            [ radiobutton => 'Wrap character', -variable => \$configure_wrap, -value => 'char' ],
            [ radiobutton => 'Wrap none',      -variable => \$configure_wrap, -value => 'none' ],
          );

}; # ITk_View_Menu

################################################################################
# internal callback subprograms
################################################################################

sub IC_shift_dialog {
   my ( $dialog, $abs_x, $abs_y, $ready_ref ) = @_;

   # $ready_ref optional

   Readonly my $C_shift_period     => 25; # ms
   Readonly my $C_shift_distance   => 10;

   my $root_width  = $dialog->vrootwidth();
   my $root_height = $dialog->vrootheight();
   my $width       = $dialog->width();
   my $height      = $dialog->height();

   if ( $width < $C_shift_distance and $height < $C_shift_distance ) {
      $dialog->after( $C_shift_period, sub { IC_shift_dialog( $dialog, $abs_x, $abs_y ), $ready_ref } );
      return;
   };

   $abs_x -= 2 * $C_shift_distance if ($abs_x + $width  + $C_shift_distance) > $root_width;
   $abs_y -= 2 * $C_shift_distance if ($abs_y + $height + $C_shift_distance) > $root_height;

   $dialog->geometry("+$abs_x+$abs_y");

   if ( ($abs_x + $width + $C_shift_distance) > $root_width or ($abs_y + $height + $C_shift_distance) > $root_height ) {
      $dialog->after( $C_shift_period, sub { IC_shift_dialog( $dialog, $abs_x, $abs_y, $ready_ref ); } );
   } elsif ( $ready_ref ) {
      $$ready_ref = TRUE;
   };

   return;

}; # IC_shift_dialog

################################################################################
# exported subprograms
################################################################################

sub BindMouseWheel {
    my ( $widget, $target ) = @_;

    $target = $widget unless $target;

    if ( $^O eq 'MSWin32') {
       $widget->bind( '<MouseWheel>' => [ sub { $target->yview('scroll', -($_[1] / 120) * 3, 'units') }, Tk::Ev('D') ] );
    } else {
       $widget->bind( '<4>' => sub { $target->yview('scroll', -3, 'units') unless $Tk::strictMotif; } );
       $widget->bind( '<5>' => sub { $target->yview('scroll', +3, 'units') unless $Tk::strictMotif; } );
    }

    return;

}; # BindMouseWheel

sub UnBindMouseWheel {
   my ( $widget ) = @_;

   if ( $OSNAME eq 'MSWin32') {
      $widget->bind( '<MouseWheel>' => undef );
   } else {
      $widget->bind( '<4>' => undef );
      $widget->bind( '<5>' => undef );
   }

   return;

}; # UnBindMouseWheel

sub set_buttons_state {
   my ( $button, $state ) = @_;

   return if not ( $button and Tk::Exists( $button ) );
   $button->configure( -state => ( $state  ?  'normal'  :  'disabled' ) ) if $button and Tk::Exists( $button );

   return;

}; # set_buttons_state

sub pointing_to_sel {
   my ( $button ) = @_;

   Readonly my $C_x_offset => 22; # from left
   Readonly my $C_y_offset => 15; # from top

   return ( ( ($button->pointerx() - $button->rootx()) > $C_x_offset ) and
            ( ($button->pointery() - $button->rooty()) > $C_y_offset )
          );

}; # pointing_to_sel

sub define_busy {
   $G_busy_command = shift;
   return;
}; # define_busy

sub busy_waitWindow {
   my ( $dialog ) = @_;

   use Exception::Class::TryCatch;

   # in case of an exception reset to busy( FALSE )
   try eval {
      &$G_busy_command( TRUE ) if $G_busy_command;
      $dialog->update();
      $dialog->grab();
      $dialog->focus();

      Readonly my $C_raise_period => $Tk::Configuration::C_Standard_period;
      $dialog->repeat( $C_raise_period => sub { $dialog->deiconify(); $dialog->raise(); } );

      $dialog->waitWindow();
      &$G_busy_command( FALSE ) if $G_busy_command;
   };
   if ( catch my $error ) {
      &$G_busy_command( FALSE ) if $G_busy_command;
      $error->rethrow();
   };

   return;

}; # busy_waitWindow

sub standard_menu_items {
   my ( $widget ) = @_;

   return(
           ITk_Edit_Menu( $widget ),
           '-separator',
           ITk_Search_Menu( $widget ),
           '-separator',
           [ 'cascade'  => 'View',
             -tearoff   => FALSE,
             -menuitems => [ ITk_View_Menu( $widget ) ]
           ],
         );

}; # standard_menu_items

sub Tk_dialog {
   my ( $parent, $title ) = @_;

   my $dialog = $parent->Toplevel();

   $dialog->iconimage( Tk::Images::image_of( $parent, Tk::Images::LogicBuddy( $parent ) ) ) if Tk::Images::LogicBuddy( $parent );

   $dialog->title( $title );

   if ( $^O ne 'MSWin32' ) {
      my ( $x, $y ) = get_geometry( $parent );
      $dialog->geometry( "+$x+$y" );
   };

   return $dialog;

}; # Tk_dialog

sub Tk_scrolled_dialog {
   my ( $parent, $title, $width ) = @_;

   $width = $Tk::Configuration::C_standard_width if not defined $width;

   my $dialog = Tk_dialog( $parent, $title );
   my $pane = $dialog->Scrolled( "Pane",
                                 -scrollbars => 'osoe',
                                 -sticky     => 'nw',
                                 -height     => $Tk::Configuration::C_scrolled_height,
                                 -width      => $width,
                               )->pack( -side => 'top', -expand => TRUE, -fill => 'both' );
   $pane->Subwidget('xscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $pane->Subwidget('yscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   LogicBuddy::Tk::Utilities::BindMouseWheel( $pane );
   $pane->focus();

   return ( $dialog, $pane );

}; # Tk_scrolled_dialog

sub info_dialog {
   my ( $heading, $info ) = @_;

   my $dialog = Tk_dialog( $LogicBuddy::Tk::G_top, LogicBuddy::Singular::Info() );

   $dialog->Label( -text  => $heading
                 )->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );

   $dialog->Label( -text    => $info,
                   -justify => 'left',
                 )->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );

   my $buttons = $dialog->Frame( -relief      => 'groove',
                                 -borderwidth => 2
                               )->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );

   my $quit_button = $buttons->Button( -command  => sub { $dialog->destroy() if Tk::Exists($dialog); },
                                       -text     => LogicBuddy::Singular::Close(),
                                       -image    => Tk::Images::cancel( $dialog ),
                                       -compound => 'left',
                                     )->pack( -side => 'left',  -padx => $C_Pad_Standard, -pady => $C_Pad );

   $dialog->bind( '<Key-Escape>' => sub { $dialog->destroy() if Tk::Exists($dialog); } );

   busy_waitWindow( $dialog );

   return;

}; # info_dialog

sub working_dialog {
   my ( $heading, $info ) = @_;

   my $dialog = Tk_dialog( $LogicBuddy::Tk::G_top, LogicBuddy::Singular::Info() );

   $dialog->Label( -text  => $heading
                 )->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );

   $dialog->Label( -text    => $info,
                   -justify => 'left',
                 )->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );

   $dialog->update();

   return $dialog;

}; # working_dialog

sub error_dialog {
   my ( $heading, $error_message ) = @_;

   my $dialog = Tk_dialog( $LogicBuddy::Tk::G_top, LogicBuddy::Singular::Error() );

   $dialog->Label( -text       => $heading,
                   -foreground => $Tk::Configuration::C_error_foreground,
                 )->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );

   $dialog->Label( -text    => $error_message,
                   -justify => 'left',
                 )->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );

   my $buttons = $dialog->Frame( -relief      => 'groove',
                                 -borderwidth => 2
                               )->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );

   my $quit_button = $buttons->Button( -command  => sub { $dialog->destroy() if Tk::Exists($dialog); },
                                       -text     => LogicBuddy::Singular::Quit(),
                                       -image    => Tk::Images::cancel( $dialog ),
                                       -compound => 'left',
                                     )->pack( -side => 'left', %C_Standard_Pads );

   $dialog->bind( '<Key-Escape>' => sub { $dialog->destroy() if Tk::Exists($dialog); } );

   busy_waitWindow( $dialog );

   return;

}; # error_dialog

sub get_geometry {
   my ( $parent ) = @_;

   Readonly my $x_shift => 200;
   Readonly my $y_shift => 100;

   my $x = $parent->pointerx();
   my $y = $parent->pointery();
   unless ( $x and $y ) {
      $x = $parent->rootx();
      $y = $parent->rooty();
   };
   $x = $parent->screenwidth()  - $x_shift if $x + $x_shift > $parent->screenwidth();
   $y = $parent->screenheight() - $y_shift if $y + $y_shift > $parent->screenheight();
   $x = 0 unless $x > 0;
   $y = 0 unless $y > 0;

   return( $x, $y );

}; # get_geometry

sub menu {
   my ( @entries ) = @_;

   my $result;
   my $border_width = 1;
   my $dialog = Tk_dialog( $LogicBuddy::Tk::G_top, EMPTY );
   $dialog->overrideredirect( TRUE );
   $dialog->configure( -background => 'black' );
   my $frame = $dialog->Frame( -background  => $Tk::Configuration::C_background,
                               -pady        => $C_Pad_Standard,
                             )->pack( -side => 'top', -expand => TRUE, -fill => 'both', -padx => 1, -pady => 1 );

   foreach my $entry ( @_ ) {
      my ( $type, $comment, $value, $image ) = @$entry;
      if ( $type eq '-error' ) {
         $dialog->configure( -background => $Tk::Configuration::C_error_foreground );
         $border_width = 5;
         $frame->Label( -text       => $comment,
                        -font       => LogicBuddy::Tk::Fonts::bold_dialog_font(),
                        -anchor     => 'w',
                        -justify    => 'left',
                        -wraplength => $Tk::Configuration::C_min_width,
                        -background => $Tk::Configuration::C_background,
                        %Tk::Configuration::C_Active_color
                      )->pack( -side => 'top', -fill => 'x', -padx => $C_Pad_Standard, -pady => $C_Pad );
      } elsif ( $type eq '-title' ) {
         $frame->Label( -text       => $comment,
                        -font       => LogicBuddy::Tk::Fonts::bold_dialog_font(),
                        -anchor     => 'w',
                        -justify    => 'left',
                        -wraplength => $Tk::Configuration::C_min_width,
                        -background => $Tk::Configuration::C_background
                      )->pack( -side => 'top', -fill => 'x', %C_Standard_Pads );
      } elsif ( $type eq '-text' ) {
         $frame->Label( -text       => $comment,
                        -font       => LogicBuddy::Tk::Fonts::dialog_font(),
                        -anchor     => 'w',
                        -justify    => 'left',
                        -wraplength => $Tk::Configuration::C_min_width,
                        -background => $Tk::Configuration::C_background
                      )->pack( -side => 'top', -fill => 'x', -padx => $C_Pad_Standard, -pady => $C_Pad );
      } elsif ( $type eq '-button' ) {
         my $button = $frame->Button( -command  => sub { $result = $value; $dialog->destroy() if Tk::Exists($dialog); },
                                      -text     => $comment,
                                      -image    => $image,
                                      -compound => 'left',
                                      -font     => LogicBuddy::Tk::Fonts::dialog_font(),
                                    )->pack( -side => 'top', -anchor => 'w', -padx => $C_Pad_Standard, -pady => $C_Pad );
         $button->configure( -background         => $Tk::Configuration::C_background,
                             -activebackground   => $Tk::Configuration::C_background,
                             -borderwidth        => 0,
                             -highlightthickness => 0,
                           );
      } else {
         die "$LogicBuddy::Tk::Dialogs::illegal_menu_type_exception '$type'";
      };
   };

   $frame->pack( -side => 'top', -expand => TRUE, -fill => 'both', -padx => $border_width, -pady => $border_width );

   Readonly my $shift_period => $Tk::Configuration::C_Standard_period;
   $dialog->after( $shift_period, sub { IC_shift_dialog( $dialog, get_geometry( $LogicBuddy::Tk::G_top ) ); } );
   busy_waitWindow( $dialog );

   return $result;

}; # menu

sub unpack_all_children {
   my ( $widget ) = @_;

   foreach my $child ( $widget->packSlaves() ) {
      unpack_all_children( $child );
   };
   $widget->packForget();
   return;

}; # unpack_all_children

sub ungrid_all_children {
   my ( $widget ) = @_;

   foreach my $child ( $widget->gridSlaves() ) {
      ungrid_all_children( $child );
   };
   $widget->gridRemove();
   return;

}; # ungrid_all_children

################################################################################
# end LogicBuddy::Tk::Utilities
################################################################################

1;

