=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

LogicBuddy::Tk::Viewer - module

=head1 SYNOPSIS

   use LogicBuddy::Tk::Viewer;

   $viewer  = new LogicBuddy::Tk::Viewer( $parent );

              $viewer->pack( ... );
              $viewer->packForget();
              $viewer->configure_cursor( $cursor );
   $buttons = $viewer->control_buttons( $parent );
              $viewer->message( $message, ... );
              $viewer->new_message( $message );
              $viewer->show_message();
              $viewer->solution( $text );
   @labels  = $viewer->input_labels();
   @labels  = $viewer->output_labels();
   $type    = $viewer->logic_type();
              $viewer->define_table( \&subprogram );
              $viewer->define_input_labels( @input_labels );
              $viewer->define_output_labels( @output_labels );
              $viewer->embed_into_solution( $widget, $help );

=head1 ABSTRACT

LogicBuddy::Tk::Viewer abstracts viewing LogicBuddy.

=head1 DESCRIPTION

=head2 METHODS

=head3 new

   $viewer = new LogicBuddy::Tk::Viewer( $parent );
   instantiates a LogicBuddy viewer

=head3 pack

   $viewer->pack( ... );

Packs the viewer

=head3 packForget

   $viewer->packForget();

Unpacks the $viewer and all its subwidgets.

=head3 configure_cursor

   $viewer->configure_cursor( $cursor );

Sets the cursor to $cursor

=head3 control_buttons

   $buttons = $viewer->control_buttons( $parent );

Creates a frame with the controlbuttons logical into $parent.
The $buttons are not packed.

=head3 message

   $viewer->message( $message, ... );

Puts the messages into the message dialogs.

=head3 new_message

   $viewer->new_message( $message );

Remembers the $message.
This massages will be displayed when the next $viewer->message is called.

=head3 show_message

   $viewer->show_message();

Raises the message dialog.

=head3 solution

   $viewer->solution( $text );

Puts the $text into the solution window.

=head3 input_labels

   @labels = $viewer->input_labels();

Returns the list of defined input variable names.

=head3 output_labels

  @labels = $viewer->output_labels();

Returns the list of defined output variable names.


=head3 logic_type

   $type = $viewer->logic_type();

Returns the type of logic selected ( AND OR - NAND ).

=head3 define_table

   $viewer->define_table( \&subprogram );

Generates the truth table using &subprogram.

&subprogram is called like &subprogram( $i1, $i2 ... );
$i1 ... are the logical values ( 0 or 1 ) of the input variable in n=1..number_of_inputs

=head3 define_input_labels

   $viewer->define_input_labels( @input_labels );

Defines new input variable names

=head3 define_output_labels

   $viewer->define_output_labels( @output_labels );

Defines new output variable names

=head3 embed_into_solution

   $viewer->embed_into_solution( $widget, $help );

Embeds the $widget into the solution window and binds the $help command to the solution window.
If $widget is defined the current text in the solution window is saved.
If $widget is not defined the last text in the solution window is restored.
If $help is defined the $help command is bound to the solution window.
If $help is defined the standard help command is bound to the solution window.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de Praktische-Elektronik.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;

use Tk;
use Tk::Pane;
use Tk::Images;
use Tk::ROText;
use Tk::Help;

use LogicBuddy::Configuration;
use LogicBuddy::Plural;
use LogicBuddy::Singular;
use LogicBuddy::Utilities;
use LogicBuddy::Tk::Configuration;

use LogicBuddy::Tk::Utilities;
use LogicBuddy::Tk::Balloon;
use LogicBuddy::Tk::Messages;
use LogicBuddy::Tk::Solution;
use LogicBuddy::Tk::Help;
use LogicBuddy::Tk::Editor;
use LogicBuddy::Tk::Table::Multiple;

use LogicBuddy::Espresso;

package LogicBuddy::Tk::Viewer;

use English '-no_match_vars';

################################################################################
# global constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

use Readonly;
Readonly my $C_Pad                => $Tk::Configuration::C_Pad;
Readonly my %C_Pads               => %Tk::Configuration::C_Pads;
Readonly my $C_Pad_Standard       => $Tk::Configuration::C_Pad_Standard;
Readonly my %C_Standard_Pads      => %Tk::Configuration::C_Standard_Pads;

################################################################################
# global variable
################################################################################

################################################################################
## class LogicBuddy::Tk::Viewer
#################################################################################

sub new {
   my $proto  = shift;
   my $parent = shift;

   my $class = ref($proto) || $proto;
   my $self;

   $self->{parent}              = $parent;
   $self->{Balloon}             = undef;
   $self->{button_frame}        = undef;
   $self->{buttons}             = []; # All buttons in @{ $self->{buttons} } will be disabled when busy.
   $self->{language}            = LogicBuddy::Configuration::language();
   $self->{selected_logic}      = 'NAND';
   $self->{HelpDialog}          = undef;
   $self->{Help}                = undef;
   $self->{Viewer}              = undef;
   $self->{Solution}            = undef;
   $self->{Table}               = undef;
   $self->{language_button}     = undef;
   $self->{language_info_shown} = FALSE;
   $self->{Editor}              = undef;

   bless $self, $class;

   $self->{Espresso}           = new LogicBuddy::Espresso( $self );

   $self->ITk_viewer( $parent );
   $self->{Messages}           = new LogicBuddy::Tk::Messages( $self->{Viewer}, $self->{Balloon} );

   my $info = LogicBuddy::Singular::Info_1();
   $info   .= "\n" . LogicBuddy::Singular::Windows_note() if not $LogicBuddy::C_DOS_window;
   $self->solution( $info );

   return $self;

}; # new

################################################################################
# internal subprograms
################################################################################

################################################################################
# internal Tk subprograms
################################################################################

sub ITk_viewer {
   my ( $self, $parent ) = @_;

   $self->{Viewer} = $parent->Frame( -background => $Tk::Configuration::C_background );

   $self->{Balloon} = LogicBuddy::Tk::Balloon->new( $self->{Viewer} );

   $self->{Solution} = LogicBuddy::Tk::Solution->new( $self->{Viewer}, $self->{Balloon} );
   $self->{Solution}->pack( -side => 'top', -fill => 'x', -expand => FALSE, -padx => 2, -pady => 2 );

   $self->{Table} = new LogicBuddy::Tk::Table::Multiple( $self->{Viewer}, $self->{Balloon}, $self );
   $self->{Table}->pack( -side => 'top', -fill => 'both', -expand => TRUE, -anchor => 'nw' );

   $self->{Editor} = new LogicBuddy::Tk::Editor( $self->{parent}, $self->{Balloon}, $self );

   $self->{Viewer}->bind( '<Key-F1>'   => sub { LogicBuddy::Tk::Help::main_help( $self->{parent} ); Tk::break } );
   $self->{Viewer}->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::main_help( $self->{parent} ); Tk::break } );

   return $self->{Viewer};

}; # ITk_viewer

sub ITk_help {
   my ( $self ) = @_;

   if ( defined $self->{HelpDialog} and Tk::Exists($self->{HelpDialog}) ) {
      $self->{HelpDialog}->update();
      $self->{HelpDialog}->deiconify();
      $self->{HelpDialog}->raise( $self->{Viewer} );
      return;
   };

   my $dialog = LogicBuddy::Tk::Utilities::Tk_dialog( $self->{Viewer}, LogicBuddy::Singular::Help() );
   $dialog->bind( '<Key-F1>'   => sub { LogicBuddy::Tk::Help::help_dialog( $self->{parent} ); Tk::break } );
   $dialog->bind( '<Button-3>' => sub { LogicBuddy::Tk::Help::help_dialog( $self->{parent} ); Tk::break } );

   my $text = $dialog->Scrolled( 'ROText',
                                 -scrollbars => 'osoe',
                                 %Tk::Configuration::C_Text_configuration
                               )->pack( -side => 'top', -expand => TRUE, -fill => 'both', %C_Standard_Pads );
   $text->Subwidget('xscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $text->Subwidget('yscrollbar')->configure( -width => $Tk::Configuration::C_scrollbar_width );
   $text->menu( undef ); # no popup menu

   $self->{HelpDialog} = $dialog;
   $self->{Help}       = $text;

   $dialog->protocol( 'WM_DELETE_WINDOW' => sub { $dialog->iconify() } );

   return;

}; # ITk_help

sub ITk_logic_button {
   my ( $self, $logic_button ) = @_;

   foreach my $logic ( 'AND OR', 'NAND' ) {  #, 'NOR' ) {
      $logic_button->radiobutton( -label    => $logic,
                                  -command  => sub { $self->solve() },
                                  -variable => \$self->{selected_logic},
                                  -value    => $logic,
                                  -font     => $Tk::Configuration::C_standard_font
                                );
   };

   return;

}; # ITk_logic_button

################################################################################
# internal callback subprograms
################################################################################

sub IC_toggle_language {
   my ( $self ) = @_;

   if ( $self->{language} eq 'English' ) {
      $self->I_set_language( 'German' );
   } else {
      $self->I_set_language( 'English' );
   };

   return if $self->{language_info_shown};

   LogicBuddy::Tk::Utilities::info_dialog( LogicBuddy::Singular::Changed_language(),
                                           LogicBuddy::Singular::Language_of_balloon_help_will_change_on_restart( '.' )
                                         );
   $self->{language_info_shown} = TRUE;

   return;

}; # IC_toggle_language

sub IC_exit {
   my ( $self ) = @_;

   if ( $self->{parent} and  Tk::Exists $self->{parent} ) {
      $self->{parent}->destroy();
   } else {
      exit;
   };
   return;

}; # IC_exit

sub IC_copy_to_editor {
   my ( $self ) = @_;

   $self->solve();

   my $function = LogicBuddy::Singular::Input() . ' = ' . ( join ', ', $self->{Table}->input_labels() ) . ".\n\n" .
                 $self->{Solution}->get() .
                 LogicBuddy::Singular::Output() . ' = ' . ( join ', ', $self->{Table}->output_labels() ) . ".\n";

   $self->{Editor}->enter_function( $function );

   return;

}; # IC_copy_to_editor

################################################################################
# internal subprograms
################################################################################

sub I_set_language {
   my ( $self, $language ) = @_;

   if ( $language ) {
      $self->{language} = $language;
      LogicBuddy::Locale::SetLanguage( $language );
      LogicBuddy::Configuration::set_language( $language );
      $self->{Editor}->set_language( $language );
   };

   if ( $self->{language} eq 'German' ) {
      $self->{language_button}->configure( -text => 'Deutsch',
                                           -image => Tk::Images::flag_de( $self->{Viewer} )
                                         );
      $self->{Balloon}->attach( $self->{language_button}, LogicBuddy::Singular::Select_language() );
   } else {
      $self->{language_button}->configure( -text => 'English',
                                           -image => Tk::Images::flag_en( $self->{Viewer} )
                                         );
      $self->{Balloon}->attach( $self->{language_button}, LogicBuddy::Singular::Select_language() );
   };

   return;

}; # I_set_language

################################################################################
# exported subprograms
################################################################################

sub packForget {
   my $self = shift;
   $self->{Viewer}->packForget();
   return;
}; # packForget

sub pack {
   my $self = shift;
   $self->{Viewer}->pack( @_ );
   return;
}; # pack

sub configure_cursor {
   my $self   = shift;
   my $cursor = shift;
   $self->{Viewer}->configure( -cursor => $cursor ) if $self->{Viewer} and Tk::Exists( $self->{Viewer} );
   return;
}; # configure_cursor

sub control_buttons {
   my ( $self, $parent ) = @_;

   # All buttons in @{ $self->{buttons} } will be disabled when busy.

   return ( $self->{button_frame} ) if $self->{button_frame};

   my $button_frame      = $parent->Frame();
   $self->{button_frame} = $button_frame;
   $button_frame->optionAdd( "*Button*Relief" => 'flat' );

   my $exit_button = $button_frame->Button( -command => sub { $self->IC_exit() },
                                            -text    => 'Exit',
                                            -image   => Tk::Images::exit( $parent )
                                          )->pack( -side => 'right', %C_Pads );
   $self->{Balloon}->attach( $exit_button, LogicBuddy::Singular::Cancel() );

   $self->{balloon_button} = $self->{Balloon}->Button( $button_frame )->pack( -side => 'right', %C_Pads );

   $self->{language_button} = $button_frame->Button( -command => sub { $self->IC_toggle_language(); },
                                                     -text    => 'English',
                                                     -image   => Tk::Images::flag_en( $parent )
                                                   )->pack( -side => 'right', %C_Pads );
   $self->{Balloon}->attach( $self->{language_button}, LogicBuddy::Singular::Select_Language() );

   $self->{help_button} = $button_frame->Button( -command => sub { LogicBuddy::Tk::Help::main_help( $self->{parent}) },
                                                 -text    => LogicBuddy::Singular::Help(),
                                                 -image   => Tk::Images::help( $parent )
                                               )->pack( -side => 'right', %C_Pads );
   $self->{Balloon}->attach( $self->{help_button}, LogicBuddy::Singular::Help_Description_for_Balloon() );

   my $message_button = $button_frame->Button( -command => sub { $self->{Messages}->show_message(); },
                                               -text    => LogicBuddy::Plural::Show_message(),
                                               -image   => Tk::Images::viewtext( $parent )
                                             )->pack( -side => 'right', %C_Pads );
   $self->{Balloon}->attach( $message_button, LogicBuddy::Plural::Show_message() );

   my $solution_button = $button_frame->Button( -command => sub { $self->{Solution}->show(); },
                                                -text    => LogicBuddy::Plural::Show(),
                                                -image   => Tk::Images::NAND( $parent )
                                              )->pack( -side => 'right', %C_Pads );
   $self->{Balloon}->attach( $solution_button, LogicBuddy::Plural::Show_solution_dialog() );
   push @{ $self->{buttons} }, $solution_button;

   $self->{edit_button} = $button_frame->Button( -command => sub { $self->{Editor}->edit(); },
                                                 -text    => LogicBuddy::Singular::Edit(),
                                                 -image   => Tk::Images::C_edit( $parent )
                                               )->pack( -side => 'right', %C_Pads );
   $self->{Balloon}->attach( $self->{edit_button}, LogicBuddy::Singular::Edit_logical_function() );
   push @{ $self->{buttons} }, $self->{edit_button};

   $self->{copy_button} = $button_frame->Button( -command => sub { $self->IC_copy_to_editor() },
                                                  -text    => LogicBuddy::Singular::Copy(),
                                                  -image   => Tk::Images::editcopy( $parent )
                                                )->pack( -side => 'right', %C_Pads );
   $self->{Balloon}->attach( $self->{copy_button}, LogicBuddy::Singular::Copy_logical_function_to_editor() );
   push @{ $self->{buttons} }, $self->{copy_button};

   my $solve_button = $button_frame->Button( -command => sub { $self->solve() },
                                             -text    => LogicBuddy::Singular::Solve(),
                                             -image   => Tk::Images::b_work( $parent )
                                           )->pack( -side => 'left', %C_Pads );
   $self->{Balloon}->attach( $solve_button, LogicBuddy::Singular::Solve_logical_function() );
   push @{ $self->{buttons} }, $solve_button;

   my $table_button = $button_frame->Button( -command => sub { $self->{Table}->new_table(); $self->{Solution}->delete(); },
                                             -text    => LogicBuddy::Singular::Table(),
                                             -image   => Tk::Images::table( $parent )
                                           )->pack( -side => 'left', %C_Pads );
   $self->{Balloon}->attach( $table_button, LogicBuddy::Singular::Define_new_table() );
   push @{ $self->{buttons} }, $table_button;

   if ( $self->{Editor} ) {
      my $status_button = $self->{Editor}->status_button( $button_frame )->pack( -side => 'left',  -padx => $C_Pad_Standard, -pady => $C_Pad );
      push @{ $self->{buttons} }, $status_button;
   };

   my $logic_button = $button_frame->Menubutton( -textvariable => \$self->{selected_logic},
                                                 -font         => $Tk::Configuration::C_standard_font,
                                                 -tearoff      => FALSE
                                               )->pack( -side => 'left',  -padx => 50, -pady => $C_Pad );
   $self->{Balloon}->attach( $logic_button, LogicBuddy::Singular::Select_logic() );
   push @{ $self->{buttons} }, $logic_button;
   $self->ITk_logic_button( $logic_button );

   $self->I_set_language( $self->{language} );

   return ( $button_frame );

}; # control_buttons

sub solve {
   my ( $self ) = @_;
   $self->{Espresso}->generate();
   return;
}; # solve

sub message {
   my ( $self, @messages ) = @_;
   $self->{Messages}->message( @messages );
   return;
}; # message

sub insert_message {
   my ( $self, @messages ) = @_;
   $self->{Messages}->insert_message( @messages );
   return;
}; # insert_message

sub new_message {
   my ( $self, $message ) = @_;
   $self->{Messages}->new_message( $message );
   return;
}; # new_message

sub show_message {
   my ( $self ) = @_;
   $self->{Messages}->show_message();
   return;
}; # show_message

sub solution {
   my ( $self, $text ) = @_;
   $self->{Solution}->solution( $text );
   return;
}; # solution

sub input_labels {
   my ( $self ) = @_;
   return $self->{Table}->input_labels();
}; # input_labels

sub output_labels {
   my ( $self ) = @_;
   return $self->{Table}->output_labels();
}; # output_labels

sub values_of {
   my ( $self, $input_index ) = @_;
   return $self->{Table}->values_of( $input_index );
}; # values_of

sub logic_type {
   my ( $self ) = @_;
   return $self->{selected_logic};
}; # logic_type

sub define_table {
   my ( $self, $command ) = @_;
   $self->{Table}->define_table( $command );
   return;
}; # define_table

sub define_input_labels {
   my ( $self, @input_labels ) = @_;
   $self->{Table}->define_input_labels( @input_labels );
   return;
}; # define_input_labels

sub define_output_labels {
   my ( $self, @output_labels ) = @_;
   $self->{Table}->define_output_labels( @output_labels );
   return;
}; # define_output_labels

sub changed {
   my ( $self ) = @_;
   $self->{Editor}->changed() if $self->{Editor};
   return;
}; # changed

sub embed_into_solution {
   my ( $self, $widged, $help ) = @_;
   $self->{Solution}->embed( $widged, $help );
   return;
}; # embed_into_solution

sub busy {
   my ( $self, $busy ) = @_;

   if ( $busy ) {
      foreach my $button ( @{ $self->{buttons} } ) {
         $button->configure( -state => 'disabled' );
      };
      $self->{parent}->configure( -cursor => $Tk::Configuration::C_busy_cursor );
   } else {
      foreach my $button ( @{ $self->{buttons} } ) {
         $button->configure( -state => 'normal' );
      };
      $self->{parent}->configure( -cursor => $Tk::Configuration::C_standard_cursor );
   };

   $self->{Editor}->busy( $busy ) if $self->{Editor};

   return;

}; # busy

################################################################################
# end LogicBuddy::Tk::Viewer
################################################################################

1;

