=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de www.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

Tk::Dialogs - module

=head1 SYNOPSIS

use Tk::Dialogs;

   $dialog = Tk::Dialogs::Tk_dialog( $parent, $title, '$con_image_name );

   ( $dialog, $pane ) = Tk::Dialogs::Tk_scrolled_dialog( $parent, $title, $width, $icon_image_name );
   
   Tk::Dialogs::busy( $dialog, $busy );

   Tk::Dialogs::disable_close_window_button( $dialog );
   Tk::Dialogs::enable_close_window_button( $dialog );

   $selection = Tk::Dialogs::menu( $parent, [ $type, $comment, $value, $image ], ... );

   ( $x, $y ) = Tk::Dialogs::get_geometry( $parent );

   $dialog = Tk::Dialogs::show_file( $parent, $file );

   Tk::Dialogs::busy_waitWindow( $dialog );

   Tk::Dialogs::define_busy_command( sub { ... } );

   Tk::Dialogs::delay_messages()
   Tk::Dialogs::throw_pending_messages()

   Tk::Dialogs::info_dialog( $parent, $heading, $info, $no_grab, $width, $lines );
   Tk::Dialogs::note_dialog( $parent, $heading, $info, $no_grab, $width, $lines );
   Tk::Dialogs::error_dialog( $parent, $heading, $error_message );

=head1 ABSTRACT

Tk::Dialogs provides Tk dialogs.

=head1 DESCRIPTION

=head2 Operations

=head3 Tk::Dialogs::Tk_dialog

   $dialog = Tk::Dialogs::Tk_dialog( $parent, $title, $con_image_name );

creates a dialog with $title in $parent. Returns undef if $parent not exists.
If $con_image_name is defined it is used for icon image of the dialog.

=head3 Tk::Dialogs::Tk_scrolled_dialog

   ( $dialog, $pane ) = Tk::Dialogs::Tk_scrolled_dialog( $parent, $title, $width, $icon_image_name );
   ( $dialog, $pane ) = Tk::Dialogs::Tk_scrolled_dialog( $parent, $title, $width );

Creates a dialog with $title in $parent. Returns () if $parent not exists.
$pane is the pane widget in $dialog. Insert widgets into $pane but not $dialog.
If $width is not defined the default width is used.
If $con_image_name is defined it is used for icon image of the dialog.

=head2 Dialog support

=head3 Tk::Dialogs::busy

   Tk::Dialogs::busy( $dialog, $busy );

grabs the $dialog if $busy otherwise releases the grab.
The cursor is set according to the $busy state.

=head3 Tk::Dialogs::disable_close_window_button

   Tk::Dialogs::disable_close_window_button( $dialog );

Disables the main close button of the $dialog window.

=head3 Tk::Dialogs::enable_close_window_button

   Tk::Dialogs::enable_close_window_button( $dialog );

Enables the main close button of the $dialog window.

=head3 Tk::Dialogs::menu

   $selection = Tk::Dialogs::menu( $parent, [ $type, $comment, $value, $image ], ... );

Opens a dialog menu with a list of entries.
If a button is clicked the $value is returned.

   $type = '-error' uses error decorations;

   $type = '-title' inserts a tile with text $title.

   $type = '-text' inserts the $text.

   $type = '-button' inserts a button with $image and $comment.
                     If the button is pressed the dialog is closed and $value returned.

=head2 Waiting for window

=head3 Tk::Dialogs::busy_waitWindow

   Tk::Dialogs::busy_waitWindow( $dialog );

$dialog stays in foreground until it is clodsed.

=head3 Tk::Dialogs::define_busy_command

   Tk::Dialogs::define_busy_command( sub { ... } );

Defines the command executed while busy_waitWindow is running.

=head3 Tk::Dialogs::get_geometry

   ( $x, $y ) = Tk::Dialogs::get_geometry( $parent );

Returns the geometry of the $parent widget.

=head3 Tk::Dialogs::show_file

   $dialog = Tk::Dialogs::show_file( $parent, $file );

Shows the $file.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de www.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;

use Tk;
use Tk::Images;
use Tk::ROText;
use Tk::Pane;

use Tk::StandardConfiguration;

use Tk::Utilities;

package Tk::Dialogs;

################################################################################
# exceptions
################################################################################

use Readonly;

Readonly $Tk::Dialogs::C_illegal_menu_type_exception => 'Tk::Dialogs: Illegal menu type';

################################################################################
# global constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

Readonly my $C_standard_cursor   => $Tk::Configuration::C_Standard_Cursor;
Readonly my $C_busy_cursor       => $Tk::Configuration::C_Watch_Cursor;

Readonly my $C_standard_width    => $Tk::Configuration::C_standard_width;
Readonly my $C_min_width         => $Tk::Configuration::C_min_width;
Readonly my $C_scrollbar_width   => $Tk::Configuration::C_scrollbar_width;

Readonly my $C_scrolled_height   => $Tk::Configuration::C_scrolled_height;
Readonly my $C_text_height       => $Tk::Configuration::C_text_height;
Readonly my $C_text_width        => $Tk::Configuration::C_text_width;

Readonly my $C_padx              => $Tk::Configuration::C_Pad;
Readonly my $C_pady              => $Tk::Configuration::C_Pad;
Readonly my %C_pad_options       => ( -padx => $C_padx,
                                      -pady => $C_pady
                                    );

Readonly my $C_background        => $Tk::Configuration::C_background;
Readonly my $C_Menu_background   => $Tk::Configuration::C_Menu_background;
Readonly my $C_error_foreground  => $Tk::Configuration::C_error_foreground;
Readonly my %C_Active_color      => %Tk::Configuration::C_Active_color;

Readonly my $C_font_family       => $Tk::Configuration::C_font_family;
Readonly my $C_dialog_font_size  => $Tk::Configuration::C_dialog_font_size;

Readonly my $C_bold_dialog_font  => $Tk::Configuration::C_bold_dialog_font;
Readonly my $C_dialog_font       => $Tk::Configuration::C_dialog_font;

################################################################################
# global variable
################################################################################

my $G_busy_command     = undef;
my %G_standard_cursor  = ();

################################################################################
# internal subprograms
################################################################################

################################################################################
# internal Tk subprograms
################################################################################

################################################################################
## internal callback subprograms
#################################################################################

sub IC_shift_dialog {
   my $dialog    = shift;
   my $abs_x     = shift;
   my $abs_y     = shift;
   my $ready_ref = shift; # optional

   my $root_width  = $dialog->vrootwidth();
   my $root_height = $dialog->vrootheight();
   my $width       = $dialog->width();
   my $height      = $dialog->height();

   if ( $width < 10 and $height < 10 ) {
      $dialog->after( 25, sub { IC_shift_dialog( $dialog, $abs_x, $abs_y ), $ready_ref } );
      return;
   };

   $abs_x -= 20 if ($abs_x + $width  + 10)  > $root_width;
   $abs_y -= 20 if ($abs_y + $height + 10) > $root_height;

   $dialog->geometry("+$abs_x+$abs_y");

   if ( ($abs_x + $width + 10) > $root_width or ($abs_y + $height + 10) > $root_height ) {
      $dialog->after( 25, sub { IC_shift_dialog( $dialog, $abs_x, $abs_y, $ready_ref ); } );
   } elsif ( $ready_ref ) {
      $$ready_ref = TRUE;
   };

   return;

}; # IC_shift_dialog

################################################################################
## exported subprograms
#################################################################################

sub Tk_dialog {
   my ( $parent, $title, $icon_image_name ) = @_;

   return if not Tk::Exists( $parent );

   my $dialog = $parent->Toplevel();

   if ( $icon_image_name ) {
      my $image_file = Tk::Images::file_of( $icon_image_name );
      $dialog->iconimage( Tk::Images::image_of( $dialog, $image_file ) ) if $image_file;
   };
   $dialog->minsize( 300, 100 );

   $dialog->title( $title );

   if ( $^O ne 'MSWin32' ) {
      my ( $x, $y ) = get_geometry( $parent );
      $dialog->geometry( "+$x+$y" );
   };

   return $dialog;

}; # Tk_dialog

sub Tk_scrolled_dialog {
   my ( $parent, $title, $width, $icon_image_name ) = @_;
   
   $width = $C_standard_width if not defined $width;

   my $dialog = Tk_dialog( $parent, $title, $icon_image_name );
   my $pane = $dialog->Scrolled( "Pane",
                                 -scrollbars => 'osoe',
                                 -sticky     => 'nw',
                                 -height     => $C_scrolled_height,
                                 -width      => $width,
                               )->pack( -side => 'top', -expand => TRUE, -fill => 'both' );
   $pane->Subwidget("xscrollbar")->configure( -width => $C_scrollbar_width );
   $pane->Subwidget("yscrollbar")->configure( -width => $C_scrollbar_width );
   Tk::Utilities::BindMouseWheel( $pane );
   $pane->focus();

   return ( $dialog, $pane );

}; # Tk_scrolled_dialog

sub disable_close_window_button {
   my ( $dialog ) = @_;
   $dialog->protocol( 'WM_DELETE_WINDOW', sub { $dialog->bell() } );
   return;
}; # disable_close_window_button

sub enable_close_window_button {
   my ( $dialog ) = @_;
   $dialog->protocol( 'WM_DELETE_WINDOW', sub { $dialog->destroy() if Tk::Exists( $dialog ) } );
   return;
}; # enable_close_window_button

sub busy {
   my ( $dialog, $busy ) = @_;

   return if not Tk::Exists( $dialog );

   $dialog->update();
   if ( $busy ) {
      my $type = $dialog;
      if ( $type =~ m/ \A ( [^=]+ ) = HASH /xms ) {
         my $class = $1;
         my $cursor = $dialog->cget( '-cursor' );
         $G_standard_cursor{$class} = $cursor if $cursor;
      };
      $dialog->configure( -cursor => $C_busy_cursor );
      eval {
         $dialog->grab() if $dialog->grabStatus() eq 'none';
      };
   } else {
      my $type = $dialog;
      if ( $type =~ m/ \A ( [^=]+ ) = HASH /xms ) {
         my $class = $1;
         if ( exists $G_standard_cursor{$class} ) {
            $dialog->configure( -cursor => $G_standard_cursor{$class} );
         } else {
            $dialog->configure( -cursor => EMPTY );
         };
      } else {
         $dialog->configure( -cursor => EMPTY );
      };
      $dialog->grabRelease() if $dialog->grabStatus() ne 'none';
   };
   $dialog->update();

   foreach my $child ( $dialog->packSlaves() ) {
      busy( $child, $busy );
   };
   foreach my $child ( $dialog->gridSlaves() ) {
      busy( $child, $busy );
   };

   return;

}; # busy

sub menu {
   my ( $parent, @entries ) = @_;

   my $result;
   my $border_width = 1;
   my $dialog = Tk_dialog( $parent, EMPTY );

   # This is necessary because there may be a after or repeat event on  $self->{parent}.
   return EMPTY if not $dialog;

   $dialog->overrideredirect( TRUE );
   $dialog->configure( -background => 'black' );
   my $frame = $dialog->Frame( -background  => $C_Menu_background,
                               -pady        => $C_pady
                             )->pack( -side => 'top', -expand => TRUE, -fill => 'both', -padx => 1, -pady => 1 );

   foreach my $entry ( @entries ) {
      my ( $type, $comment, $value, $image ) = @$entry;
      if ( $type eq '-error' ) {
         $dialog->configure( -background => $C_error_foreground );
         $border_width = 5;
         $frame->Label( -text       => $comment,
                        -font       => $C_bold_dialog_font,
                        -anchor     => 'w',
                        -justify    => 'left',
                        -wraplength => $C_min_width,
                        -background => $C_Menu_background,
                        %C_Active_color
                      )->pack( -side => 'top', -fill => 'x', -padx => $C_padx, -pady => 0 );
      } elsif ( $type eq '-title' ) {
         $frame->Label( -text       => $comment,
                        -font       => $C_bold_dialog_font,
                        -anchor     => 'w',
                        -justify    => 'left',
                        -wraplength => $C_min_width,
                        -background => $C_Menu_background
                      )->pack( -side => 'top', -fill => 'x', %C_pad_options );
      } elsif ( $type eq '-text' ) {
         $frame->Label( -text       => $comment,
                        -font       => $C_dialog_font,
                        -anchor     => 'w',
                        -justify    => 'left',
                        -wraplength => $C_min_width,
                        -background => $C_Menu_background
                      )->pack( -side => 'top', -fill => 'x', -padx => $C_padx, -pady => 0 );
      } elsif ( $type eq '-button' ) {
         my $button = $frame->Button( -command            => sub { $result = $value; $dialog->destroy() if Tk::Exists($dialog); },
                                      -text               => $comment,
                                      -image              => $image,
                                      -compound           => 'left',
                                      -font               => $C_dialog_font,
                                      -background         => $C_Menu_background,
                                      -activebackground   => $C_Menu_background,
                                      -borderwidth        => 0,
                                      -highlightthickness => 0,
                                    )->pack( -side => 'top', -anchor => 'w', -padx => $C_padx, -pady => 0 );
      } else {
         die "$Tk::Dialogs::C_illegal_menu_type_exception '$type'";
      };
   };

   $frame->pack( -side => 'top', -expand => TRUE, -fill => 'both', -padx => $border_width, -pady => $border_width );
   $dialog->after( 100, sub { IC_shift_dialog( $dialog, get_geometry( $parent ) ); } );

   $dialog->update();
   $dialog->grab();
   $dialog->waitWindow();

   return $result;

}; # menu

sub busy_waitWindow {
   my ( $dialog ) = @_;

   use Exception::Class::TryCatch;

   # in case of an exception reset to busy( FALSE )
   try eval {
      &$G_busy_command( TRUE ) if $G_busy_command;
      $dialog->update();
      $dialog->grab();
      $dialog->focus();

      $dialog->repeat( 1000 => sub { $dialog->deiconify(); $dialog->raise(); } );

      $dialog->waitWindow();
      &$G_busy_command( FALSE ) if $G_busy_command;
   };
   if ( catch my $error ) {
      &$G_busy_command( FALSE ) if $G_busy_command;
      $error->rethrow();
   };

   return;

}; # busy_waitWindow

sub show_file {
   my ( $parent, $source_file ) = @_;

   Tk::Utilities::clear_clipboard();

   my $dialog = Tk_dialog( $parent, "Source $source_file" );
   my $message = $dialog->Scrolled( 'ROText',
                                    -scrollbars => 'osoe',
                                    -height     => $C_text_height,
                                    -width      => $C_text_width,
                                    -wrap       => 'none',
                                    -background => $C_background,
                                    %C_pad_options
                                  )->pack( -side => 'top', -expand => TRUE, -fill => 'both' );
   $message->Subwidget("xscrollbar")->configure( -width => $C_scrollbar_width );
   $message->Subwidget("yscrollbar")->configure( -width => $C_scrollbar_width );

   my $SRC;
   my @messages = ();
   if ( open $SRC, '<', $source_file ) {
      @messages = <$SRC>;
      close $SRC;
   } else {
      @messages = ( $! );
   };

   foreach my $line ( @messages ) {
      $message->insert( 'end', $line );
   };

   return $dialog;

}; # show_file

sub get_geometry {
   my ( $parent )= @_;

   my $x = $parent->pointerx();
   my $y = $parent->pointery();
   unless ( $x and $y ) {
      $x = $parent->rootx();
      $y = $parent->rooty();
   };
   $x = $parent->screenwidth()  - 200 if $x + 200 > $parent->screenwidth();
   $y = $parent->screenheight() - 100 if $y + 100 > $parent->screenheight();
   $x = 0 unless $x > 0;
   $y = 0 unless $y > 0;

   return( $x, $y );

}; # get_geometry

################################################################################
# end Tk::Dialogs
################################################################################

1;
