=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de www.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

Tk::Help - module

=head1 SYNOPSIS

   use Tk::Help;

   $dialog = Tk::Help::help_dialog( $parent, $title, $image, $legend, $image_width, $image_height, $icon_image_name, $label, $balloon_message );
   Tk::Help::insert_help( $dialog, @comments );
   Tk::Help::insert_Tk_help( $dialog, @comments );
   Tk::Help::insert_documentation( @comments );
   Tk::Help::insert_help_image( $dialog, $image, $legend, $image_width, $image_height );
   Tk::Help::generate_html_contents();
   Tk::Help::set_title( $title );
   Tk::Help::set_language( $language );
   Tk::Help::set_ignore_image_path( $pattern );
   Tk::Help::set_image_path( $path );
   Tk::Help::set_icon_path( $path );
   Tk::Help::set_header( $html_text );
   Tk::Help::set_footer( $html_text );
   Tk::Help::generate_html( $file );
   Tk::Help::bind_help( $help_command, @widgets );
   if ( Tk::Help::generating_html_contents() ) 

=head1 ABSTRACT

Tk::Help provides help for Tk and HTML.

=head1 DESCRIPTION

Tk::Help provides operations to create help widgets.
The contents of the widgets can be generated into a HTML-file.

There are two types of subprograms provided by Tk::Help.

1. Creating Tk help windows

2. Generating HTML version of help.

=head3 Creating Tk help windows

A help window is created by a sequence like

   sub my_help {
      my ( $parent ) = @_;

      my $dialog = Tk::Help::help_dialog( $parent,                                       # $parent
                                          "My Title",                                    # $title
                                          'My_Screenshot.png',                           # $image
                                          'Legend of the screen shot'                    # $legend
                                          undef, undef,                                  # $image_width, $image_height
                                          'IconImage',                                   # $icon_image_name
                                          'my_help'                                      # $label
                                        );

      Tk::Help::insert_help( $dialog, EMPTY, $any_text );

      Tk::Help::insert_help( $dialog, EMPTY,
         [ sub { your_help( $dialog ) }, 'Description of the link', 'your_help' ]
      );

      Tk::Help::insert_help( $dialog, EMPTY, [ caption => "A heading" ] );

      Tk::Help::insert_help( $dialog, [ image => 'balloon' ], "A line with preceding image (icon)" );
      Tk::Help::insert_help( $dialog, [ bold => 'X' ],
         [ sub { another_help( $dialog ) }, "A line with preceding bold text 'X'", 'another_help' ]
      );

      Tk::Help::insert_help( $dialog, @colums_of_a_row );

      return $dialog;

   }; # my_help

The help dialog is structured like table with 1 to 4 columns.
Every call of insert_help inserts a new row.

=head3 Generating HTML

The HTML version of the help is generated when the help subprograms are called in the sequence they
shall be occur in the HTML file.

The code is like

   # create $top_widget etc of yout perl/Tk application

   Tk::Help::generate_html_contents();
   Tk::Help::set_title( 'LogicBuddy' );
   Tk::Help::set_header( $header );
   Tk::Help::set_footer( $footer );
   Tk::Help::set_icon_path( '../icons/' );
   Tk::Help::set_translator( \&IC_translate );
   Tk::Help::generate_html( 'Documentation.html' );
   Tk::Help::set_language( 'German' );

   my_help( $top_widget );
   your_help( $top_widget );
   another_help( $top_widget );

   # Run your perl/Tk application

   Tk::MainLoop;

=head2 SUBPROGRAMS

=head3 help_dialog

   $dialog = Tk::Help::help_dialog( $parent, $title, $image, $legend, $image_width, $image_height, $icon_image_name, $label, $balloon_message );

Creates a new top level dialog referring to the $parent widget with $title.
If $image is defined it will be packed under the title.
If the $legend is defined it will be under the $image.
The parameters $image_width are $image_height optional.
The $icon_image_name will be used for icon_image of the dialog.
The $label is necessary to identify the help in the HTML text.
The optional $balloon_message will be over the image displayed.

=head3 insert_help

   Tk::Help::insert_help( $dialog, @comments );

insert_help puts help comments into the help $dialog.
The list @comments describe a row of comment columns.
Each element of @comments is a text or an anonymous list.

The anonymous list has the structure [ type => $parameter ].

   [ head => 'Text with heading font' ]

   [ caption => 'Text with caption font' ]

   [ bold    => 'Text with bold font' ]

   [ fixed   => 'Text with fixed font' ]

   [ COLOR   => 'Text with color font' ] COLOR is one of red magenta green blue forestgreen firebrick mediumblue.

   [ image   => name_of_image ] inserts an image name_of_image.png or name_of_image.gif ... in the image paths

   [ html => 'text', 'URL' ] inserts a link to URL into HTML. The link is described by text.

   [ sub { another_help( $dialog ) }, 'Description', 'Label' ]

sub inserts a link calling the subprogram. In HTML a link to 'Label' will be generated.
The 'Label' is the label of the dialog called.

   Tk::Help::insert_help( $dialog, EMPTY, [ caption => "A heading" ] );

Inserts a row text with a empty first column followed by a caption text.

   Tk::Help::insert_help( $dialog, [ image => 'balloon' ], "A line with preceding image (icon)" );

inserts a row with a balloon icon in the first column and a text in the second.

   Tk::Help::insert_help( $dialog, [ bold => 'X' ],
      [ sub { another_help( $dialog ) }, "A line with preceding bold text 'X'", 'another_help' ]
   );

Inserts a row with a bold X and a link to another_help.   

=head3 insert_help_image

   Tk::Help::insert_help_image( $dialog, $image, $legend, $image_width, $image_height );

Inserts a image with optional $legend, $image_width and $image_height:

=head3 insert_Tk_help

   Tk::Help::insert_Tk_help( $dialog, @comments );

is like insert_help but has only effect on Tk.

=head3 insert_documentation

   Tk::Help::insert_documentation( @comments );

Is like insert_help but has only effect on HTML.

=head3 generate_html_contents

   Tk::Help::generate_html_contents();

Activates the generation of HTML text.

=head3 set_title

   Tk::Help::set_title( $title );

Defines the $title of the HTML file generated by generate_html.

An error will occur when set_title is not called before generate_html.

=head3 set_language

   Tk::Help::set_language( $language );

Defines the $language used in the HTML text.

=head3 set_ignore_image_path

   Tk::Help::set_ignore_image_path( $pattern );

Defines a $pattern substituted by the image or icon path.

=head3 set_image_path

   Tk::Help::set_image_path( $path );

Defines the image path.

=head3 set_icon_path

   Tk::Help::set_icon_path( $path );

Defines the icon path.

=head3 set_header set_footer

   Tk::Help::set_header( $header );
   Tk::Help::set_footer( $footer );

Define the header and footer of the HTML text.
The header will be inserted just behind <body> and footer just before </body>.

=head3 set_translator

   Tk::Help::set_translator( \&IC_translator );

   $translated_text = &$IC_translator( $text );

Will be called some internally texts in Tk::Help to a language.
The language translated to depends on IC_translator and is independent of set_language.

=head3 bind_help

   Tk::Help::bind_help( $help_command, @widgets );

binds the $help_command to <Key-F1> and <Button-3> of all @widgets and the
widgets packed or gridded into any widget.
If <Key-F1> or <Button-3> is already bound to a widget then $help_command is not bound.

=head3 generate_html

   Tk::Help::generate_html( $file );

Generates the HTML text produced by into a file.

An error will occur if the title is not defined by set_title.

=head3 generating_html_contents

   if ( Tk::Help::generating_html_contents() ) 

is true if generate_html_contents was called.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de www.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;

use File::HomeDir;
use Locale::Language;

use Tk;
use Tk::Images;

use Tk::StandardConfiguration;

use Tk::Utilities;
use Tk::Dialogs;

package Tk::Help;

################################################################################
# exceptions
################################################################################

################################################################################
# global constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;
use constant EMPTY => '';

use Readonly;

Readonly my $C_standard_width     => $Tk::Configuration::C_standard_width;
#                                                        Try and error
Readonly my $C_help_width         => $C_standard_width - 100;
Readonly my %C_grid_options       => ( -sticky => 'nw',
                                       -padx   => 1,
                                       -pady   => 1,
                                     );

Readonly my $C_background         => $Tk::Configuration::C_background;
Readonly my %C_background         => ( -background => $Tk::Configuration::C_background );

Readonly my %C_standard           => ( -anchor  => 'nw',
                                       -justify => 'left',
                                       %C_background
                                     );

Readonly my @C_colors             => qw( red magenta green blue forestgreen firebrick mediumblue slategrey );
Readonly my $C_color_alternatives => join ' | ', @C_colors;
Readonly my $C_color_pattern      => qr{ \A ( $C_color_alternatives ) \z }xms;

Readonly my @C_dialog_minsize     => ( $C_standard_width, $Tk::Configuration::C_scrolled_height );

################################################################################
# global variable
################################################################################

my $G_language           = 'English';
my $G_Title              = undef;
my $G_Header             = undef;
my $G_Footer             = undef;
my $G_Description        = undef;
my $G_Keywords           = undef;
my $G_Author             = undef;
my $G_help_html          = undef;
my $G_help_table         = FALSE;
my %G_references         = ();
my $G_help_html_contents = FALSE;
my @G_html               = ();
my @G_contents           = ();
my $G_ignore_image_path  = undef;
my $G_image_path         = 'images';
my $G_icon_path          = 'icons';
my $G_Translator         = undef;

################################################################################
# internal subprograms
################################################################################

sub I_translate {
   my ( $text ) = @_;

   if ( $G_Translator ) {
      return &$G_Translator( $text );
   } else {
      $text =~ s/_/ /gxms;
      return $text;
   };

}; # I_translate

################################################################################
# HTML
################################################################################

sub I_open_row {
   push @G_html, "<tr>\n";
   return;
}; # I_open_row

sub I_close_row {
   push @G_html, "</tr>\n";
   return;
}; # I_close_row

sub I_insert_image {
   my ( $image ) = @_;

   if ( $G_ignore_image_path ) {
      $image =~ s{ \A $G_ignore_image_path }{$G_icon_path}xms;
   } elsif ( $G_icon_path ) {
      $image =~ s{ \A ( [.] / (icons/)? )? }{$G_icon_path}xms;
   };
   push @G_html, "   <td valign=\"top\"><p><img src=\"$image\" class=\"adapted_img\"></p></td>\n";

   return;

}; # I_insert_image

sub I_convert_to_html {
   my ( $text ) = @_;

   my $C_tab_lenght = 10;

   $text =~ s/&/&amp;/gxms;
   $text =~ s/"/&quot;/gxms;
   $text =~ s/</&lt;/gxms;
   $text =~ s/>/&gt;/gxms;
   while ( $text =~ m/ ( (?: \A | \n ) [^\n]*? ) \t /gxms ) {
      my $tab = '&nbsp;' x ( $C_tab_lenght - length($1) );
      $text =~ s/ ( (?: \A | \n ) [^\n]*? ) \t /$1$tab/xms;
   };
   $text =~ s/ \A (.*?) \t /&nbsp;&nbsp;&nbsp;/gxms;
   $text =~ s/ \n /<br>/gxms;

   return $text;

}; # I_convert_to_html

sub I_insert_reference {
   my ( $reference, $text ) = @_;

   if ( $G_help_html ) {
      my $margin = '';
      if ( $text =~ s/ \A ( \s* ) //xms ) {
         $margin = I_convert_to_html( $1 );
      };
      $text = I_convert_to_html( $text );
      if ( $reference ) {
         $reference =~ s/ \s+ /_/gxms;
         $G_references{$reference}{reference} = $text;
         push @G_html, "   <td valign=\"middle\"><p>$margin<a href=\"#$reference\">$text</a></p></td>\n";
      } else {
         push @G_html, "   <td valign=\"middle\"><p>$margin$text</p></td>\n";
      };
   };

   return;

}; # I_insert_reference

sub I_insert_html {
   my ( $reference, $text ) = @_;

   if ( $G_help_html ) {
      $text = I_convert_to_html( $text );
      if ( $reference ) {
         push @G_html, "   <td valign=\"middle\"><p><a href=$reference>$text</a></p></td>\n";
      } else {
         push @G_html, "   <td valign=\"middle\"><p>$text</p></td>\n";
      };
   };

   return;

}; # I_insert_html

sub I_insert_head {
   my ( $text ) = @_;
   $text = I_convert_to_html( $text );
   push @G_html, "   <td valign=\"middle\"><h2>$text</h2></td>\n";
   return;
}; # I_insert_head

sub I_insert_caption {
   my ( $text ) = @_;
   $text = I_convert_to_html( $text );
   push @G_html, "   <td valign=\"middle\"><h3>$text</h3></td>\n";
   return;
}; # I_insert_caption

sub I_insert_bold {
   my ( $text ) = @_;
   $text = I_convert_to_html( $text );
   push @G_html, "   <td valign=\"middle\"><p class=\"bold\">$text</p></td>\n";
   return;
}; # I_insert_bold

sub I_insert_fixed {
   my ( $text ) = @_;
   $text = I_convert_to_html( $text );
   push @G_html, "   <td valign=\"middle\"><pre>$text</pre></td>\n";
   return;
}; # I_insert_fixed

sub I_insert_colour {
   my ( $colour, $text ) = @_;
   $text = I_convert_to_html( $text );
   push @G_html, "   <td valign=\"middle\"><p class=\"bold\" style=\"color:$colour;\">$text</p></td>\n";
   return;
}; # I_insert_colour

sub I_insert_text {
   my ( $text ) = @_;
   $text = I_convert_to_html( $text );
   push @G_html, "   <td valign=\"middle\"><p>$text</p></td>\n";
   return;
}; # I_insert_text

################################################################################
# internal Tk subprograms
################################################################################

sub ITk_display_image { 
   my ( $parent, $image, $title ) = @_;

   my ( $toplevel, $dialog ) = Tk::Dialogs::Tk_scrolled_dialog( $parent, $title );
   $toplevel->configure( %C_background );
   $toplevel->minsize( @C_dialog_minsize );

   eval { Tk::Utilities::BindMouseWheel( $dialog ) };

   my $label = $dialog->Label( -image  => $image,
                               -anchor => 'nw',
                               %C_background
                             )->pack( -side => 'top' );
   eval { Tk::Utilities::BindMouseWheel( $label, $dialog ) };

   $dialog->waitWindow();

   return;

}; # ITk_display_image

################################################################################
## internal callback subprograms
#################################################################################

################################################################################
## exported subprograms
#################################################################################

sub help_dialog {
   my ( $parent, $title, $image, $legend, $image_width, $image_height, $icon_image_name, $label, $balloon_message ) = @_;

   my ( $toplevel, $dialog ) = Tk::Dialogs::Tk_scrolled_dialog( $parent, $title, undef, $icon_image_name );
   $toplevel->minsize( @C_dialog_minsize );
   $dialog->configure( %C_background );
   my $balloon               = $dialog->Balloon( -background => $Tk::Configuration::C_Balloon_background,
                                                 -font       => $Tk::Configuration::C_small_font,
                                               );

   $dialog->Label( -text    => $title,
                   -font    => $Tk::Configuration::C_heading_font,
                   %C_standard
                 )->grid( %C_grid_options, -columnspan => 4 );

   Tk::Utilities::BindMouseWheel( $dialog );

   if ( $image ) {
      my $file = Tk::Images::file_of( $image );
      if ( $file ) {
         my $large_image = Tk::Images::image_of( $dialog, $file );
         my $small_image = $dialog->Photo ();
         $small_image->copy ( $large_image, -subsample => 2 );
         my $button = $dialog->Button( -image  => $small_image,
                                       -anchor => 'nw',
                                       -relief => 'flat',
                                       -borderwidth => 0,
                                       -command => sub { ITk_display_image( $parent, $large_image, $title ) },
                                       %C_background
                                     )->grid( %C_grid_options, -columnspan => 4 );
         $balloon->attach( $button, -msg => $balloon_message || 'Click to enlarge the image.' );
         Tk::Utilities::BindMouseWheel( $button, $dialog );
      };
   };

   return $dialog if not $G_help_html;

   if ( not $label ) {
      $label = $title;
      $label =~ s/ \s+ /_/gxms;
   };
   $G_references{$label}{target} = $label;

   push @G_html, "</table>\n" if $G_help_table;

   if ( $G_help_html_contents and 2 < scalar @G_html ) {
      my $to_contents = I_translate( 'to_contents' );
      push @G_html, "<ul type=\"square\"><li><a href=\"#Contents\">$to_contents</a></li></ul>\n";
   };

   push @G_html, "<span id=\"$label\"></span><h2>$title</h2>\n";
   if ( $image ) {
      if ( $G_ignore_image_path and ($image =~ s{ \A $G_ignore_image_path }{$G_image_path}xms) ) {
         # NOP
      } elsif ( $G_image_path ) {
         $image =~ s{ \A ( [.] / )? }{$G_image_path/}xms;
      };
      my $size = ( $image_width and $image_height )  ?  "width=\"$image_width\" height=\"$image_height\""  :  EMPTY;
      push @G_html, "<p><img src=\"$image\" $size class=\"adapted_img\"></p>\n";
   };
   push @G_html, "<p>$legend</p>\n" if $legend;   
   push @G_html, "<table cellspacing=\"1\">\n";

   $title =~ s/ \A Help \s+ //xms;
   push @G_contents, "<li><a href=\"#$label\">$title</a></li>\n" if $G_help_html_contents;

   $G_help_table = TRUE;

   return $dialog;

}; # help_dialog

sub insert_documentation {
   my ( @comments ) = @_;

   return if not $G_help_html;

   I_open_row();

   foreach my $comment ( @comments ) {
      if ( ref $comment and ( ref $comment eq 'Tk::Photo' )  ) {
         # I_insert_image( $comment ); nicht HTML
      } elsif ( ref $comment eq 'ARRAY' ) {
         my ( $type, $parameter, $reference ) = @$comment;
         if ( ref $type eq 'CODE' ) {
            # sub { .... }
            I_insert_reference( $reference, $parameter );
         } elsif ( $type eq 'head' ) {
            I_insert_head( $parameter );
         } elsif ( $type eq 'caption' ) {
            I_insert_caption( $parameter );
         } elsif ( $type eq 'bold' ) {
            I_insert_bold( $parameter );
         } elsif ( $type eq 'fixed' ) {
            I_insert_fixed( $parameter );
         } elsif ( $type =~ m/ $C_color_pattern /xms ) {
            my $colour = $1;
            I_insert_colour( $colour, $parameter );
         } elsif ( $type eq 'image' ) {
            my $file = Tk::Images::file_of( $parameter );
            if ( $file ) {
               I_insert_image( $file );
            } else {
               print STDERR "image file $parameter not found\n";
            };
         } elsif ( $type eq 'html' ) {
            I_insert_html( $reference, $parameter );
         } else {
           I_insert_text( $parameter );
         };
      } else {
         I_insert_text( $comment );
      };

   };

   I_close_row();

   return;

}; # insert_documentation

sub insert_Tk_help {
   my ( $dialog, @comments ) = @_;

   my @labels = ();
   foreach my $comment ( @comments ) {
      my $label;
      if ( ref $comment and ( ref $comment eq 'Tk::Photo' )  ) {
         $label = $dialog->Label( -image  => $comment,
                                  %C_standard
                                );
      } elsif ( ref $comment eq 'ARRAY' ) {
         my ( $type, $parameter, $reference ) = @$comment;
         if ( ref $type eq 'CODE' ) {
            # sub { .... }
            $label = $dialog->Label( -text       => $parameter,
                                     -font       => $Tk::Configuration::C_dialog_font,
                                     -foreground => 'blue',
                                     %C_standard
                                   );
            $label->bind( '<Button-1>' => sub { &$type( @_ ) } );
         } elsif ( $type eq 'head' ) {
            $label = $dialog->Label( -text    => $parameter,
                                     -font    => $Tk::Configuration::C_heading_dialog_font,
                                     %C_standard
                                   );
         } elsif ( $type eq 'caption' ) {
            $label = $dialog->Label( -text    => $parameter,
                                     -font    => $Tk::Configuration::C_caption_dialog_font,
                                     %C_standard
                                   );
         } elsif ( $type eq 'bold' ) {
            $label = $dialog->Label( -text    => $parameter,
                                     -font    => $Tk::Configuration::C_bold_dialog_font,
                                     %C_standard
                                   );
         } elsif ( $type eq 'fixed' ) {
            $label = $dialog->Label( -text    => $parameter,
                                     -font    => $Tk::Configuration::C_fixed_font,
                                     %C_standard
                                   );
         } elsif ( $type =~ m/ $C_color_pattern /xms ) {
            my $color = $1;
            $label     = $dialog->Label( -text       => $parameter,
                                         -font       => $Tk::Configuration::C_bold_dialog_font,
                                         -foreground => $color,
                                         %C_standard
                                       );
         } elsif ( $type eq 'image' ) {
            my $file = Tk::Images::file_of( $parameter );
            if ( $file ) {
               my $image = Tk::Images::image_of( $dialog, $file );
               $label = $dialog->Label( -image  => $image,
                                        %C_standard
                                      );
            } else {
               print STDERR "image file $parameter not found\n";
               $label = $dialog->Label( -text       => "image $parameter",
                                        -font       => $Tk::Configuration::C_dialog_font,
                                        -wraplength => $C_help_width,
                                        %C_standard
                                      );
            };
         } elsif ( $type eq 'html' ) {
            $label = $dialog->Label( -text       => $parameter,
                                     -font       => $Tk::Configuration::C_dialog_font,
                                     -wraplength => $C_help_width,
                                     %C_standard
                                   );
         } else {
            $label = $dialog->Label( -text       => $parameter,
                                     -font       => $Tk::Configuration::C_dialog_font,
                                     -wraplength => $C_help_width,
                                     %C_standard
                                   );
         };
      } else {
         $label = $dialog->Label( -text       => $comment,
                                  -font       => $Tk::Configuration::C_dialog_font,
                                  -wraplength => $C_help_width,
                                  %C_standard
                                );
      };

      # $dialog may not be scrolled. We try.
      eval { Tk::Utilities::BindMouseWheel( $label, $dialog ) };

      push @labels, $label;
   };

   my $first = shift @labels;
   $first->grid( @labels, %C_grid_options );

   return;

}; # insert_Tk_help

sub insert_help {
   my ( $dialog, @comments ) = @_;

   insert_Tk_help( $dialog, @comments );

   insert_documentation( @comments );

   return;

}; # insert_help

sub insert_help_image {
   my ( $dialog, $image, $legend, $image_width, $image_height ) = @_;

   return if not $G_help_html;

   push @G_html, "</table>\n" if $G_help_table;

   if ( $G_ignore_image_path ) {
      $image =~ s{ \A $G_ignore_image_path }{$G_image_path}xms;
   } elsif ( $G_image_path ) {
      $image =~ s{ \A ( [.] / )? }{$G_image_path}xms;
   };
   my $size = ( $image_width and $image_height )  ?  "width=\"$image_width\" height=\"$image_height\""  :  EMPTY;
   push @G_html, "<p><img src=\"$image\" $size class=\"adapted_img\"></p>\n";
   push @G_html, "<p>$legend</p>\n" if $legend;   

   push @G_html, "<table>\n" if $G_help_table;

   return;

}; # insert_help_image

sub set_title {
   ( $G_Title ) = @_;
   return;
}; # set_title

sub set_language {
   ( $G_language ) = @_;
   return;
}; # set_language

sub set_ignore_image_path {
   ( $G_ignore_image_path ) = @_;
   return;
}; # set_ignore_image_path

sub set_image_path {
   ( $G_image_path ) = @_;
   return;
}; # set_image_path

sub set_icon_path {
   ( $G_icon_path ) = @_;
   return;
}; # set_icon_path

sub set_header {
   ( $G_Header ) = @_;
   return;
}; # set_header

sub set_footer {
   ( $G_Footer ) = @_;
   return;
}; # set_footer

sub set_translator {
   ( $G_Translator ) = @_;
   return;
}; # set_translator

sub generate_html {
   my ( $file ) = @_;

   my $language_code = Locale::Language::language2code( $G_language || 'English' );

   open $G_help_html, '>:utf8', $file or die "error opening $G_help_html: $!" if $file;

   print $G_help_html '<!DOCTYPE html>',
                      "\n<html lang=\"$language_code\">\n<head>\n",
                      "<meta charset=\"utf-8\"/>\n",
                      '<link rel="stylesheet" href="tk.css"/>' . "\n",
                      '<link rel="icon" href="/favicon.ico" type="image/x-icon" />' . "\n",
                      '<link rel="shortcut icon" href="/favicon.ico" type="image/x-icon" />' . "\n",
                      '<meta name="viewport" content="width=device-width, initial-scale=1.0"/>' . "\n";
   if ( $G_Description ) {
      print $G_help_html "<meta name=\"description\" content=\"$G_Description\" />\n";
   };
   if ( $G_Keywords ) {
      print $G_help_html "<meta name=\"keywords\" content=\"$G_Keywords\" />\n";
   };
   if ( $G_Author ) {
      print $G_help_html "<meta name=\"author\" content=\"$G_Author\" />\n";
   };
   if ( $G_Title ) {
      print $G_help_html "<TITLE>$G_Title</TITLE>\n",
                         "</HEAD>\n",
                         "<BODY>\n";
      print $G_help_html "$G_Header\n" if $G_Header;
      print $G_help_html "<H1>$G_Title</H1>\n";
   } else {
     Carp::croak( "Missing Title. Did you call Tk::Help::set_title?\n" );
   };

   return;

}; # generate_html

sub bind_help;

sub bind_help {
   my ( $help_command, @widgets ) = @_;

   foreach my $widget ( @widgets ) {
      my %current_bindings = map { ($_ => TRUE) } $widget->bind();
      $widget->Tk::bind( '<Key-F1>'   => $help_command ) if not exists $current_bindings{'<Key-F1>'};
      $widget->Tk::bind( '<Button-3>' => $help_command ) if not exists $current_bindings{'<Button-3>'};
      bind_help( $help_command, $widget->packSlaves(), $widget->gridSlaves() );
   };

   return;

}; # bind_help

sub generate_html_contents {
   $G_help_html_contents = TRUE;
   return;
}; # generate_html_contents

sub generating_html_contents {
   return $G_help_html_contents;
}; # generating_html_contents

sub BEGIN {

   binmode STDERR, 'utf8';

   foreach my $root ( Cwd::cwd(), File::HomeDir->my_home(), '/', '/usr/local', '/usr/local/perl' ) {
      if ( $root and -e $root and -d "$root/Documentation" ) {
         Tk::Images::add_path( "$root/Documentation" );
         return;
      } elsif ( $root and -e $root and -d "$root/Doc" ) {
         Tk::Images::add_path( "$root/Doc" );
         return;
      };
   };

}; # BEGIN

sub END {

   if ( $G_help_html ) {

      if ( $G_help_html_contents ) {
         my $Contents = I_translate( 'contents' );
         print $G_help_html "<span id=\"Contents\"></span><h2>$Contents</h2></a>\n<ul type=\"circle\">\n";
         print $G_help_html @G_contents;
         print $G_help_html "</ul>\n<hr>\n";
      };

      print $G_help_html @G_html, "</table>\n";

      if ( $G_help_html_contents ) {
         my $to_contents = I_translate( 'to_contents' );
         print $G_help_html "<ul type=\"square\"><li><a href=\"#Contents\">$to_contents</a></li></ul>\n";
      };

      print $G_help_html "$G_Footer\n" if $G_Footer;

      print $G_help_html "</BODY>\n</HTML>\n";
      close $G_help_html;
   };

   foreach my $label ( keys %G_references ) {
      if ( (exists $G_references{$label}{reference}) and (not exists $G_references{$label}{target}) ) {
         print STDERR "$label with no target ( text = '$G_references{$label}{reference}' )\n";
      } elsif ( (exists $G_references{$label}{target}) and (not exists $G_references{$label}{reference}) ) {
         print STDERR "$label with no reference ( label = '$G_references{$label}{target}' )\n";
      };
   };

}; # END

################################################################################
# end Tk::Help
################################################################################

1;
