=pod

 Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de www.dr-k.de
 The software is licensed under the GNU Library General Public License.

=head1 SYNOPSIS:

use Tk::Images;

 Tk::Images::add_paths( @paths );

 $image_file = Tk::Images::file_of ( 'Image_Name' );
 $image      = Tk::Images::image_of ( $widget, $image_file );

 $image      = Tk::Images::image( $widget );
 $image      = Tk::Images::image( $widget, $type );

 Tk::Images::clear_cache();

=head1 DESCRIPTION:

Tk::Images provides images.

The packages Tk::PNG and TK::JPG should be installed.
Though Tk::Images does work without these packages.

 Tk::Images::add_paths( @paths );

add the @paths to the list of paths to be searched.

 $image_file = Tk::Images::file_of ( 'Image_Name' );

Returns an image file with Image_Name.
The file is searched in all paths defined by Tk::Images::add_paths and
if not found in@INC and subdirectories icons of @INC.
The file extensions gif xpm, png and jpg are supported.

 $image = Tk::Images::image_of ( $widget, $image_file );

Generates a TK::Photo object of $image_file. If there raises a problem
hen generating undef is returned. 

 Tk::Images::clear_cache();

The AUTOLAODed images are cached for faster access. This prevents to load images changed.
clear_cache deletes all cache information.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de www.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;

use Tk;
use MIME::Base64;

eval "use Tk::JPEG";
eval "use Tk::PNG";
eval "use Tk::Animation";

################################################################################

package Tk::Images;

################################################################################
# internal constants
################################################################################

################################################################################
# global variables
################################################################################

my %G_auto_cache = ();
my %G_flags      = ();
my @G_paths      = ();

################################################################################
# exported subprograms
################################################################################

sub clear_cache {
   %G_auto_cache = ();
   %G_flags      = ();
}; # clear_cache

sub add_path {
   push @G_paths, @_;
}; # add_path

sub file_of ($) {
   my ( $image ) = @_;

   return unless $image;

   return $image if ( ($image =~ m/\.(png|gif|xpm|jpg)$/) and (-f $image) );
   if ( $image =~ m/^\// ) {
      return $image if ( ($image =~ m/\.(png|gif|xpm|jpg)$/) and (-f $image) );
      return;
   };

   foreach my $path ( @G_paths, @INC ) {
      next if ref $path;
      foreach my $type ( '', '.png', '.xpm', '.jpg', '.gif' ) {
         next unless ( $type or $image =~ m/\.(png|gif|xpm|jpg|gif)$/ );
         return "$path/icons/$image$type" if ( -e "$path/icons/$image$type" ) and ( -f "$path/icons/$image$type" );
      };
   };
   foreach my $path ( @G_paths, @INC ) {
      next if ref $path;
      foreach my $type ( '', '.png', '.xpm', '.jpg', '.gif' ) {
         next unless ( $type or $image =~ m/\.(png|gif|xpm|jpg|gif)$/ );
         return "$path/$image$type"       if ( -e "$path/$image$type" ) and ( -f "$path/$image$type" );
      };
   };
   return;

}; # file_of

sub image_of {
   my ( $parent, $image_file ) = @_;

   Carp::carp( "parent widget not defined" ) if not defined $parent;
   Carp::carp( "image file not defined" ) if not defined $image_file;

   return if not $image_file;

   my $ref_image_file = ref $image_file;
   my $image;

   if ( not $ref_image_file ) {
      return unless ( -e $image_file );
      if ( $image_file =~ m/ [.]gif \z /xms ) {
         eval { $image = $parent->Animation( -format => 'gif', -file => $image_file ) };
         return $image if not $@;
      };
      eval { $image = $parent->Photo ( -file => $image_file ) };
      return if ( $@ );
   } elsif ( $ref_image_file eq 'Tk::Photo' ) {
      return $image_file;
   } elsif ( $ref_image_file eq 'GLOB' ) {
      my $data = join ('', <$image_file>);
      eval { $image = $parent->Photo ( -data => MIME::Base64::encode_base64($data) ) };
      return if ( $@ );
   };
   return $image;

}; # image_of

sub flag_of {
   my ( $parent, $language ) = @_;

   $language = 'unknown' unless defined $language;

   return $G_flags{$language} if exists $G_flags{$language};

   my $file = file_of("flag_$language");
   if ( $file ) {
      $G_flags{$language} = image_of( $parent, $file );
      return $G_flags{$language};
   } else {
      $file = file_of("flag_white");
      if ( $file ) {
         $G_flags{white} = image_of( $parent, $file );
      } else {
         $G_flags{white} = undef;
      };
      $G_flags{$language} = $G_flags{white};
      return $G_flags{white};
   };

}; # flag_of

################################################################################
# AUTOLOAD
################################################################################

sub AUTOLOAD {
   my ( $parent, $type ) = @_;

   my $program  =  $Tk::Images::AUTOLOAD;
   $program     =~ s/.*:://;
   $program    .= $type if $type;

   return $G_auto_cache{$program} if exists $G_auto_cache{$program};

   my $file = file_of($program);
   if ( $file ) {
      $G_auto_cache{$program} = image_of( $parent, File::Spec->rel2abs($file) );
      return $G_auto_cache{$program};
   } else {
      $G_auto_cache{$program} = undef;
      return;
   };

}; # AUTOLOAD

################################################################################
# end Tk::Images
################################################################################

1;
