=pod

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de www.dr-k.de
The software is licensed under the GNU Library General Public License.

=head1 NAME

Tk::StandardConfiguration - module

=head1 SYNOPSIS

   BEGIN {
      Readonly $Tk::Configuration::C_fixed_font_size 13;
   }; # BEGIN

   use Tk::StandardConfiguration;
   $font_string = LogicBuddy::Tk::Configuration::font_of( %font );
   LogicBuddy::Tk::Configuration::set_options( $top_widget );

=head1 ABSTRACT

Tk::StandardConfiguration provides Tk constants and some operations.

=head1 DESCRIPTION

Tk::StandardConfiguration provides Tk constants like $Tk::Configuration::C_background.

=head2 How to use

Tk::StandardConfiguration provides Readonly constants.
Thus it cannot be changed.

But every constant can be defined in a BEGIN
Then that constant will be used and further more it will be used in Tk::StandardConfiguration to define other constants.

For example the code

   BEGIN {
      Readonly $Tk::Configuration::C_fixed_font_size 13;
   }; # BEGIN
   use Tk::StandardConfiguration;

will change the size of the $Tk::Configuration::C_fixed_font.

A good solution is to provide a package MyProject::Tk::Configuration.
This package has that BEGIN block and uses Tk::StandardConfiguration;
All programs in MyProject use only MyProject::Tk::Configuration.

=head2 SUBPROGRAMS

The operations provided for useful for initial configuration.

=head3 font_of

   $font_string = Tk::Configuration::font_of( %font );

Returns the font string of the %font.

=head3 set_options

   Tk::Configuration::set_options( $top_widget );

sets global options of the $top_widget.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de www.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;
use warnings;

package Tk::Configuration;

################################################################################
# global constants
################################################################################

use Readonly;

# timing
eval { Readonly $Tk::Configuration::C_Standard_period           => 100 }; # ms

# size
eval { Readonly $Tk::Configuration::C_Border_size               => 4 };
eval { Readonly $Tk::Configuration::C_Border_width              => 2 };
eval { Readonly $Tk::Configuration::C_Highlight_Thickness       => 0 };
eval { Readonly $Tk::Configuration::C_Menu_length               => 20 };
eval { Readonly $Tk::Configuration::C_min_width                 => 320 };
eval { Readonly $Tk::Configuration::C_min_height                => 240 };
eval { Readonly $Tk::Configuration::C_scrollbar_width           => ( $^O eq 'MSWin32' )  ?  20  :  10 };
eval { Readonly $Tk::Configuration::C_standard_width            => 650 };
eval { Readonly $Tk::Configuration::C_standard_height           => 600 };
eval { Readonly $Tk::Configuration::C_scrolled_height           => 400 };
eval { Readonly $Tk::Configuration::C_scrolled_width            => 800 };
eval { Readonly $Tk::Configuration::C_Pad                       => 0 };
eval { Readonly %Tk::Configuration::C_Pads                      => ( -padx => $Tk::Configuration::C_Pad,
                                                                     -pady => $Tk::Configuration::C_Pad
                                                                   );
     };
eval { Readonly $Tk::Configuration::C_Pad_Standard              => 5 };
eval { Readonly %Tk::Configuration::C_Standard_Pads             => ( -padx => $Tk::Configuration::C_Pad_Standard,
                                                                     -pady => $Tk::Configuration::C_Pad_Standard
                                                                   );
     };
eval { Readonly $Tk::Configuration::C_Pad_Wide                  => 10 };
eval { Readonly %Tk::Configuration::C_pad_options               => ( -padx => $Tk::Configuration::C_Pad,
                                                                     -pady => $Tk::Configuration::C_Pad
                                                                   );
     };

# other
eval { Readonly $Tk::Configuration::C_button_relief             => 'flat' };
eval { Readonly $Tk::Configuration::C_Standard_Cursor           => 'left_ptr' };
eval { Readonly $Tk::Configuration::C_Watch_Cursor              => 'watch' };
eval { Readonly $Tk::Configuration::C_balloon_initial_wait      => 1000 }; # milliseconds

# colors
eval { Readonly $Tk::Configuration::C_error_foreground          => 'red' };
eval { Readonly $Tk::Configuration::C_for_information_color     => 'darkorange2' };
eval { Readonly $Tk::Configuration::C_info_color                => 'chartreuse3' };
eval { Readonly $Tk::Configuration::C_notice_color              => 'blue' };
eval { Readonly $Tk::Configuration::C_standard_color            => 'black' };
eval { Readonly $Tk::Configuration::C_Menu_continue_color       => 'DarkOrange' };
eval { Readonly $Tk::Configuration::C_System_background         => 'grey99' };
eval { Readonly $Tk::Configuration::C_background                => 'ivory' };
eval { Readonly $Tk::Configuration::C_Balloon_background        => 'ivory' };
eval { Readonly $Tk::Configuration::C_Menu_background           => '#efefef' };
eval { Readonly $Tk::Configuration::C_Entry_background          => 'ivory2' };
eval { Readonly $Tk::Configuration::C_selected_background       => 'LightCyan' };
eval { Readonly $Tk::Configuration::C_message_background        => 'Azure1' };

eval { Readonly $Tk::Configuration::C_Active_foreground         => 'blue' };
eval { Readonly $Tk::Configuration::C_foreground                => 'black' };

eval { Readonly %Tk::Configuration::C_Active_color              => ( -foreground       => $Tk::Configuration::C_Active_foreground,
                                                                     -activeforeground => $Tk::Configuration::C_Active_foreground
                                                                   );
     };

# fonts
eval { Readonly $Tk::Configuration::C_font_family               => ( $^O eq 'MSWin32' )  ?  'Arial'    :  'Nimbus Sans L' };
eval { Readonly $Tk::Configuration::C_fixed_font_family         => ( $^O eq 'MSWin32' )  ?  'Courier'  :  'Nimbus Mono L' };

eval { Readonly $Tk::Configuration::C_dialog_font_size          => ( $^O eq 'MSWin32' )  ?    8        :  8 };
eval { Readonly $Tk::Configuration::C_fixed_font_size           => ( $^O eq 'MSWin32' )  ?    8        :  8 };
eval { Readonly $Tk::Configuration::C_heading_font_size         => ( $^O eq 'MSWin32' )  ?   12        :  12 };
eval { Readonly $Tk::Configuration::C_menu_font_size            => ( $^O eq 'MSWin32' )  ?    8        :  8 };
eval { Readonly $Tk::Configuration::C_small_font_size           => ( $^O eq 'MSWin32' )  ?    8        :  8 };
eval { Readonly $Tk::Configuration::C_standard_font_size        => ( $^O eq 'MSWin32' )  ?   10        :  10 };

Readonly my $C_font_family                                      => $Tk::Configuration::C_font_family;
Readonly my $C_fixed_font_family                                => $Tk::Configuration::C_fixed_font_family;

Readonly my $C_dialog_font_size                                 => $Tk::Configuration::C_dialog_font_size;
Readonly my $C_fixed_font_size                                  => $Tk::Configuration::C_fixed_font_size;
Readonly my $C_heading_font_size                                => $Tk::Configuration::C_heading_font_size;
Readonly my $C_menu_font_size                                   => $Tk::Configuration::C_menu_font_size;
Readonly my $C_small_font_size                                  => $Tk::Configuration::C_small_font_size;
Readonly my $C_standard_font_size                               => $Tk::Configuration::C_standard_font_size;

eval { Readonly $Tk::Configuration::C_bold_font                 => [ $C_font_family,       $C_standard_font_size, 'bold', 'roman' ] };
eval { Readonly $Tk::Configuration::C_bold_dialog_font          => [ $C_font_family,       $C_dialog_font_size,   'bold', 'roman' ] };
eval { Readonly $Tk::Configuration::C_caption_dialog_font       => [ $C_font_family,       $C_standard_font_size, 'bold', 'roman' ] };
eval { Readonly $Tk::Configuration::C_dialog_font               => [ $C_font_family,       $C_dialog_font_size,           'roman' ] };
eval { Readonly $Tk::Configuration::C_bold_fixed_font           => [ $C_fixed_font_family, $C_fixed_font_size,    'bold', 'roman' ] };
eval { Readonly $Tk::Configuration::C_fixed_font                => [ $C_fixed_font_family, $C_fixed_font_size,            'roman' ] };
eval { Readonly $Tk::Configuration::C_heading_dialog_font       => [ $C_font_family,       $C_heading_font_size,  'bold', 'roman' ] };
eval { Readonly $Tk::Configuration::C_heading_font              => [ $C_font_family,       $C_heading_font_size,  'bold', 'roman' ] };
eval { Readonly $Tk::Configuration::C_label_font                => [ $C_font_family,       $C_small_font_size,    'bold', 'roman' ] };
eval { Readonly $Tk::Configuration::C_menufont                  => [ $C_font_family,       $C_menu_font_size,     'bold', 'roman' ] };
eval { Readonly $Tk::Configuration::C_small_font                => [ $C_font_family,       $C_small_font_size,            'roman' ] };
eval { Readonly $Tk::Configuration::C_small_bold_font           => [ $C_font_family,       $C_small_font_size,    'bold', 'roman' ] };
eval { Readonly $Tk::Configuration::C_standard_font             => [ $C_font_family,       $C_standard_font_size, 'bold', 'roman' ] };
eval { Readonly $Tk::Configuration::C_light_standard_font       => [ $C_font_family,       $C_standard_font_size,         'roman' ] };

################################################################################
# exported subprograms
################################################################################

sub set_options {
   my ( $top ) = @_;

   $top->optionAdd( "*Background*"        => $Tk::Configuration::C_Sytem_background );
   $top->optionAdd( "*Button*Relief"      => $Tk::Configuration::C_button_relief );
   $top->optionAdd( "*highlightThickness" => $Tk::Configuration::C_Highlight_Thickness );
   $top->optionAdd( "*activeForeground"   => $Tk::Configuration::C_Active_foreground );
   $top->optionAdd( "*Font"               => font_of( $Tk::Configuration::C_menufont ) );

   return;

}; # set_options

sub font_of {
   my ( $font ) = @_;

   return $font if $^O ne 'MSWin32';
   if ( defined $font->[3] ) {
      return [ -family => "{$font->[0]}", -size => $font->[1], -weight => $font->[2], -slant => $font->[3] ];
   } else {
      return [ -family => "{$font->[0]}", -size => $font->[1], -slant => $font->[2] ];
   };

}; # font_of

################################################################################
# end Tk::StandardConfiguration
################################################################################

1;
