=pod

 Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de www.dr-k.de
 The software is licensed under the GNU Library General Public License.

=head2 SYNOPSIS:

   use Tk::Utilities;

   Tk::Utilities::BindMouseWheel( $widget );
   Tk::Utilities::BindMouseWheel( $widget, $target );
   Tk::Utilities::UnBindMouseWheel( $widget );

   Tk::Utilities::bind_default_keys( $top, \$result, $value );
   Tk::Utilities::bind_default_keys( $top, undef, $value );
   Tk::Utilities::bind_default_keys( $top );

   Tk::Utilities::bind_destroy_key_values( $top, \$result, ( key => $value ) );
   Tk::Utilities::bind_destroy_key_values( $top, undef, ( key => $value ) );
   Tk::Utilities::bind_destroy_key_values( $top, $ignored, () );
   Tk::Utilities::bind_destroy_key_values( $top, $ignored );
   Tk::Utilities::bind_destroy_key_values( $top );

   Tk::Utilities::bind_keys( $top, ( key => \$command ) );
   Tk::Utilities::bind_keys( $top );
   Tk::Utilities::unbind_keys( $top );

   ($menu = $length) = Tk::Utilities::continue_menu($menu, $length, $condition, $text );
   ($menu = $length) = Tk::Utilities::continue_menu($menu, $length, $condition );
   ($menu = $length) = Tk::Utilities::continue_menu($menu, $length );

   Tk::Utilities::delete_menu_entries( $menu_button );

   Tk::Utilities::set_buttons_state( $button, $state );

   Tk::Utilities::clear_clipboard();

   Tk::Utilities::unpack_all_children( $widget );
   Tk::Utilities::ungrid_all_children( $widget );

=head1 ABSTRACT:

   Tk::Utilities provides global GUI components.

=head1 DESCRIPTION:

=head2 Mouse bindings

=head3 Tk::Utilities::BindMouseWheel

   Tk::Utilities::BindMouseWheel( $widget );

Binds the mouse wheel to $widget;

   Tk::Utilities::BindMouseWheel( $widget, $target );

Binds the mouse wheel events of $widget to $target.

=head3 Tk::Utilities::UnBindMouseWheel

   Tk::Utilities::UnBindMouseWheel( $widget );

Binds no mouse wheel to $widget;

=head2 Menu support

=head3 Tk::Utilities::continue_menu

   ($menu = $length) = Tk::Utilities::continue_menu($menu, $length, $condition, $text );
   ($menu = $length) = Tk::Utilities::continue_menu($menu, $length, $condition );
   ($menu = $length) = Tk::Utilities::continue_menu($menu, $length );

Inserts into $menu a sub menu if the menu is to long or $condition is TRUE.
The label of the sub menu is $text.
The default of $text is 'continued'.
The default of $condition is FALSE.

=head3 Tk::Utilities::delete_menu_entries

   Tk::Utilities::delete_menu_entries( $menu_button );

deletes all entries in the $menu.

=head2 KEY BINDING:

   The following operations define key bindings to widgets.
   Only the last binding on a widget is effective.

=head3 Tk::Utilities::bind_default_keys

   Tk::Utilities::bind_default_keys( $top, \$result, $value );

binds the Return and Escape key to widget $top.
If the Return key is pressed the $result is set to $value and $top destroyed.
If the Escape key is pressed the $top widget is destroyed.

   Tk::Utilities::bind_default_keys( $top );

If the Return or Escape key is pressed the $top widget is destroyed.

=head3 Tk::Utilities::bind_destroy_key_values

   Tk::Utilities::bind_destroy_key_values( $top, \$result, ( key => $value ) );

Binds keys to the widget $top.
If a key of the hash is pressed its associated $value is assigned to $result and
$top is destroyed.

   Tk::Utilities::bind_destroy_key_values( $top, undef, ( key => $value ) );

Binds keys to the widget $top.
If a key of the hash is pressed $top is destroyed.

   Tk::Utilities::bind_destroy_key_values ( $top, $ignored, () );
   Tk::Utilities::bind_destroy_key_values ( $top, $ignored );
   Tk::Utilities::bind_destroy_key_values ( $top );

Unbinds any keys bound to widget $top.

=head3 Tk::Utilities::bind_keys

   Tk::Utilities::bind_keys( $top, ( key => \&command ) );

Binds keys to the widget $top.
If a key of the hash is pressed its associated &command($top) is called.

   Tk::Utilities::bind_keys ( $top );

Unbinds any keys bound to widget $top.

=head3 Tk::Utilities::unbind_keys

   Tk::Utilities::unbind_keys( $top );

Unbinds any keys bound to widget $top.

=head2 Button support

=head3 Tk::Utilities::set_buttons_state

   Tk::Utilities::set_buttons_state( $button, $state );

Sets the $button to normal if $state otherwise to disabled.

=head2 Clipboard support

=head3 Tk::Utilities::clear_clipboard

   Tk::Utilities::clear_clipboard();

Clears the primary X11 clipboard.

=head2 Unpack and ungrid

=head3 Tk::Utilities::unpack_all_children

   Tk::Utilities::unpack_all_children( $widget );

Unpacks $widget and all its children.

=head3 Tk::Utilities::ungrid_all_children

   Tk::Utilities::ungrid_all_children( $widget );

Ungrids $widget and all its children.

=head1 AUTHOR

Heiner Kuhlmann, E<lt>hkuhlmann@dr-k.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright Heiner Kuhlmann : hkuhlmann@dr-k.de www.dr-k.de
The software is licensed under the GNU Library General Public License.

=cut

use strict;

use Tk;

use Tk::StandardConfiguration;

################################################################################
# Tk::Utilities
################################################################################

package Tk::Utilities;

################################################################################
# constants
################################################################################

use constant FALSE => 0;
use constant TRUE  => 1;

use Readonly;

Readonly my $C_Menu_length         => $Tk::Configuration::C_Menu_length;
Readonly my $C_Menu_continue_color => $Tk::Configuration::C_Menu_continue_color;

################################################################################
# exported Tk operations
################################################################################

sub BindMouseWheel {
    my ( $widget, $target ) = @_;

    $target = $widget unless $target;
    if ( $^O eq 'MSWin32' ) {
       $widget->bind( '<MouseWheel>' => [ sub { $target->yview('scroll', -($_[1] / 120) * 3, 'units') }, Tk::Ev('D') ] );
       $target->bind( '<MouseWheel>' => [ sub { $target->yview('scroll', -($_[1] / 120) * 3, 'units') }, Tk::Ev('D') ] );
       $target->focus();
    } else {
       $widget->bind( '<4>' => sub { $target->yview('scroll', -3, 'units') unless $Tk::strictMotif; } );
       $widget->bind( '<5>' => sub { $target->yview('scroll', +3, 'units') unless $Tk::strictMotif; } );
    };
    return;

}; # BindMouseWheel

sub UnBindMouseWheel {
    my ( $widget ) = @_;

    if ( $^O eq 'MSWin32' ) {
       $widget->bind( '<MouseWheel>' => undef );
    } else {
       $widget->bind( '<4>' => undef );
       $widget->bind( '<5>' => undef );
    }
    return;

}; # UnBindMouseWheel

sub bind_destroy_key_values ($;$%) {
   my ( $top, $r_result, %keys ) = @_;

   if ( scalar(keys %keys) ) {
      $top->bind ( '<Key>' => undef );
   } elsif ( ref $r_result ) {
      $top->bind ( '<Key>' => [ sub { if ( exists($keys{$_[1]}) ) {
                                         $$r_result = $keys{$_[1]};
                                         $top->destroy();
                                      };
                                    }, Tk::Ev('K')
                              ]
                 );
    } else {
      $top->bind ( '<Key>' => [ sub { if ( exists($keys{$_[1]}) ) {
                                         $top->destroy();
                                      };
                                    }, Tk::Ev('K')
                              ]
                 );
    };
   return;

}; # bind_destroy_key_values

sub unbind_keys ($) {
   my ( $top ) = @_;
   $top->bind ( '<Key>' => undef );
   return;
}; # unbind_keys

sub bind_keys ($;%) {
   my ( $top, %keys ) = @_;

   if ( scalar(keys %keys) ) {
      $top->bind ( '<Key>' => [ sub { if ( exists($keys{$_[1]}) ) {
                                         &{$keys{$_[1]}} ( $top );
                                      };
                                    }, Tk::Ev('K')
                              ]
                 );
   } else {
      $top->bind ( '<Key>' => undef );
   };
   return;

}; # bind_keys

sub bind_default_keys ($;$$) {
   my ( $top, $r_result, $value ) = @_;

   if ( ref $r_result ) {
      $top->bind ( '<Key>' => [ sub { if ( $_[1] eq 'Return' ) {
                                         $$r_result = $value;
                                         $top->destroy();
                                      } elsif ( $_[1] eq 'Escape' ) {
                                         $top->destroy();
                                      };
                                    }, Tk::Ev('K')
                             ]
                 );
   } else {
      $top->bind ( '<Key>' => [ sub {
                                    if ( $_[1] =~ m/^(Return|Escape)$/ ) {
                                       $top->destroy();
                                    };
                                  },
                             Tk::Ev('K')
                           ]
                 );
   };
   return;

}; # bind_default_keys

sub continue_menu {
   my ( $menu, $length, $condition, $text ) = @_;

   $text = 'continued' unless defined $text;

   if ( $condition or ($length > $C_Menu_length) ) {
      my $submenu = $menu->Menu( -tearoff => FALSE, -borderwidth => 1 );
      $menu->add( 'cascade',
                  -label            => "  $text",
                  -image            => Tk::Images::hint( $menu ),
                  -compound         => 'left',
                  -menu             => $submenu,
                  -foreground       => $C_Menu_continue_color,
                  -activeforeground => $C_Menu_continue_color
                );
      $menu = $submenu;
      $length = 0;
   };

   return ( $menu, $length );

}; # continue_menu

sub set_buttons_state {
   my ( $button, $state ) = @_;
   $button->configure( -state => ( $state  ?  'normal'  :  'disabled' ) ) if $button and Tk::Exists( $button );
   return;
}; # set_buttons_state

sub clear_clipboard {
   if ( $^O eq 'MSWin32' ) {
      system "echo off | clip";
   } else {
      # Unter Linux gibt es keine einfache Weise das Clipboard zu löschen.
      # workaround for openSyse 64 bit mit KDE 4 5?
      system "qdbus org.kde.klipper /klipper org.kde.klipper.klipper.clearClipboardHistory > /dev/null 2>&1";
      # sollte überall laufen aber ...
      system "xsel --clipboard --clear";
   };
   return;
}; # clear_clipboard

sub delete_menu_entries {
   my ( $menu_button ) = @_;

   return if not $menu_button;
   my $menu = $menu_button->cget( '-menu' );
   my $last = $menu->index('last');
   if ( $last ne 'none' ) {
      foreach my $index ( 0 .. $menu->index('last') ) {
         $menu->delete( $index );
      };
   };
   return;

}; # delete_menu_entries

sub unpack_all_children {
   my ( $widget ) = @_;

   foreach my $child ( $widget->packSlaves() ) {
      unpack_all_children( $child );
   };
   $widget->packForget();
   return;

}; # unpack_all_children

sub ungrid_all_children {
   my ( $widget ) = @_;

   foreach my $child ( $widget->gridSlaves() ) {
      ungrid_all_children( $child );
   };
   $widget->gridRemove();
   return;

}; # ungrid_all_children

################################################################################
# end Tk::Utilities
################################################################################

1;
